// {{{ copyright

/********************************************************************
 *
 * $Id: App.java,v 1.13 2000/07/05 14:07:43 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog;

// {{{ imports

import java.io.File;
import java.io.IOException;

import java.util.Locale;
import java.util.MissingResourceException;
import java.util.PropertyResourceBundle;
import java.util.ResourceBundle;

import de.qfs.lib.command.CommandDistributor;
import de.qfs.lib.config.Configuration;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.option.IntegerOption;
import de.qfs.lib.option.OptionSet;
import de.qfs.lib.util.MapResourceBundle;

import de.qfs.apps.qflog.logview.LogView;

// }}}

/**
 * Convenience class that gives access to the major pieces of the qflog
 * application.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.13 $
 */
public class App
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (App.class);

    /**
     * The instance.
     */
    private static App theApp = new App ();

    /**
     * The Options.
     */
    private OptionSet options;

    /**
     * The resources.
     */
    private MapResourceBundle resources;

    /**
     * The CommandDistributor.
     */
    private CommandDistributor cd = new CommandDistributor();

    /**
     * The configFile of the App.
     */
    private File configFile;

    /**
     * The serverName of the App.
     */
    private String serverName;

    /**
     * The serverHost of the App.
     */
    private String serverHost;

    // }}}

    // {{{ constructor

    /**
     * Create a new App. Private due to singleton pattern.
     */
    private App()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "App()", "");
        }
        reloadResources();
        initOptions();
    }

    // }}}

    // {{{ instance

    /**
     * Get the singleton App instance.
     *
     * @return  The single App instance.
     */
    public static App instance()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "instance()", "");
        }
        return theApp;
    }


    // }}}

    //----------------------------------------------------------------------
    // static convenience shortcuts
    //----------------------------------------------------------------------
    // {{{ getResources

    /**
     * Get the resources of the application.
     *
     * @return  The resources of the App.
     */
    public final static MapResourceBundle getResources()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getResources()", "");
        }
        return instance()._getResources();
    }

    // }}}
    // {{{ getCommandDistributor

    /**
     * Get the CommandDistributor used by the application.
     *
     * @return  The CommandDistributor.
     */
    public static CommandDistributor getCommandDistributor()
    {
        return instance()._getCommandDistributor();
    }

    // }}}
    // {{{ getOptions

    /**
     * Get the Options for the application.
     *
     * @return  The Options.
     */
    public static OptionSet getOptions()
    {
        return instance()._getOptions();
    }

    // }}}

    //----------------------------------------------------------------------
    // instance methods
    //----------------------------------------------------------------------
    // {{{ _getResources

    /**
     * Get the resources of the application.
     *
     * @return  The resources of the App.
     */
    private final MapResourceBundle _getResources()
    {
        return resources;
    }

    // }}}
    // {{{ _getCommandDistributor

    /**
     * Get the CommandDistributor used by the application.
     *
     * @return  The CommandDistributor.
     */
    private final CommandDistributor _getCommandDistributor()
    {
        return cd;
    }

    // }}}
    // {{{ _getOptions

    /**
     * Get the Options for the application.
     *
     * @return  The Options.
     */
    private final OptionSet _getOptions()
    {
        return options;
    }

    // }}}

    // {{{ reloadResources

    /**
     * (Re)load the resources for the application.
     */
    public void reloadResources ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "reloadResources()", "");
        }
        try {
            resources = new MapResourceBundle();
            ((MapResourceBundle)resources).fetchProperties
                ("/de/qfs/lib/resources/properties/qflib",
                 MapResourceBundle.class);
            ((MapResourceBundle)resources).fetchProperties
                ("/de/qfs/apps/qflog/logview/resources/properties/logview",
                 LogView.class);
            ((MapResourceBundle)resources).fetchProperties
                ("/de/qfs/apps/qflog/logview/resources/properties/loglevelview",
                 LogView.class);
            ((MapResourceBundle)resources).fetchProperties
                ("/de/qfs/apps/qflog/logview/resources/properties/logtableview",
                 LogView.class);
            ((MapResourceBundle)resources).fetchProperties
                ("/de/qfs/apps/qflog/resources/properties/logserver",
                 getClass());
        } catch (MissingResourceException ex) {
            if (logger.level >= Log.ERR) {
                logger.log("reloadResources()", ex);
            }
        }
    }

    // }}}

    // {{{ getConfigFile

    /**
     * Get the configFile of the application.
     *
     * @return  The configFile of the application.
     */
    public final File getConfigFile()
    {
        return configFile;
    }

    // }}}
    // {{{ setConfigFile

    /**
     * Set the configFile of the application.
     *
     * @param   configFile      The configFile to set.
     */
    public final void setConfigFile(File configFile)
    {
        this.configFile = configFile;
    }

    // }}}
    // {{{ getServerName

    /**
     * Get the serverName of the application.
     *
     * @return  The serverName of the application.
     */
    public final String getServerName()
    {
        return serverName;
    }

    // }}}
    // {{{ setServerName

    /**
     * Set the serverName of the application.
     *
     * @param   serverName      The serverName to set.
     */
    public final void setServerName(String serverName)
    {
        this.serverName = serverName;
    }

    // }}}
    // {{{ getServerHost

    /**
     * Get the serverHost of the application.
     *
     * @return  The serverHost of the application.
     */
    public final String getServerHost()
    {
        return serverHost;
    }

    // }}}
    // {{{ setServerHost

    /**
     * Set the serverHost of the application.
     *
     * @param   serverHost      The serverHost to set.
     */
    public final void setServerHost(String serverHost)
    {
        this.serverHost = serverHost;
    }

    // }}}

    // {{{ initOptions

    /**
     * Initialize the Options for the application.
     */
    private void initOptions()
    {
        options = new OptionSet ("LogServer Options");
        options.add(new IntegerOption ("maxClients", 4));
        options.add(new IntegerOption ("maxNamedClients", 2));
        options.add(new IntegerOption ("maxDeadClients", 2));
        options.add(new IntegerOption ("maxDeadNamedClients", 1));
        Configuration.instance().register(options);
    }

    // }}}
}
