// {{{ copyright

/********************************************************************
 *
 * $Id: LogServerImpl.java,v 1.10 2000/07/05 14:07:43 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog;

// {{{ imports

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;

import java.util.Hashtable;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.logrmi.LogListener;
import de.qfs.lib.logrmi.RemoteLogLevelListener;
import de.qfs.lib.logrmi.LogServerImplBase;

import de.qfs.apps.qflog.logview.LogLevelTreeModel;

// }}}

/**
 * An implementation of the LogServer interface.
 *
 * This particular server will bring up a window for every connected
 * client.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.10 $
 */
public class LogServerImpl
    extends LogServerImplBase
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (LogServerImpl.class);

    // }}}

    // {{{ constructors

    /**
     * Construct a new LogServerImpl.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public LogServerImpl ()
         throws RemoteException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "LogServerImpl()", "");
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The LogServer interface
    //----------------------------------------------------------------------
    // {{{ getId

    /**
     * Get an Id for a new client. Different clients may connect under the same
     * name, so this identification mechanism is required to differentiate
     * between them.
     *
     * @return  A unique handle identifying the client to the server.
     */
    public int getId(String clientName)
        throws RemoteException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getId(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "clientName: " + clientName);
        }
        return Model.instance().getNewId(clientName);
    }

    // }}}
    // {{{ getLogListener

    /**
     * Get a LogListener for a client.
     *
     * @param   id      The client identifier obtained with {@link #getId
     *                  getId}.
     *
     * @return  The LogListener
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public LogListener getLogListener (int id)
         throws RemoteException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getLogListener(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "id: " + id);
        }

        LogListenerImpl listener = Model.instance().getListener(id);
        if (listener == null) {
            listener = new LogListenerImpl (id);
            Model.instance().setListener(id, listener);
        }
        return listener;
    }

    // }}}
    // {{{ getLogLevelListener

    /**
     * Get a LogLevelListener for a client.
     *
     * @param   id      The client identifier obtained with {@link #getId
     *                  getId}.
     *
     * @return  The LogLevelListener
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public RemoteLogLevelListener getLogLevelListener (int id)
         throws RemoteException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getLogLevelListener(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "id: " + id);
        }

        try {
            LogLevelTreeModel levels = new LogLevelTreeModel ();
            RemoteLogLevelCallbackAdapter adapter =
                new RemoteLogLevelCallbackAdapter (levels);
            Model.instance().setLogLevelListener(id, adapter);
            return adapter;
        } catch (Exception ex) {
            if (logger.level >= Log.ERR) {
                logger.log("getLogListener(int)", ex);
            }
            return null;
        }
    }

    // }}}
    // {{{ disconnect

    /**
     * Disconnect the client from the server.
     *
     * @param   id      The client identifier obtained with {@link #getId
     *                  getId}.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public void disconnect(int id)
         throws RemoteException
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "disconnect(int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "id: " + id);
        }
        LogListener logListener = Model.instance().getListener(id);
        if (logListener != null) {
            logListener.disconnect();
        }

        RemoteLogLevelListener logLevelListener =
            Model.instance().getLogLevelListener(id);
        if (logLevelListener != null) {
            logLevelListener.disconnect();
        }
    }

    // }}}
}

