// {{{ copyright

/********************************************************************
 *
 * $Id: MainFrame.java,v 1.16 2000/07/05 14:07:43 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog;

// {{{ imports

import java.awt.Event;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseAdapter;
import java.awt.BorderLayout;
import java.awt.Point;

import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.KeyStroke;
import javax.swing.WindowConstants;
import javax.swing.table.JTableHeader;

import de.qfs.lib.config.Configurator;
import de.qfs.lib.config.ConfigWrapper;
import de.qfs.lib.gui.DefaultTableModelSorter;
import de.qfs.lib.gui.FocusFixer;
import de.qfs.lib.gui.SortedTableHelper;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.MapResourceBundle;

import de.qfs.apps.qflog.command.CommandFileExit;
import de.qfs.apps.qflog.command.CommandFileOpen;
import de.qfs.apps.qflog.command.CommandFileSaveAs;
import de.qfs.apps.qflog.command.CommandEditRemove;
import de.qfs.apps.qflog.command.CommandEditOptions;
import de.qfs.apps.qflog.command.CommandOpenLogView;
import de.qfs.apps.qflog.command.CommandHelpAbout;

// }}}

/**
 * This is the main Frame of the qflog application.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.16 $
 */
public class MainFrame extends JFrame
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (MainFrame.class);

    /**
     * The Configurator for the MainFrame.
     */
    private Configurator cfg = new Configurator ("LogFrame");

    /**
     * The table to display.
     */
    private JTable table;

    // }}}

    // {{{ constructor

    /**
     * Create a new MainFrame.
     */
    public MainFrame ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "MainFrame()", "");
        }
    }

    // }}}
    // {{{ initialization

    // {{{ init

    /**
     * Initialize the MainFrame.
     */
    public void init ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "init()", "");
        }
        setTitle("qflog");
        ImageIcon icon = (ImageIcon)
            App.getResources().getIcon("logServer.icon", null);
        if (icon != null) {
            setIconImage(icon.getImage());
        }

        initMenu();
        initTable();

        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing (java.awt.event.WindowEvent e) {
                App.getCommandDistributor().assignCommand
                    (new CommandFileExit (e.getSource()));
            }
        });

        cfg.add(ConfigWrapper.makeWrapper(this, "MainFrame"));
        FocusFixer.instance().addWindow(this);
    }

    // }}}
    // {{{ actions

    /**
     * Exit the application.
     */
    protected ActionListener actionFileExit = new ActionListener () {
        public void actionPerformed (ActionEvent e) {
            App.getCommandDistributor().assignCommand
                (new CommandFileExit (MainFrame.this));
        }
    };

    /**
     * Open a log file.
     */
    protected ActionListener actionFileOpen = new ActionListener () {
        public void actionPerformed (ActionEvent e) {
            App.getCommandDistributor().assignCommand
                (new CommandFileOpen (MainFrame.this, MainFrame.this));
        }
    };

    // protected ActionListener actionFileSave = new ActionListener () {
        // public void actionPerformed (ActionEvent e) {
            // App.getCommandDistributor().assignCommand
                // (new CommandFileSave (MainFrame.this));
        // }
    // };

    /**
     * Save a log file.
     */
    protected ActionListener actionFileSaveAs = new ActionListener () {
        public void actionPerformed (ActionEvent e) {
            App.getCommandDistributor().assignCommand
                (new CommandFileSaveAs (MainFrame.this, MainFrame.this,
                                        table));
        }
    };

    /**
     * Remove a listener.
     */
    protected ActionListener actionEditRemove = new ActionListener () {
        public void actionPerformed (ActionEvent e) {
            App.getCommandDistributor().assignCommand
                (new CommandEditRemove (MainFrame.this, MainFrame.this,
                                        table));
        }
    };

    /**
     * Edit the Options.
     */
    protected ActionListener actionEditOptions = new ActionListener () {
        public void actionPerformed (ActionEvent e) {
            App.getCommandDistributor().assignCommand
                (new CommandEditOptions (MainFrame.this, MainFrame.this));
        }
    };

    /**
     * Open a view for one listener.
     */
    protected ActionListener actionOpenLogView = new ActionListener () {
        public void actionPerformed (ActionEvent e) {
            App.getCommandDistributor().assignCommand
                (new CommandOpenLogView (MainFrame.this, table));
        }
    };

    /**
     * Show version information
     */
    protected ActionListener actionHelpAbout = new ActionListener () {
        public void actionPerformed (ActionEvent e) {
            App.getCommandDistributor().assignCommand
                (new CommandHelpAbout (MainFrame.this, MainFrame.this));
        }
    };

    // }}}
    // {{{ menus

    // {{{ initMenu

    /**
     * Initialize the menus.
     */
    protected void initMenu ()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "initMenu()", "");
        }
        JMenuBar mb = new JMenuBar ();
        setJMenuBar(mb);

        // File
        JMenu mFile = makeMenu("file");
        mb.add(mFile);

        JMenuItem mi = makeMenuItem ("fileOpen", actionFileOpen);
        mFile.add(mi);

        mi = makeMenuItem ("fileSaveAs", actionFileSaveAs);
        mFile.add(mi);

        mi = makeMenuItem ("fileExit", actionFileExit);
        mFile.add(mi);

        // Edit
        JMenu mEdit = makeMenu("edit");
        mb.add(mEdit);

        mi = makeMenuItem ("editShow", actionOpenLogView);
        mEdit.add(mi);
        mi.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0));

        mi = makeMenuItem ("editRemove", actionEditRemove);
        mEdit.add(mi);
        mi.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));

        mi = makeMenuItem ("editOptions", actionEditOptions);
        mEdit.add(mi);

        // Edit
        JMenu mHelp = makeMenu("help");
        mb.add(mHelp);

        mi = makeMenuItem ("helpAbout", actionHelpAbout);
        mHelp.add(mi);

    }

    // }}}
    // {{{ makeMenu

    /**
     * Build a menu from information in the application's resources.
     *
     * @param   resource        The base name for the resource Strings to use.
     *
     * @return  The new JMenu.
     */
    protected JMenu makeMenu (String resource)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "makeMenu(String)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "resource: " + resource);
        }
        MapResourceBundle rb = App.instance().getResources();

        JMenu menu = new JMenu (rb.getString("menu." + resource + ".name",
                                             "MISSING: " + resource));
        String mnemo = rb.getString("menu." + resource + ".mnemonic", null);
        if (mnemo != null) {
            menu.setMnemonic(mnemo.charAt(0));
        }
        return menu;
    }

    // }}}
    // {{{ makeMenuItem

    /**
     * Build a menu item from information in the application's resources.
     *
     * @param   resource        The base name for the resource Strings to use.
     * @param   action          The action to associate with the menu item.
     *
     * @return  The new JMenuItem.
     */
    protected JMenuItem makeMenuItem (String resource, ActionListener action)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "makeMenuItem(String,ActionListener)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "resource: " + resource + ", " +
                       "action: " + action);
        }
        MapResourceBundle rb = App.instance().getResources();

        JMenuItem mi = new JMenuItem
            (rb.getString("action." + resource + ".name",
                          "MISSING NAME: " + resource));
        String mnemo = rb.getString("action." + resource + ".mnemonic", null);
        if (mnemo != null) {
            mi.setMnemonic(mnemo.charAt(0));
        }
        mi.setActionCommand
            (rb.getString("action." + resource + ".command",
                          "MISSING COMMAND: " + resource));
        mi.addActionListener(action);

        return mi;

    }

    // }}}

    // }}}
    // {{{ initTable

    /**
     * Initialize the table.
     */
    private void initTable()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "initTable()", "");
        }

        getContentPane().setLayout(new BorderLayout ());

        table = new ViewTable (Model.instance());

        JTableHeader header = new JTableHeader (table.getColumnModel()) {
            public boolean isFocusTraversable()
            {
                return false;
            }
        };
        header.setRequestFocusEnabled(false);
        table.setTableHeader(header);

        SortedTableHelper helper = new SortedTableHelper (table);
        helper.prepareTable();
        helper.saveSelection(true);
        cfg.add(ConfigWrapper.makeWrapper(table, "Table"));
        DefaultTableModelSorter sorter =
            (DefaultTableModelSorter) helper.getTableModelSorter();
        sorter.setConfigurableName("sorter");
        cfg.add(sorter);

        JPanel bg = new JPanel ();
        bg.setLayout(new BorderLayout ());
        bg.setBackground(table.getBackground());
        bg.add(new JScrollPane (table), BorderLayout.CENTER);
        getContentPane().add(bg, BorderLayout.CENTER);

        table.registerKeyboardAction
            (actionOpenLogView, "open",
             KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0),
             JComponent.WHEN_IN_FOCUSED_WINDOW);

        table.addMouseListener(new MouseAdapter () {
            public void mousePressed(MouseEvent e) {
                if (e.getClickCount() > 1) {
                    int row = table.rowAtPoint(new Point (e.getX(), e.getY()));
                    if (row >= 0) {
                        App.getCommandDistributor().assignCommand
                            (new CommandOpenLogView (e.getSource(), table,
                                                     row));
                    }
                }
            }
        });
    }

    // }}}

    // }}}

    // {{{ getTable

    /**
     * Get the table.
     *
     * @return  The table.
     */
    public final JTable getTable()
    {
        return table;
    }

    // }}}
    // {{{ getConfigurator

    /**
     * Get the configurator of the LogFrame.
     *
     * @return  The configurator of the LogFrame.
     */
    public final Configurator getConfigurator()
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getConfigurator()", "");
        }
        return cfg;
    }

    // }}}

}
