// {{{ copyright

/********************************************************************
 *
 * $Id: RemoteLogLevelCallbackAdapter.java,v 1.1 2000/07/04 13:47:31 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog;

// {{{ imports

import java.rmi.RemoteException;

import de.qfs.lib.log.Log;
import de.qfs.lib.log.LogLevelCallback;
import de.qfs.lib.log.LogLevelEvent;
import de.qfs.lib.log.LogLevelListener;
import de.qfs.lib.logrmi.RemoteLogLevelCallback;
import de.qfs.lib.logrmi.RemoteLogLevelListenerImplBase;

// }}}

/**
 *
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.1 $
 */
public class RemoteLogLevelCallbackAdapter
    extends RemoteLogLevelListenerImplBase
    implements LogLevelCallback
{
    // {{{ variables

    /**
     * Maximum number of failures.
     */
    private final static int MAX_FAILURES = 3;

    /**
     * The LogLevelListener to forward LogLevelEvents to.
     */
    LogLevelListener listener;

    /**
     * The RemoteLogLevelCallback to forward callbacks to.
     */
    RemoteLogLevelCallback callback;

    /**
     * Number of failures when calling the remote callback. When this reaches
     * MAX_FAILURES, the callback is disconnected.
     */
    private int failures;

    // }}}

    // {{{ RemoteLogLevelCallbackAdapter

    /**
     * Create a new RemoteLogLevelAdapter.
     *
     * @throws  RemoteException
     */
    public RemoteLogLevelCallbackAdapter()
        throws RemoteException
    {
    }

    /**
     * Create a new RemoteLogLevelAdapter.
     *
     * @param   listener The LogLevelListener to forward LogLevelEvents to.
     *
     * @throws  RemoteException
     */
    public RemoteLogLevelCallbackAdapter(LogLevelListener listener)
        throws RemoteException
    {
        this.listener = listener;
    }

    // }}}

    // {{{ getListener

    /**
     * Get the LogLevelListener of the RemoteLogLevelCallbackAdapter.
     *
     * @return  The listener.
     */
    public LogLevelListener getListener()
    {
        return listener;
    }

    // }}}

    //----------------------------------------------------------------------
    // The LogLevelCallback interface
    //----------------------------------------------------------------------
    // {{{ getLogLevels

    /**
     * Get the current log levels from the callback. This method is needed to
     * synchronize the LogLevelListener and the LogLevelCallback after the
     * connection has been established.
     *
     * @return  An object array that contains an alternating sequence of
     *          class/package names and log levels. The levels are Integers
     *          that may be null for classes for which a level has been
     *          requested but not explicitly set.
     */
    public Object[] getLogLevels()
    {
        try {
            if (callback != null) {
                return callback.getLogLevels();
            }
        } catch (RemoteException ex) {
            failed();
        }
        return null;
    }

    // }}}
    // {{{ setLogLevel

    /**
     * Callback method for a LogLevelListener to change the log level for
     * a class or package.
     *
     * @param   source  The listener that causes the change. It will not be
     *                  notified to avoid recursion.
     * @param   name    The name of the affected class or package.
     * @param   level   The new log level.
     */
    public void setLogLevel(String name, int level)
    {
        try {
            if (callback != null) {
                callback.setLogLevel(name, level);
            }
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}
    // {{{ removeLogLevel

    /**
     * Callback method for a LogLevelListener to remove the log level for
     * a class or package.
     *
     * @param   source  The listener that causes the change. It will not be
     *                  notified to avoid recursion.
     * @param   name    The name of the affected class or package.
     */
    public void removeLogLevel(String name)
    {
        try {
            if (callback != null) {
                callback.removeLogLevel(name);
            }
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}
    // {{{ getOutputLevel

    /**
     * Get the current output log level from the callback's application.
     *
     * @return  The current output log level.
     */
    public int getOutputLevel()
    {
        try {
            if (callback != null) {
                return callback.getOutputLevel();
            }
        } catch (RemoteException ex) {
            failed();
        }
        return Log.DEFAULT_OUTPUT_LEVEL;
    }

    // }}}
    // {{{ setOutputLevel

    /**
     * Set the output log level for the callback's application.
     *
     * @param level     Output log level to set.
     */
    public void setOutputLevel(int level)
    {
        try {
            if (callback != null) {
                callback.setOutputLevel(level);
            }
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}
    // {{{ getPreQueueLevel

    /**
     * Return the current pre-queue level from the callback's application.
     *
     * @return  The current pre-queue level.
     */
    public int getPreQueueLevel()
    {
        try {
            if (callback != null) {
                return callback.getPreQueueLevel();
            }
        } catch (RemoteException ex) {
            failed();
        }
        return Log.DEFAULT_PRE_QUEUE_LEVEL;
    }

    // }}}
    // {{{ setPreQueueLevel

    /**
     * Set the pre-queue level for the callback's application.
     *
     * @param level     The pre-queue level to set.
     */
    public void setPreQueueLevel(int level)
    {
        try {
            if (callback != null) {
                callback.setPreQueueLevel(level);
            }
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}
    // {{{ isQueueing

    /**
     * Check whether the callback's application is using a log queue.
     *
     * @return  Whether the callback's application is queueing log entries.
     *
     * @see     Log#isQueueing
     */
    public boolean isQueueing()
    {
        try {
            if (callback != null) {
                return callback.isQueueing();
            }
        } catch (RemoteException ex) {
            failed();
        }
        return true;
    }

    // }}}
    // {{{ setQueueing

    /**
     * Set whether the callback's application should use a log queue.
     *
     * @param   queue   Whether to queue log entries.
     *
     * @see     Log#setQueueing
     */
    public void setQueueing(boolean queue)
    {
        try {
            if (callback != null) {
                callback.setQueueing(queue);
            }
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}
    // {{{ getQueueSize

    /**
     * Get the size of log queue of the callback's application.
     *
     * @return  The size of the log queue.
     *
     * @see     Log#getQueueSize
     */
    public int getQueueSize()
    {
        try {
            if (callback != null) {
                return callback.getQueueSize();
            }
        } catch (RemoteException ex) {
            failed();
        }
        return Log.DEFAULT_QUEUE_SIZE;
    }

    // }}}
    // {{{ setQueueSize

    /**
     * Set the size of log queue of the callback's application.
     *
     * @param   size    The size of the log queue.
     *
     * @see     Log#setQueueSize
     */
    public void setQueueSize(int size)
    {
        try {
            if (callback != null) {
                callback.setQueueSize(size);
            }
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}
    // {{{ isDropOnOverflow

    /**
     * Check whether the callback's application is dropping entries when
     * its log queue overflows.
     *
     * @return  Whether entries are dropped on overflow.
     *
     * @see     Log#isDropOnOverflow
     */
    public boolean isDropOnOverflow()
    {
        try {
            if (callback != null) {
                return callback.isDropOnOverflow();
            }
        } catch (RemoteException ex) {
            failed();
        }
        return false;
    }

    // }}}
    // {{{ setDropOnOverflow

    /**
     * Set whether the callback's application should drop entries when
     * its log queue overflows.
     *
     * @param   drop    Whether to drop entries on overflow.
     *
     * @see     Log#setDropOnOverflow
     */
    public void setDropOnOverflow(boolean drop)
    {
        try {
            if (callback != null) {
                callback.setDropOnOverflow(drop);
            }
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}
    // {{{ getFlushBufferSize

    /**
     * Get the size of the flush buffer of the callback's application.
     *
     * @return  The flush buffer size or 0 to indicate no flush buffer.
     */
    public int getFlushBufferSize()
    {
        try {
            if (callback != null) {
                return callback.getFlushBufferSize();
            }
        } catch (RemoteException ex) {
            failed();
        }
        return Log.DEFAULT_FLUSH_BUFFER_SIZE;
    }

    // }}}
    // {{{ setFlushBufferSize

    /**
     * Set the size of the flush buffer of the callback's application. Setting
     * it to 0 will turn the flush buffer off.
     *
     * @param   size    The size of the flush buffer.
     */
    public void setFlushBufferSize(int size)
    {
        try {
            if (callback != null) {
                callback.setFlushBufferSize(size);
            }
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}
    // {{{ getFlushTriggerLevel

    /**
     * Get the level that will trigger a flush of the messages saved in the
     * flush buffer of the callback's application.
     *
     * @return  The flush buffer's trigger level.
     */
    public int getFlushTriggerLevel()
    {
        try {
            if (callback != null) {
                return callback.getFlushTriggerLevel();
            }
        } catch (RemoteException ex) {
            failed();
        }
        return Log.DEFAULT_FLUSH_TRIGGER_LEVEL;
    }

    // }}}
    // {{{ setFlushTriggerLevel

    /**
     * Set the level that will trigger a flush of the messages saved in the
     * flush buffer of the callback's application.
     *
     * @param   level   The trigger level to set.
     */
    public void setFlushTriggerLevel(int level)
    {
        try {
            if (callback != null) {
                callback.setFlushTriggerLevel(level);
            }
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}
    // {{{ getPostFlushSize

    /**
     * Get the number of messages to pass unfiltered through the pre-queue
     * stage after a flush happened in the callback's application.
     *
     * @return  The number of messages to pass.
     */
    public int getPostFlushSize()
    {
        try {
            if (callback != null) {
                return callback.getPostFlushSize();
            }
        } catch (RemoteException ex) {
            failed();
        }
        return Log.DEFAULT_POST_FLUSH_SIZE;
    }

    // }}}
    // {{{ setPostFlushSize

    /**
     * Set the number of messages to pass unfiltered through the pre-queue
     * stage after a flush happened in the callback's application.
     *
     * @param   size    The number of messages to pass.
     */
    public void setPostFlushSize(int size)
    {
        try {
            if (callback != null) {
                callback.setPostFlushSize(size);
            }
        } catch (RemoteException ex) {
            failed();
        }
    }

    // }}}
    // {{{ disconnect

    /**
     * End the connection with the RemoteLogLevelListener.
     *
     * @throws  RemoteException If something RMI specific goes wrong.
     */
    public void disconnect()
         throws RemoteException
    {
    }

    // }}}

    //----------------------------------------------------------------------
    // The LogLevelListener interface
    //----------------------------------------------------------------------
    // {{{ classAdded

    /**
     * Notify the listener that a Logger for a formerly unknown class has
     * requested its log level, adding the class to the log level tree.
     *
     * @param   event   The LogLevelEvent holding the details.
     */
    public void classAdded(LogLevelEvent event)
    {
        if (listener != null) {
            listener.classAdded(event);
        }
    }

    // }}}
    // {{{ levelChanged

    /**
     * Notify the listener that a log level has been changed.
     *
     * @param   event   The LogLevelEvent holding the details.
     */
    public void levelChanged(LogLevelEvent event)
    {
        if (listener != null) {
            listener.levelChanged(event);
        }
    }

    // }}}
    // {{{ levelRemoved

    /**
     * Notify the listener that a log level has been unset.
     *
     * @param   event   The LogLevelEvent holding the details.
     */
    public void levelRemoved(LogLevelEvent event)
    {
        if (listener != null) {
            listener.levelRemoved(event);
        }
    }

    // }}}
    // {{{ setLogLevelCallback

    /**
     * Set the listener to forward LogLevelEvents to.
     *
     * @param   callback The RemoteLogLevelCallback to forward callbacks to.
     */
    public void setLogLevelCallback(RemoteLogLevelCallback callback)
    {
        this.callback = callback;
        if (listener != null && callback != null) {
            listener.setLogLevelCallback(this);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ failed

    /**
     * Take care of RMI failures.
     */
    private void failed()
    {
        if (++failures >= MAX_FAILURES) {
            // it doesn't make any sense to try to disconnect from the callback
            callback = null;
        }
    }

    // }}}
}
