// {{{ copyright

/********************************************************************
 *
 * $Id: Start.java,v 1.29 2000/07/05 14:07:43 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog;

// {{{ imports

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import java.net.InetAddress;
import java.net.UnknownHostException;

import java.rmi.AccessException;
import java.rmi.ConnectException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.ServerException;
import java.rmi.registry.Registry;
import java.rmi.registry.LocateRegistry;

import java.util.Hashtable;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.UIManager;
import javax.swing.SwingUtilities;

import de.qfs.lib.config.Configuration;
import de.qfs.lib.config.Configurator;
import de.qfs.lib.gui.EventQueue;
import de.qfs.lib.gui.Message;
import de.qfs.lib.gui.SplashWindow;
import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.logrmi.LogClient;
import de.qfs.lib.logrmi.RemoteLogFilter;
import de.qfs.lib.option.Option;
import de.qfs.lib.util.ArgsParser;
import de.qfs.lib.util.MapResourceBundle;
import de.qfs.lib.util.MissingParameterException;
import de.qfs.lib.util.UnexpectedParameterException;
import de.qfs.lib.util.UnknownOptionException;

import de.qfs.apps.qflog.logview.LogFrame;
import de.qfs.apps.qflog.command.CommandFileOpen;
import de.qfs.apps.qflog.command.CommandFileSaveAs;

// }}}

/**
 * Main entry for the logserver.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.29 $
 */
public class Start
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (Start.class);

    /**
     * File name for the config file.
     */
    private final static String CFG_NAME =
        System.getProperty ("user.home") +
        File.separator + ".qflog" + File.separator + "config";

    /**
     * Default RMI registry host name.
     */
    //  private final static String SERVER_HOST = "localhost";

    /**
     * Default name for the log server.
     */
    private final static String SERVER_NAME = "qflog";

    /**
     * The command line options.
     */
    private static Hashtable options;

    /**
     * The main Configurator for the application.
     */
    private static Configurator cfg;

    /**
     * The first non option argument.
     */
    private static int argstart;

    // }}}

    // {{{ main

    /**
     * Main point of entry for the logserver.
     *
     * @param   args    The arguments for the server.
     */
    public static void main(final String[] args)
    {
        Log.setQueueing(false);
        Log.setOutputLevel(2);

        // play it safe
        new JDK11BugThread ().start();

        // for the tab hack
        EventQueue.install();

        new Splash(App.getResources());

        // Resources may be needed to handle errors from now on
        Message.setResources(App.getResources());
        Option.setResources(App.getResources());

        // get the options and patch them up
        getOpts(args);

        if (options.containsKey("noquery")) {
            // override
            options.remove("query");
        } else if (! options.containsKey("query")) {
            // set default value
            options.put("query", "");
        }

        // check for -version
        if (options.containsKey("version")) {
            Version.main(new String[0]);
            System.exit(0);
        }

        // set Look and Feel
        try {
            UIManager.setLookAndFeel
                (UIManager.getSystemLookAndFeelClassName());
        } catch (Exception ex) {
            if (logger.level >= Log.ERR) {
                logger.log ("main(String[])", ex);
            }
        }

        SwingUtil.prepareKeymaps();

        if (options.containsKey("logserver")) {
            try {
                String ls = (String) options.get("logserver");
                RemoteLogFilter.logRemote(ls, "qflog");
                Log.setPreQueueLevel(10);
            } catch (Exception ex) {
                //  Log.setLevel(10);
                //  Log.setPreQueueLevel(10);
                if (logger.level >= Log.WRN) {
                    logger.log(Log.WRN, "main(String[])", ex);
                }
            }
        } else {
            //  Log.setLevel(10);
            //  Log.setPreQueueLevel(10);
            //  Log.setQueueing(false);
        }
        Logger.setLogLevels(options);

        // restore Configuration
        String configFile = CFG_NAME;
        if (options.containsKey("configfile")) {
            configFile = (String) options.get("configfile");
        } else {
            // Create ~/.qflog if necessary
            File tmp = new File
                (System.getProperties().getProperty("user.home") +
                 File.separator + ".qflog");
            if (tmp.exists()) {
                if (!tmp.isDirectory()) {
                    // not perfect - but so what
                    File xtmp = new File
                        (System.getProperties().getProperty("user.home") +
                         File.separator + ".xqflogx");
                    //  try {
                        tmp.renameTo(xtmp);
                        tmp.mkdir();
                        xtmp.renameTo(new File (tmp, "config"));
                    //  } catch (IOException ex) {
                        //  if (logger.level >= Log.ERR) {
                            //  logger.log("main(String[])", ex);
                        //  }
                    //  }
                }
            } else {
                //  try {
                    tmp.mkdir();
                //  } catch (IOException ex) {
                    //  if (logger.level >= Log.ERR) {
                        //  logger.log("main(String[])", ex);
                    //  }
                //  }
            }
        }
        try {
            FileInputStream fis = new FileInputStream (configFile);
            Configuration.instance().restore (fis);
            fis.close();
        } catch (Exception ex) {
            if (logger.level >= Log.ERR) {
                logger.log("main(String[])", ex);
            }
        }

        // get the server hostname
        String serverhost;
        if (options.containsKey("serverhost")) {
            serverhost = (String) options.get("serverhost");
        } else {
            try {
                serverhost = InetAddress.getLocalHost().getHostName();
            } catch (UnknownHostException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("main(String[])", ex);
                }
                // this should still work locally
                serverhost = "";
            }
        }
        if (options.containsKey("port")) {
            serverhost = serverhost + ":" + options.get("port");
        }

        // get the server name
        String servername = SERVER_NAME;
        if (options.containsKey("servername")) {
            servername = (String) options.get("servername");
        }

        LogServerImpl server = null;
        if (!options.containsKey("noserver")
            || options.containsKey("query")) {
            try {
                server = new LogServerImpl();
            } catch (RemoteException ex) {
                // shouldn't happen
                if (logger.level >= Log.ERR) {
                    logger.log("main(String[])", ex);
                }
            }
        }

        // start the RMI server
        Registry reg = null;
        if (!options.containsKey("noserver")) {
            registerServer(server, servername, options);
        }

        // initialize the App
        App.instance().setConfigFile(new File (configFile));
        // make server name known
        App.instance().setServerHost(serverhost);
        App.instance().setServerName(servername);

        // workaround for very weird class load deadlock
        // doesn't really help much, but doesn't hurt either
        new javax.swing.JScrollPane();

        Model model = Model.instance();
        cfg = new Configurator("top");

        // YASDP - Yet Another Stupid Deadlock Prevention
        App.getResources().getIcon("logServer.icon", null);

        SwingUtil.invokeAndWait(new Runnable() {
            public void run() {
                final MainFrame frame = new MainFrame ();
                frame.init();
                cfg.add(frame.getConfigurator());
                frame.pack();
                cfg.register ();
                frame.setVisible(true);
                boolean load = argstart < args.length;
                for (;argstart < args.length; argstart++) {
                    App.getCommandDistributor().assignCommand
                        (new CommandFileOpen (frame, frame,
                                              new File (args[argstart])));
                }
                if (load) {
                    frame.getTable().setRowSelectionInterval(0, 0);
                    frame.getTable().requestFocus();
                }
                SplashWindow.getLastSplashWindow().remove();
            }
        });

        // Query for waiting clients
        if (server != null && options.containsKey("query")) {
            queryClients(server, options);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // helper methods
    //----------------------------------------------------------------------
    // {{{ getOpts

    /**
     * Parse the options from the command line arguments.
     *
     * @param   args    The arguments.
     */
    private static void getOpts (String[] args)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getOpts(String[])",
                       logger.level < Log.MTDDETAIL ? "" :
                       "args: " + args);
        }
        ArgsParser ap = new ArgsParser (new String[] {
            "+configfile",
            "+logserver",
            "-nocreateregistry",
            "-noquery",
            "-noserver",
            "+optionfile",
            "+port",
            "m?query",
            "+serverhost",
            "+servername",
            "-version"
        });
        ap.setDefaultOption("optionfile");
        try {
            argstart = ap.parse(args);
            options = ap.getOptions();
        } catch (MissingParameterException ex) {
            System.err.println(Message.format("error.args.missingParameter",
                                              new Object[] {ex.getOption()}));
            System.err.println(App.getResources().getString
                               ("error.args.usage", ""));
            System.exit(1);
        } catch (UnexpectedParameterException ex) {
            System.err.println(Message.format("error.args.unexpectedParamter",
                                              new Object[] {ex.getOption()}));
            System.err.println(App.getResources().getString
                               ("error.args.usage", ""));
            System.exit(1);
        } catch (UnknownOptionException ex) {
            System.err.println(Message.format("error.args.unknownOption",
                                              new Object[] {ex.getOption()}));
            System.err.println(App.getResources().getString
                               ("error.args.usage", ""));
            System.exit(1);
        }
    }

    // }}}
    // {{{ registerServer

    /**
     * Register the server with the registry.
     *
     * @param   server  The LogServer to bind.
     * @param   servername The name to bind under.
     * @param   options The command line options.
     */
    private static void registerServer(LogServerImpl server,
                                       String servername, Hashtable options)
    {
        Registry reg = null;
        try {
            int port = Registry.REGISTRY_PORT;
            if (options.containsKey("port")) {
                port = Integer.parseInt((String) options.get("port"));
            }
            server = new LogServerImpl();
            try {
                reg = LocateRegistry.getRegistry(port);
                reg.rebind(servername, server);
                if (logger.level >= Log.MSG) {
                    logger.log(Log.MSG, "main(String[])",
                               "LogServer bound");
                }
            } catch (RemoteException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("main(String[])", ex);
                }
                if (! options.containsKey("nocreateregistry")) {
                    try {
                        reg = LocateRegistry.createRegistry(port);
                        if (logger.level >= Log.DBG) {
                            logger.log(Log.DBG, "main(String[])",
                                       "registry created at port " + port);
                        }
                        if (reg != null) {
                            reg.rebind(servername, server);
                            if (logger.level >= Log.MSG) {
                                logger.log(Log.MSG, "main(String[])",
                                           "LogServer bound");
                            }
                        }
                    } catch (RemoteException ex2) {
                        if (logger.level >= Log.ERR) {
                            logger.log("main(String[])", ex2);
                        }
                    }
                }
            }
        } catch (Exception ex) {
            if (logger.level >= Log.ERR) {
                logger.log("main(String[])", ex);
            }
        }
    }

    // }}}
    // {{{ queryClients

    /**
     * Query the RMI-Registries for clients.
     *
     * @param   server  The LogServer to pass to the clients.
     * @param   options The command line options.
     */
    private static void queryClients (LogServerImpl server, Hashtable options)
    {
        Object opt = options.get("query");
        String[] query =
            (opt instanceof String[])
            ? (String[]) opt
            : new String[] {(String) opt};

        for (int i = 0; i < query.length; i++) {
            try {
                int pStart = query[i].indexOf(':');
                String host = query[i].length() > 0 ? query[i] : null;
                int port = -1;
                if (pStart >= 0) {
                    host = pStart == 0 ? null : query[i].substring(0, pStart);
                    try {
                        port = Integer.parseInt
                            (query[i].substring(pStart + 1));
                    } catch (NumberFormatException ex) {
                    }
                }
                Registry reg = LocateRegistry.getRegistry(host, port);

                // check registry for waiting clients
                String[] names = reg.list();
                for (int j = 0; j < names.length; j++) {
                    if (names[j].startsWith(LogClient.REGISTRY_BASE_NAME)) {
                        try {
                            LogClient client = (LogClient)
                                reg.lookup(names[j]);
                            client.connect(server);
                        } catch (ServerException ex) {
                            if (logger.level >= Log.ERR) {
                                logger.log("main(String[])", ex);
                            }
                        } catch (ConnectException ex) {
                            if (logger.level >= Log.ERR) {
                                logger.log("main(String[])", ex);
                            }
                            // stale client
                            try {
                                reg.unbind(names[j]);
                            } catch (RemoteException ex2) {
                                // maybe non-local, so be it
                            } catch (NotBoundException ex2) {
                                // that would be amazing, but it's possible
                            }
                        } catch (Exception ex) {
                            if (logger.level >= Log.ERR) {
                                logger.log("main(String[])", ex);
                            }
                        }
                    }
                }
            } catch (Exception ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("main(String[])", ex);
                }
            }
        }
    }

    // }}}

    // {{{ class JDK11BugThread

    /**
     * This class helps avoid a garbage collection bug in JDK 1.1.
     */
    private static class JDK11BugThread extends Thread
    {
        Class bug5 = de.qfs.apps.qflog.logview.LogFrame.class;
        Class bug6 = de.qfs.lib.gui.Message.class;

        public JDK11BugThread()
        {
            setDaemon(true);
            setName("JDK1.1 Bug Workaroud Thread");
        }

        public void run () {
            while (true) {
                try {
                    // Sleep one hour at a time
                    sleep(3600000);
                } catch (InterruptedException ex) {
                }
            }
        }

    }

    // }}}

    // {{{ class Splash

    /**
     * Splash Window for qflog.
     */
    private static class Splash
        extends SplashWindow
    {
        // {{{ Splash

        /**
         * Create a new Splash window.
         *
         * @param       rb      The resources to get the icon from.
         */
        public Splash(MapResourceBundle rb)
        {
            super(100, 100);
            Icon icon = rb.getIcon ("qflog.splash.icon", null);
            if (icon != null) {
                panel.setBorder(BorderFactory.createRaisedBevelBorder());
                panel.add(new JLabel (icon));
                panel.remove(bar);
                setSize(panel.getPreferredSize());
                SwingUtil.centerWindow(this);
                show();
            } else {
                remove();
            }
        }

        // }}}
    }

    // }}}
}
