// {{{ copyright

/********************************************************************
 *
 * $Id: TestClient.java,v 1.5 2000/07/05 14:07:43 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog;

// {{{ imports

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import java.util.Hashtable;

import javax.swing.JButton;
import javax.swing.JFrame;

import de.qfs.lib.config.Configuration;
import de.qfs.lib.config.ConfigWrapper;
import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.log.StreamFilter;
import de.qfs.lib.logrmi.RemoteLogFilter;
import de.qfs.lib.util.ArgsParser;
import de.qfs.lib.util.MalformedArgumentListException;

import de.qfs.apps.qflog.logview.LogView;

// }}}

/**
 * This little tool is used to demonstrate and test qflog.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.5 $
 */
public class TestClient
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (TestClient.class);

    /**
     * The default client name.
     */
    private final static String CLIENT_NAME = "testclient";

    /**
     * The command line options.
     */
    private static Hashtable options;

    // }}}

    // {{{ main

    /**
     * Main point of entry for the TestClient.
     *
     * @param   args    The command line arguments.
     */
    public static void main (String[] args)
    {
        getOpts(args);
        Logger.setLogLevels(options);

        int outputlevel = 2;
        if (options.containsKey("outputlevel")) {
            try {
                outputlevel =
                    Integer.parseInt((String) options.get("outputlevel"));
            } catch (NumberFormatException ex) {
            }
        }
        Log.setOutputLevel(outputlevel);

        String clientname = CLIENT_NAME;
        if (options.containsKey("clientname")) {
            clientname = (String) options.get("clientname");
        }

        // create RemoteLogFilter
        String logserver = (String) options.get("logserver");
        boolean waitfor = options.containsKey("waitforserver");

        try {
            if (waitfor) {
                int port = -1;
                if (options.containsKey("port")) {
                    try {
                        port = Integer.parseInt((String) options.get("port"));
                    } catch (NumberFormatException ex) {
                    }
                }
                boolean local = ! options.containsKey("allownonlocal");
                boolean create = options.containsKey("createregistry");
                if (logserver != null) {
                    RemoteLogFilter.logRemote(logserver, clientname, port,
                                              create, local);
                } else {
                    RemoteLogFilter.logRemote(clientname, port,
                                              create, local);
                }
            } else if (logserver != null) {
                RemoteLogFilter.logRemote(logserver, clientname);
            }
        } catch (Exception ex) {
            if (logger.level >= Log.ERR) {
                logger.log("main(String[])", ex);
            }
        }

        // create logfile
        if (options.containsKey("logfile")) {
            try {
                StreamFilter.logToFile
                    (clientname, new File ((String) options.get("logfile")),
                     StreamFilter.MODE_CREATE);
            } catch (IOException ex) {
                if (logger.level >= Log.ERR) {
                    logger.log("main(String[])", ex);
                }
            }
        }

        // restore settings from a configfile?
        if (options.containsKey("configfile")) {
            try {
                FileInputStream in  =
                    new FileInputStream ((String) options.get("configfile"));
                Configuration.instance().restore(in);
                in.close();
            } catch (IOException ex) {
                // so be it
            }
        }

        JFrame frame = new JFrame ("TestClient");
        frame.setName("TestClient.frame");
        Configuration.instance().register(ConfigWrapper.makeWrapper
                                          (frame, "TestClient.frame"));
        frame.addWindowListener(new WindowAdapter () {
            public void windowClosing(WindowEvent e) {
                exit();
            }
        });

        JButton button = new JButton ("quit");
        button.setName("quit");

        button.addActionListener(new ActionListener () {
            public void actionPerformed(ActionEvent e) {
                exit();
            }
        });
        frame.getContentPane().add(button);
        frame.pack();
        frame.show();

        // bring up a LogFrame if internal logging is on.
        if (options.containsKey("internal")) {
            final String cn = clientname;
            SwingUtil.invokeAndWait(new Runnable() {
                public void run() {
                    LogView lv = LogView.logToView(cn);
                    lv.setLoggingEnabled(true);
                    lv.getFrame().show();
                }
            });
        }

        int numloggers = 3;
        if (options.containsKey("numloggers")) {
            try {
                numloggers =
                    Integer.parseInt((String) options.get("numloggers"));
            } catch (NumberFormatException ex) {
            }
        }
        //  finally start the logging threads
        for (int i = 1; i <= numloggers; i++) {
            int pkgnum = i / 2 + 1;
            int clnum = (i - 1) % 3 + 1;
            new LoggerThread("pkg_" + pkgnum +
                             ".Logger" + (char) ('A' + pkgnum - 1) +
                             "_" + clnum, i * 1000).start();
        }

    }

    // }}}

    //----------------------------------------------------------------------
    // helper methods
    //----------------------------------------------------------------------
    // {{{ exit

    /**
     * Cleanup and exit.
     */
    private static void exit()
    {
        Log.setQueueing(false);
        RemoteLogFilter.stopLogging();
        StreamFilter.stopLogging();
        // restore settings from a configfile?
        if (options.containsKey("configfile")) {
            try {
                FileOutputStream out  =
                    new FileOutputStream ((String) options.get("configfile"));
                Configuration.instance().save(out);
                out.close();
            } catch (IOException ex) {
                // so be it
            }
        }

        System.exit(0);

    }

    // }}}
    // {{{ getOpts

    /**
     * Parse the options from the command line arguments.
     *
     * @param   args    The arguments.
     */
    private static void getOpts (String[] args)
    {
        if (logger.level >= Log.MTD) {
            logger.log(Log.MTD, "getOpts(String[])",
                       logger.level < Log.MTDDETAIL ? "" :
                       "args: " + args);
        }

        ArgsParser ap = new ArgsParser (new String[] {
            "-allownonlocal",
            "+clientname",
            "+configfile",
            "-createregistry",
            "-internal",
            "+log-*",
            "+logfile",
            "+logserver",
            "+numloggers",
            "+outputlevel",
            "+port",
            "-waitforserver",
        });

        try {
            ap.parse(args);
            options = ap.getOptions();
        } catch (MalformedArgumentListException ex) {
            if (logger.level >= Log.ERR) {
                logger.log("getOpts(String[])", ex);
            }
            options = new Hashtable ();
        }
    }

    // }}}

    // {{{ class LoggerThread

    /**
     * Helper class that logs messages in a background thread.
     */
    private static class LoggerThread
        extends Thread
    {
        // {{{ variables

        /**
         * The logger to log with.
         */
        private Logger logger;

        /**
         * Delay between messages.
         */
        private int delay;

        // }}}

        // {{{ constructor

        /**
         * Create a new LoggerThread.
         *
         * @param       name    Class name for the logger.
         * @param       delay   Delay between messages.
         */
        public LoggerThread (String name, int delay)
        {
            super ("LoggerThread-" + name);
            setDaemon(true);
            logger = new Logger (name);
            this.delay = delay;
        }

        // }}}

        // {{{ run

        /**
         * Log messages.
         */
        public void run()
        {
            while (true) {
                try {
                    sleep(delay);
                } catch (InterruptedException ex) {
                }
                int lvl = (int) Math.floor(Math.random() * 10.0) + 1;
                int mtdnum = (int) Math.floor(Math.random() * 5.0) + 1;
                int msgnum = (int) Math.floor(Math.random() * 8.0) + 1;
                if (logger.level >= lvl) {
                    logger.log(lvl, "method_" + mtdnum, "message-" + msgnum);
                }
            }
        }

        // }}}
    }

    // }}}
}
