// {{{ copyright

/********************************************************************
 *
 * $Id: InternalLogLevelCallback.java,v 1.1 2000/07/04 13:47:31 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 2000 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog.logview;

// {{{ imports

import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.log.LogLevelCallback;
import de.qfs.lib.log.LogLevelListener;

// }}}

/**
 * LogLevelCallback for a LogView that is integrated directly into an
 * application.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.1 $
 */
public class InternalLogLevelCallback
    implements LogLevelCallback
{
    // {{{ variables

    /**
     * The LogLevelListener to which the callback is connected.
     */
    private LogLevelListener listener;

    // }}}

    // {{{ constructor

    /**
     * Create a new InternalLogLevelCallback.
     *
     * @param listener The LogLevelListener to which the callback is
     *          connected.
     */
    public InternalLogLevelCallback (LogLevelListener listener)
    {
        this.listener = listener;
    }

    // }}}

    //----------------------------------------------------------------------
    // The LogLevelCallback interface
    //----------------------------------------------------------------------
    // {{{ getLogLevels

    /**
     * Get the current log levels from the callback. This method is needed to
     * synchronize the LogLevelListener and the LogLevelCallback after the
     * connection has been established.
     *
     * @return  An object array that contains an alternating sequence of
     *          class/package names and log levels. The levels are Integers
     *          that may be null for classes for which a level has been
     *          requested but not explicitly set.
     */
    public Object[] getLogLevels()
    {
        return Logger.getLogLevels();
    }

    // }}}
    // {{{ setLogLevel

    /**
     * Callback method for a LogLevelListener to change the log level for
     * a class or package.
     *
     * @param   source  The listener that causes the change. It will not be
     *                  notified to avoid recursion.
     * @param   name    The name of the affected class or package.
     * @param   level   The new log level.
     */
    public void setLogLevel(String name, int level)
    {
        Logger.setLogLevel(listener, name, level);
    }

    // }}}
    // {{{ removeLogLevel

    /**
     * Callback method for a LogLevelListener to remove the log level for
     * a class or package.
     *
     * @param   source  The listener that causes the change. It will not be
     *                  notified to avoid recursion.
     * @param   name    The name of the affected class or package.
     */
    public void removeLogLevel(String name)
    {
        Logger.removeLogLevel(listener, name);
    }

    // }}}
    // {{{ getOutputLevel

    /**
     * Get the current output log level from the callback's application.
     *
     * @return  The current output log level.
     */
    public int getOutputLevel()
    {
        return Log.getOutputLevel();
    }

    // }}}
    // {{{ setOutputLevel

    /**
     * Set the output log level for the callback's application.
     *
     * @param level     Output log level to set.
     */
    public void setOutputLevel(int level)
    {
        Log.setOutputLevel(level);
    }

    // }}}
    // {{{ getPreQueueLevel

    /**
     * Return the current pre-queue level from the callback's application.
     *
     * @return  The current pre-queue level.
     */
    public int getPreQueueLevel()
    {
        return Log.getPreQueueLevel();
    }

    // }}}
    // {{{ setPreQueueLevel

    /**
     * Set the pre-queue level for the callback's application.
     *
     * @param level     The pre-queue level to set.
     */
    public void setPreQueueLevel(int level)
    {
        Log.setPreQueueLevel(level);
    }

    // }}}
    // {{{ isQueueing

    /**
     * Check whether the callback's application is using a log queue.
     *
     * @return  Whether the callback's application is queueing log entries.
     *
     * @see     Log#isQueueing
     */
    public boolean isQueueing()
    {
        return Log.isQueueing();
    }

    // }}}
    // {{{ setQueueing

    /**
     * Set whether the callback's application should use a log queue.
     *
     * @param   queue   Whether to queue log entries.
     *
     * @see     Log#setQueueing
     */
    public void setQueueing(boolean queue)
    {
        Log.setQueueing(queue);
    }

    // }}}
    // {{{ getQueueSize

    /**
     * Get the size of log queue of the callback's application.
     *
     * @return  The size of the log queue.
     *
     * @see     Log#getQueueSize
     */
    public int getQueueSize()
    {
        return Log.getQueueSize();
    }

    // }}}
    // {{{ setQueueSize

    /**
     * Set the size of log queue of the callback's application.
     *
     * @param   size    The size of the log queue.
     *
     * @see     Log#setQueueSize
     */
    public void setQueueSize(int size)
    {
        Log.setQueueSize(size);
    }

    // }}}
    // {{{ isDropOnOverflow

    /**
     * Check whether the callback's application is dropping entries when
     * its log queue overflows.
     *
     * @return  Whether entries are dropped on overflow.
     *
     * @see     Log#isDropOnOverflow
     */
    public boolean isDropOnOverflow()
    {
        return Log.isDropOnOverflow();
    }

    // }}}
    // {{{ setDropOnOverflow

    /**
     * Set whether the callback's application should drop entries when
     * its log queue overflows.
     *
     * @param   drop    Whether to drop entries on overflow.
     *
     * @see     Log#setDropOnOverflow
     */
    public void setDropOnOverflow(boolean drop)
    {
        Log.setDropOnOverflow(drop);
    }

    // }}}
    // {{{ getFlushBufferSize

    /**
     * Get the size of the flush buffer of the callback's application.
     *
     * @return  The flush buffer size or 0 to indicate no flush buffer.
     */
    public int getFlushBufferSize()
    {
        return Log.getFlushBufferSize();
    }

    // }}}
    // {{{ setFlushBufferSize

    /**
     * Set the size of the flush buffer of the callback's application. Setting
     * it to 0 will turn the flush buffer off.
     *
     * @param   size    The size of the flush buffer.
     */
    public void setFlushBufferSize(int size)
    {
        Log.setFlushBufferSize(size);
    }

    // }}}
    // {{{ getFlushTriggerLevel

    /**
     * Get the level that will trigger a flush of the messages saved in the
     * flush buffer of the callback's application.
     *
     * @return  The flush buffer's trigger level.
     */
    public int getFlushTriggerLevel()
    {
        return Log.getFlushTriggerLevel();
    }

    // }}}
    // {{{ setFlushTriggerLevel

    /**
     * Set the level that will trigger a flush of the messages saved in the
     * flush buffer of the callback's application.
     *
     * @param   level   The trigger level to set.
     */
    public void setFlushTriggerLevel(int level)
    {
        Log.setFlushTriggerLevel(level);
    }

    // }}}
    // {{{ getPostFlushSize

    /**
     * Get the number of messages to pass unfiltered through the pre-queue
     * stage after a flush happened in the callback's application.
     *
     * @return  The number of messages to pass.
     */
    public int getPostFlushSize()
    {
        return Log.getPostFlushSize();
    }

    // }}}
    // {{{ setPostFlushSize

    /**
     * Set the number of messages to pass unfiltered through the pre-queue
     * stage after a flush happened in the callback's application.
     *
     * @param   size    The number of messages to pass.
     */
    public void setPostFlushSize(int size)
    {
        Log.setPostFlushSize(size);
    }

    // }}}
}
