// {{{ copyright

/********************************************************************
 *
 * $Id: LogDetailView.java,v 1.13 2000/07/05 14:07:44 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog.logview;

// {{{ imports

import java.awt.Color;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;

import java.text.SimpleDateFormat;

import java.util.Date;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;
import javax.swing.border.Border;

import de.qfs.lib.gui.Borders;
import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.LogEntry;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.MapResourceBundle;

// }}}

/**
 * Simple view that displays the details of a LogEntry.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.13 $
 */
public class LogDetailView extends JPanel
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (LogDetailView.class);

    /**
     * DateFormat used for timestamp output.
     */
    private static SimpleDateFormat df = new SimpleDateFormat("HH:mm:ss.SSS");

    /**
     * The names of the icon resources.
     */
    private final static String iconNames[] = {
        "logTableView.icon.err", "logTableView.icon.errDetail",
        "logTableView.icon.wrn", "logTableView.icon.wrnDetail",
        "logTableView.icon.msg", "logTableView.icon.msgDetail",
        "logTableView.icon.mtd", "logTableView.icon.mtdDetail",
        "logTableView.icon.dbg", "logTableView.icon.dbgDetail"};

    /**
     * The icons to render.
     */
    private final static Icon[] icons = new Icon[iconNames.length];

    /**
     * Label to display the level.
     */
    JLabel level;

    /**
     * Label to display the timestamp.
     */
    JLabel time;

    /**
     * Label to display the thread.
     */
    JLabel thread;

    /**
     * Label to display the class.
     */
    JLabel clazz;

    /**
     * Label to display the method.
     */
    JLabel method;

    /**
     * Text area to display the message.
     */
    MessageView detail;

    /**
     * Whether logging is enabled (true) or disabled (false);
     */
    private boolean logging;

    /**
     * The resources to use.
     */
    private transient MapResourceBundle resources;

    // }}}

    // {{{ constructor

    /**
     * Create a new LogDetailView
     */
    public LogDetailView ()
    {
    }

    // }}}

    // {{{ setLoggingEnabled

    /**
     * Enable or disable logging output caused directly or indirectly by
     * LogDetailView methods. If logging is disabled, all calls to methods
     * that might create log messages will be protected with {@link
     * de.qfs.lib.log.Log#excludeThread de.qfs.lib.log.Log.excludeThread}.
     * This is necessary, if the LogFilterTreeModel is used inside the
     * application whose logs it is supposed to filter. <p>
     *
     * The default value is false.
     *
     * @param   enable  True to enable logging, false to disable it.
     */
    public final void setLoggingEnabled(boolean enable)
    {
        logging = enable;
    }

    // }}}

    // {{{ init

    /**
     * Initialize the view.
     */
    protected void init()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "init()", "");
        }

        initIcons();

        setLayout(new BorderLayout ());

        JPanel top = new JPanel ();
        top.setLayout(new BorderLayout ());
        top.setBorder(null);

        String title = resources.getString("logTableView.detail.title",
                                           "Detail");
        JLabel titleLabel = new JLabel (title);
        titleLabel.setBorder(BorderFactory.createEtchedBorder());
        top.add(titleLabel, BorderLayout.NORTH);

        add(top, BorderLayout.NORTH);

        JPanel center = new JPanel ();
        center.setBorder(BorderFactory.createCompoundBorder
                         (BorderFactory.createLoweredBevelBorder(),
                          BorderFactory.createEmptyBorder(5, 5, 5, 5)));

        GridBagLayout layout = new GridBagLayout ();
        center.setLayout(layout);

        GridBagConstraints cons = new GridBagConstraints();

        level = new JLabel (" ");
        cons.gridx = 0;
        cons.gridy = 0;
        cons.gridwidth = 1;
        cons.weightx = 1.0;
        cons.weighty = 0;
        cons.anchor = GridBagConstraints.NORTHWEST;
        layout.setConstraints(level, cons);
        center.add(level);

        time = new JLabel (" ");
        cons.gridx = 1;
        cons.gridy = 0;
        cons.gridwidth = 1;
        cons.weightx = 1.0;
        cons.weighty = 0;
        cons.anchor = GridBagConstraints.NORTHWEST;
        layout.setConstraints(time, cons);
        center.add(time);

        thread = new JLabel (" ");
        cons.gridx = 2;
        cons.gridy = 0;
        cons.gridwidth = 2;
        cons.weightx = 2.0;
        cons.weighty = 0;
        cons.anchor = GridBagConstraints.NORTHWEST;
        layout.setConstraints(thread, cons);
        center.add(thread);

        clazz = new JLabel (" ");
        cons.gridx = 0;
        cons.gridy = 1;
        cons.gridwidth = 2;
        cons.weightx = 2.0;
        cons.weighty = 0;
        cons.anchor = GridBagConstraints.NORTHWEST;
        layout.setConstraints(clazz, cons);
        center.add(clazz);

        method = new JLabel (" ");
        cons.gridx = 2;
        cons.gridy = 1;
        cons.gridwidth = 2;
        cons.weightx = 2.0;
        cons.weighty = 0;
        cons.anchor = GridBagConstraints.NORTHWEST;
        layout.setConstraints(method, cons);
        center.add(method);

        detail = new MessageView ("");
        cons.gridx = 0;
        cons.gridy = 2;
        cons.gridwidth = 4;
        cons.weightx = 4.0;
        cons.weighty = 1.0;
        cons.fill = GridBagConstraints.BOTH;
        cons.anchor = GridBagConstraints.NORTHWEST;
        layout.setConstraints(detail, cons);
        center.add(detail);

        add(center, BorderLayout.CENTER);
    }

    // }}}
    // {{{ cleanup

    /**
     * Clear everything that might prevent garbage collection.
     */
    public void cleanup()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "cleanup()", "");
        }

        SwingUtil.cleanup(this);

        detail.cleanup();
        detail = null;

        resources = null;
    }

    // }}}
    // {{{ finalize

    /**
     * Finalize the LogDetailView.
     */
    public void finalize()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "finalize()", "");
        }
    }

    // }}}

    // {{{ setResources

    /**
     * Set the resources to get the icons from.
     *
     * @param   rb      MapResourceBundle containing the resources to use.
     */
    public void setResources(MapResourceBundle rb)
    {
        resources = rb;
    }

    // }}}
    // {{{ setEntry

    /**
     * Set the entry to display.
     *
     * @param   entry   The LogEntry to display.
     */
    public void setEntry(LogEntry entry)
    {
        if (entry != null) {
            Icon icon = null;
            if (icons[0] != null
                && entry.getLevel() >= 0
                && entry.getLevel() < icons.length) {
                icon = icons[entry.getLevel() - 1];
            }
            level.setIcon(icon);
            level.setText(icon == null ? "" + entry.getLevel() : null);
            time.setText(df.format (new Date (entry.getTimestamp())));
            thread.setText(entry.getThread());
            method.setText(entry.getMethod());
            clazz.setText(entry.getClazz());
            detail.getTextArea().setText(entry.getMessage());
            detail.getTextArea().setCaretPosition(0);
        } else {
            level.setText(" ");
            time.setText(" ");
            thread.setText(" ");
            method.setText(" ");
            clazz.setText(" ");
            detail.getTextArea().setText(" ");
        }
    }

    // }}}

    // {{{ initIcons

    /**
     * Fetch the icons.
     */
    private void initIcons()
    {
        if (icons[0] == null && resources != null) {
            for (int i = 0; i < icons.length; i++) {
                icons[i] = resources.getIcon(iconNames[i], null);
            }
        }
    }

    // }}}

    // {{{ class MessageView

    // {{{ static MessageView variables

    /**
     * The Logger used for logging MessageView methods.
     */
    private final static Logger mvLogger = new Logger (MessageView.class);

    // }}}

    /**
     * Scroll pane with a text area that draws a border around itself, when
     * the text area has the focus.
     */
    private class MessageView
        extends JScrollPane
        implements FocusListener
    {
        // {{{ variables

        /**
         * The non focus border.
         */
        private Border defaultBorder;

        /**
         * The border used when the MessageView has the focus.
         */
        private Border focusBorder;

        /**
         * The JTextArea of the MessageView.
         */
        private JTextArea textArea;

        /**
         * Listner for scrolling.
         */
        private KeyAdapter scrollListener = new KeyAdapter () {
            public void keyPressed (KeyEvent e) {
                if (e.getKeyCode() == e.VK_UP &&
                    e.getModifiers() == 0) {
                    SwingUtil.scrollDownLine(MessageView.this);
                    e.consume();
                } else if (e.getKeyCode() == e.VK_DOWN &&
                           e.getModifiers() == 0) {
                    SwingUtil.scrollUpLine(MessageView.this);
                    e.consume();
                }
            }
        };

        // }}}

        // {{{ constructor

        /**
         * Create a new MessageView.
         *
         * @param       contents        The text to display in the view.
         */
        public MessageView (String contents)
        {
            textArea = new JTextArea (contents) {
                public boolean isManagingFocus()
                {
                    return false;
                }
            };
            textArea.setLineWrap(true);
            textArea.setEditable(false);
            textArea.addFocusListener(this);

            setViewportView(textArea);

            defaultBorder = BorderFactory.createCompoundBorder
                (BorderFactory.createEmptyBorder(1, 1, 1, 1),
                 BorderFactory.createLoweredBevelBorder());
            focusBorder = BorderFactory.createCompoundBorder
                (Borders.STIPPLE,
                 BorderFactory.createLoweredBevelBorder());
            this.setBorder(defaultBorder);

            textArea.addKeyListener(scrollListener);
        }

        // }}}

        // {{{ getTextArea

        /**
         * Get the JTextArea of the MessageView.
         *
         * @return      The JTextArea of the MessageView.
         */
        public JTextArea getTextArea()
        {
            return textArea;
        }

        // }}}
        // {{{ focusGained

        /**
         * Draw a focus border around the MessageView.
         */
        public void focusGained(FocusEvent e)
        {
            this.setBorder(focusBorder);
        }


        // }}}
        // {{{ focusLost

        /**
         * Remove the focus border from the MessageView.
         */
        public void focusLost(FocusEvent e)
        {
            this.setBorder(defaultBorder);
        }

        // }}}

        // {{{ cleanup

        /**
         * Clear everything that might prevent garbage collection.
         */
        public void cleanup()
        {
            if (logging && mvLogger.level >= Log.MTD) {
                mvLogger.log(Log.MTD, "cleanup()", "");
            }
            setViewportView(null);
            this.setBorder(null);
            defaultBorder = null;
            focusBorder = null;

            textArea.removeKeyListener(scrollListener);
            scrollListener = null;
            textArea.removeFocusListener(this);
            textArea = null;

            resources = null;
        }

        // }}}
        // {{{ finalize

        /**
         * Finalize the MessageView.
         */
        public void finalize()
        {
            if (logging && mvLogger.level >= Log.MTD) {
                mvLogger.log(Log.MTD, "finalize()", "");
            }
        }

        // }}}
    }

    // }}}
}
