// {{{ copyright

/********************************************************************
 *
 * $Id: LogFilterView.java,v 1.16 2000/07/05 14:07:44 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog.logview;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.util.ResourceBundle;

import javax.swing.Action;
import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.ButtonModel;
import javax.swing.Icon;
import javax.swing.KeyStroke;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;

import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.gui.WindowsTreeCellRenderer;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.util.MapResourceBundle;

// }}}

/**
 * A LogFilterView is used for display and manipulation of various filters,
 * thus modifying the display of log messages in a LogTableView. It is not
 * very useful on its own, but usually used as part of the {@link LogView
 * LogView} component, that integrates the LogFilterView with other logging
 * related views. <p>
 *
 * The LogFilterView must be initialized by a call to {@link #init init} after
 * creation. It uses its own resources for localization, which can be modified
 * via {@link #setResources setResources}. <p>
 *
 * The LogLevelView does not have any {@link Configurable Configurable} parts
 * except for its {@link LogFilterTreeModel LogFilterTreeModel}, which must be
 * handled by the caller if its state should be saved, unless the
 * LogFilterView is used as part of a LogView. <p>
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.16 $
 */
public class LogFilterView
    extends JPanel
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (LogFilterView.class);

    /**
     * The model of log levels.
     */
    private transient LogFilterTreeModel model;

    /**
     * The tree component.
     */
    private transient JTree tree;

    /**
     * The default resources to use.
     */
    private static MapResourceBundle defaultResources;

    /**
     * The resources to use.
     */
    private transient MapResourceBundle resources;

    /**
     * Popup menu for context menu.
     */
    private transient JPopupMenu popup;

    /**
     * Mouse listener for the popup menu.
     */
    private transient PopupListener popupListener;

    /**
     * Whether logging is enabled (true) or disabled (false);
     */
    private boolean logging;

    /**
     * Button model for the toggleExtraFilter menu items.
     */
    private ButtonModel toggleButtonModel;

    //----------------------------------------------------------------------
    // The actions
    //----------------------------------------------------------------------
    /**
     * Remove log levels (non recursive).
     */
    private Action removeAction;

    /**
     * Remove log levels (recursive).
     */
    private Action removeRecursiveAction;

    /**
     * Set log levels to Log.ERR
     */
    private Action setERRAction;

    /**
     * Set log levels to Log.ERRDETAIL
     */
    private Action setERRDETAILAction;

    /**
     * Set log levels to Log.WRN
     */
    private Action setWRNAction;

    /**
     * Set log levels to Log.WRNDETAIL
     */
    private Action setWRNDETAILAction;

    /**
     * Set log levels to Log.MSG
     */
    private Action setMSGAction;

    /**
     * Set log levels to Log.MSGDETAIL
     */
    private Action setMSGDETAILAction;

    /**
     * Set log levels to Log.MTD
     */
    private Action setMTDAction;

    /**
     * Set log levels to Log.MTDDETAIL
     */
    private Action setMTDDETAILAction;

    /**
     * Set log levels to Log.DBG
     */
    private Action setDBGAction;

    /**
     * Set log levels to Log.DBGDETAIL
     */
    private Action setDBGDETAILAction;

    /**
     * Toggle extra filters.
     */
    private Action toggleFilterAction;

    /**
     * Add an extra filter.
     */
    private Action addFilterAction;

    /**
     * Remove an extra filter.
     */
    private Action removeFilterAction;

    /**
     * Remove all extra filters.
     */
    private Action clearFilterAction;

    /**
     * Expand path and all descendants.
     */
    private Action expandAllAction;

    /**
     * Collapse path and all descendants.
     */
    private Action collapseAllAction;

    /**
     * Display all nodes with a non-null level.
     */
    private Action showSetLevelsAction;

    // }}}

    // {{{ constructor

    /**
     * Create a new LogFilterView.
     *
     * @param   model   The mode to use for the tree.
     */
    public LogFilterView(LogFilterTreeModel model)
    {
        this.model = model;
        setupResources();
        model.setRootName(resources.getString("logFilterView.root.name",
                                              "default"));
    }

    // }}}

    // {{{ setLoggingEnabled

    /**
     * Enable or disable logging output caused directly or indirectly by
     * LogFilterView methods. If logging is disabled, all calls to methods
     * that might create log messages will be protected with {@link
     * de.qfs.lib.log.Log#excludeThread de.qfs.lib.log.Log.excludeThread}.
     * This is necessary, if the LogFilterTreeModel is used inside the
     * application whose logs it is supposed to filter. <p>
     *
     * The default value is false.
     *
     * @param   enable  True to enable logging, false to disable it.
     */
    public final void setLoggingEnabled(boolean enable)
    {
        logging = enable;
        model.setLoggingEnabled(enable);
    }

    // }}}

    //----------------------------------------------------------------------
    // Initialization
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the components of the view. This method is not called by the
     * constructor to enable further configuration like setting the resources,
     * before the view is initialized.
     */
    public void init()
    {
        boolean excluded = !logging;
        if (excluded) {
            Log.excludeThread();
        }
        try {
            setLayout(new BorderLayout ());
            setBorder(null);

            JPanel top = new JPanel ();
            top.setLayout(new BorderLayout ());
            top.setBorder(null);

            String title = resources.getString("logFilterView.title", "Filter");
            JLabel titleLabel = new JLabel (title);
            titleLabel.setBorder(BorderFactory.createEtchedBorder());
            top.add(titleLabel, BorderLayout.NORTH);

            tree = new JTree (model) {
                //  public boolean hasBeenExpanded(TreePath path) {
                    //  return false;
                //  }
            };
            tree.setBorder(null);
            tree.setCellRenderer(new Renderer ());

            JScrollPane treePane = new JScrollPane (tree);
            SwingUtil.constrainScroll(treePane);
            treePane.setBorder(BorderFactory.createLoweredBevelBorder());
            top.add(treePane, BorderLayout.CENTER);
            add(top, BorderLayout.CENTER);

            // Default selection
            tree.setSelectionRow(0);

            // setup and bind actions
            setupActions();

            tree.registerKeyboardAction
                (removeAction, "remove",
                 KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0),
                 tree.WHEN_FOCUSED);
            tree.registerKeyboardAction
                (removeRecursiveAction, "removeRecursive",
                 KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, KeyEvent.CTRL_MASK),
                 tree.WHEN_FOCUSED);

            tree.registerKeyboardAction
                (setERRAction, "setERR",
                 KeyStroke.getKeyStroke(KeyEvent.VK_1, 0),
                 tree.WHEN_FOCUSED);
            tree.registerKeyboardAction
                (setERRDETAILAction, "setERRDETAIL",
                 KeyStroke.getKeyStroke(KeyEvent.VK_2, 0),
                 tree.WHEN_FOCUSED);
            tree.registerKeyboardAction
                (setWRNAction, "setWRN",
                 KeyStroke.getKeyStroke(KeyEvent.VK_3, 0),
                 tree.WHEN_FOCUSED);
            tree.registerKeyboardAction
                (setWRNDETAILAction, "setWRNDETAIL",
                 KeyStroke.getKeyStroke(KeyEvent.VK_4, 0),
                 tree.WHEN_FOCUSED);
            tree.registerKeyboardAction
                (setMSGAction, "MSGERR",
                 KeyStroke.getKeyStroke(KeyEvent.VK_5, 0),
                 tree.WHEN_FOCUSED);
            tree.registerKeyboardAction
                (setMSGDETAILAction, "setMSGDETAIL",
                 KeyStroke.getKeyStroke(KeyEvent.VK_6, 0),
                 tree.WHEN_FOCUSED);
            tree.registerKeyboardAction
                (setMTDAction, "setMTD",
                 KeyStroke.getKeyStroke(KeyEvent.VK_7, 0),
                 tree.WHEN_FOCUSED);
            tree.registerKeyboardAction
                (setMTDDETAILAction, "setMTDDETAIL",
                 KeyStroke.getKeyStroke(KeyEvent.VK_8, 0),
                 tree.WHEN_FOCUSED);
            tree.registerKeyboardAction
                (setDBGAction, "setDBG",
                 KeyStroke.getKeyStroke(KeyEvent.VK_9, 0),
                 tree.WHEN_FOCUSED);
            tree.registerKeyboardAction
                (setDBGDETAILAction, "setDBGDETAIL",
                 KeyStroke.getKeyStroke(KeyEvent.VK_0, 0),
                 tree.WHEN_FOCUSED);

            tree.registerKeyboardAction
                (toggleFilterAction, "filter",
                 KeyStroke.getKeyStroke(KeyEvent.VK_F, KeyEvent.CTRL_MASK),
                 tree.WHEN_FOCUSED);

            tree.registerKeyboardAction
                (addFilterAction, "addFilter",
                 KeyStroke.getKeyStroke(KeyEvent.VK_A, KeyEvent.CTRL_MASK),
                 tree.WHEN_FOCUSED);

            tree.registerKeyboardAction
                (removeFilterAction, "removeFilter",
                 KeyStroke.getKeyStroke(KeyEvent.VK_R, KeyEvent.CTRL_MASK),
                 tree.WHEN_FOCUSED);

            tree.registerKeyboardAction
                (clearFilterAction, "clearFilter",
                 KeyStroke.getKeyStroke(KeyEvent.VK_K, KeyEvent.CTRL_MASK),
                 tree.WHEN_FOCUSED);

            tree.registerKeyboardAction
                (expandAllAction, "expandAll",
                 KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, KeyEvent.ALT_MASK),
                 tree.WHEN_FOCUSED);

            tree.registerKeyboardAction
                (collapseAllAction, "collapseAll",
                 KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, KeyEvent.ALT_MASK),
                 tree.WHEN_FOCUSED);

            // popup menu
            popup = new JPopupMenu ();
            fillMenu(popup);
            popupListener = new PopupListener ();
            tree.addMouseListener(popupListener);

            showSetLevels();

        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}
    // {{{ cleanup

    /**
     * Try to remove everything that might prevent the LogFilterView from
     * being garbage collected.
     */
    public void cleanup()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "cleanup()", "");
        }

        SwingUtil.cleanup(this);

        tree.removeMouseListener(popupListener);
        popupListener = null;
        SwingUtil.cleanup(popup);
        popup = null;

        tree.resetKeyboardActions();
        tree = null;
        resources = null;

        model.cleanup();
        model = null;
    }

    // }}}
    // {{{ finalize

    /**
     * Finalize the LogFilterView.
     */
    public void finalize()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "finalize()", "");
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Access
    //----------------------------------------------------------------------
    // {{{ getModel

    /**
     * Get the view's associated LogFilterTreeModel.
     *
     * @return  The model of the view.
     */
    public final LogFilterTreeModel getModel()
    {
        return model;
    }

    // }}}
    // {{{ getTree

    /**
     * Get the JTree component of the view.
     *
     * @return  The tree.
     */
    public final JTree getTree()
    {
        return tree;
    }

    // }}}
    // {{{ fillMenu

    /**
     * Fill a JMenu with menu items for the LogFilterView.
     *
     * @param   menu    The JMenu to fill.
     */
    public void fillMenu(JMenu menu)
    {
        JMenuItem mi = menu.add(removeAction);
        mi = menu.add(removeRecursiveAction);

        menu.addSeparator();

        mi = menu.add(setERRAction);
        mi = menu.add(setERRDETAILAction);
        mi = menu.add(setWRNAction);
        mi = menu.add(setWRNDETAILAction);
        mi = menu.add(setMSGAction);
        mi = menu.add(setMSGDETAILAction);
        mi = menu.add(setMTDAction);
        mi = menu.add(setMTDDETAILAction);
        mi = menu.add(setDBGAction);
        mi = menu.add(setDBGDETAILAction);

        menu.addSeparator();

        mi = new JCheckBoxMenuItem
            ((String) toggleFilterAction.getValue(Action.NAME),
             (Icon) toggleFilterAction.getValue(Action.SMALL_ICON));
        if (toggleButtonModel == null) {
            toggleButtonModel = mi.getModel();
            toggleButtonModel.setSelected(model.isPackageFilteringEnabled());
        mi.addActionListener(toggleFilterAction);
        } else {
            mi.setModel(toggleButtonModel);
        }
        menu.add(mi);
        mi = menu.add(addFilterAction);
        mi = menu.add(removeFilterAction);
        mi = menu.add(clearFilterAction);

        menu.addSeparator();

        menu.add(showSetLevelsAction);
    }

    /**
     * Fill a JPopupMenu with menu items for the LogLevelView.
     *
     * @param   menu    The JPopupMenu to fill.
     */
    public void fillMenu(JPopupMenu menu)
    {
        menu.add(removeAction);
        menu.add(removeRecursiveAction);

        menu.addSeparator();

        menu.add(setERRAction);
        menu.add(setERRDETAILAction);
        menu.add(setWRNAction);
        menu.add(setWRNDETAILAction);
        menu.add(setMSGAction);
        menu.add(setMSGDETAILAction);
        menu.add(setMTDAction);
        menu.add(setMTDDETAILAction);
        menu.add(setDBGAction);
        menu.add(setDBGDETAILAction);

        menu.addSeparator();

        JCheckBoxMenuItem mi = new JCheckBoxMenuItem
            ((String) toggleFilterAction.getValue(Action.NAME),
             (Icon) toggleFilterAction.getValue(Action.SMALL_ICON));
        if (toggleButtonModel == null) {
            toggleButtonModel = mi.getModel();
            toggleButtonModel.setSelected(model.isPackageFilteringEnabled());
            mi.addActionListener(toggleFilterAction);
        } else {
            mi.setModel(toggleButtonModel);
        }
        menu.add(mi);
        menu.add(addFilterAction);
        menu.add(removeFilterAction);
        menu.add(clearFilterAction);

        menu.addSeparator();

        menu.add(showSetLevelsAction);
    }

    // }}}

    //----------------------------------------------------------------------
    // Configuration
    //----------------------------------------------------------------------
    // {{{ setResources

    /**
     * Override the default resources used by the LogLevelView. If this method
     * is not called, the LogLevelView will use its default resource settings.
     *
     * @param   rb      ResourceBundle containing the resources to use.
     */
    public void setResources(ResourceBundle rb)
    {
        resources.addResources(rb);
        model.setRootName(resources.getString("logFilterView.root.name",
                                              "default"));
    }


    // }}}
    // {{{ showSetLevels

    /**
     * Expand or collapse tree nodes so that all nodes with a non-null level
     * are displayed.
     */
    public void showSetLevels()
    {
        SwingUtilities.invokeLater(new Runnable () {
            public void run() {
                showSetLevelsAction.actionPerformed(null);
            }
        });
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ setupActions

    /**
     * Set up the actions for the LogLevelView.
     */
    private void setupActions()
    {
        // do this only once
        if (removeAction == null) {
            removeAction = new RemoveLevelAction (false);
            setActionProperties(removeAction, "removeLevel");
            removeRecursiveAction = new RemoveLevelAction (true);
            setActionProperties(removeRecursiveAction, "removeLevelRecursive");

            setERRAction = new ChangeLevelAction (Log.ERR);
            setActionProperties(setERRAction, "setLevelERR");
            setERRDETAILAction = new ChangeLevelAction (Log.ERRDETAIL);
            setActionProperties(setERRDETAILAction, "setLevelERRDETAIL");
            setWRNAction = new ChangeLevelAction (Log.WRN);
            setActionProperties(setWRNAction, "setLevelWRN");
            setWRNDETAILAction = new ChangeLevelAction (Log.WRNDETAIL);
            setActionProperties(setWRNDETAILAction, "setLevelWRNDETAIL");
            setMSGAction = new ChangeLevelAction (Log.MSG);
            setActionProperties(setMSGAction, "setLevelMSG");
            setMSGDETAILAction = new ChangeLevelAction (Log.MSGDETAIL);
            setActionProperties(setMSGDETAILAction, "setLevelMSGDETAIL");
            setMTDAction = new ChangeLevelAction (Log.MTD);
            setActionProperties(setMTDAction, "setLevelMTD");
            setMTDDETAILAction = new ChangeLevelAction (Log.MTDDETAIL);
            setActionProperties(setMTDDETAILAction, "setLevelMTDDETAIL");
            setDBGAction = new ChangeLevelAction (Log.DBG);
            setActionProperties(setDBGAction, "setLevelDBG");
            setDBGDETAILAction = new ChangeLevelAction (Log.DBGDETAIL);
            setActionProperties(setDBGDETAILAction, "setLevelDBGDETAIL");

            toggleFilterAction = new AbstractAction () {
                public void actionPerformed(ActionEvent e) {
                    model.togglePackageFiltering();
                    if (toggleButtonModel != null) {
                        toggleButtonModel.setSelected
                        (model.isPackageFilteringEnabled());
                    }
                }
            };
            setActionProperties(toggleFilterAction, "toggleFilter");

            addFilterAction = new AbstractAction () {
                public void actionPerformed(ActionEvent e) {
                    TreePath[] paths = tree.getSelectionPaths();

                    if (paths != null) {
                        for (int i = 0; i < paths.length; i++) {
                            FilterNode node = (FilterNode)
                                paths[i].getLastPathComponent();
                            model.addFilterPackage(node);
                        }
                    }
                }
            };
            setActionProperties(addFilterAction, "addFilter");

            removeFilterAction = new AbstractAction () {
                public void actionPerformed(ActionEvent e) {
                    TreePath[] paths = tree.getSelectionPaths();

                    if (paths != null) {
                        for (int i = 0; i < paths.length; i++) {
                            FilterNode node = (FilterNode)
                                paths[i].getLastPathComponent();
                            model.removeFilterPackage(node);
                        }
                    }
                }
            };
            setActionProperties(removeFilterAction, "removeFilter");

            clearFilterAction = new AbstractAction () {
                public void actionPerformed(ActionEvent e) {
                    model.clearFilterPackages();
                }
            };
            setActionProperties(clearFilterAction, "clearFilter");

            expandAllAction  = new AbstractAction () {
                public void actionPerformed(ActionEvent e) {
                    TreePath path = tree.getSelectionPath();
                    if (path != null) {
                        boolean excluded = ! logging;
                        if (excluded) {
                            Log.excludeThread();
                        }
                        try {
                            SwingUtil.expandAll(tree, path);
                        } finally {
                            if (excluded) {
                                Log.includeThread();
                            }
                        }
                    }
                }
            };

            collapseAllAction  = new AbstractAction () {
                public void actionPerformed(ActionEvent e) {
                    TreePath path = tree.getSelectionPath();
                    if (path != null) {
                        boolean excluded = ! logging;
                        if (excluded) {
                            Log.excludeThread();
                        }
                        try {
                            SwingUtil.collapseAll(tree, path);
                        } finally {
                            if (excluded) {
                                Log.includeThread();
                            }
                        }
                    }
                }
            };

            showSetLevelsAction = new ShowSetLevelsAction ();
            setActionProperties(showSetLevelsAction, "showSetLevels");
        }
    }

    // }}}
    // {{{ setActionProperties

    /**
     * Set name and icon for an action from the LogLevelView's resources.
     *
     * @param   action   The action to prepare.
     * @param   resource Identifier for the action's resources.
     */
    private void setActionProperties(Action action, String resource)
    {
        String name = resources.getString
            ("logFilterView.action." + resource + ".name", null);
        if (name != null) {
            action.putValue(Action.NAME, name);
        }

        String iconName = resources.getString
            ("logFilterView.action." + resource + ".icon", null);
        if (iconName != null) {
            Icon icon =  resources.getIcon (iconName, null);
            if (icon != null) {
                action.putValue(Action.SMALL_ICON, icon);
            }
        }
    }

    // }}}
    // {{{ setupResources

    /**
     * Initialize the resources from the default resources, fetching them if
     * necessary.
     */
    private void setupResources()
    {
        if (defaultResources == null) {
            defaultResources = new MapResourceBundle ();
            defaultResources.fetchProperties
                ("/de/qfs/apps/qflog/logview/resources/properties/logfilterview",
                 LogFilterView.class);
        }
        resources = defaultResources;
    }

    // }}}

    //----------------------------------------------------------------------
    // Action classes
    //----------------------------------------------------------------------
    // {{{ class ChangeLevelAction

    // {{{ static ChangeLevelAction variables

    /**
     * The Logger used for logging ChangeLevelAction methods.
     */
    private final static Logger claLogger =
        new Logger (ChangeLevelAction.class);

    // }}}

    /**
     * Change the level of selected nodes in the filter tree.
     */
    private class ChangeLevelAction
        extends AbstractAction
    {
        // {{{ variables

        /**
         * The level to set.
         */
        private int level;

        // }}}

        // {{{ constructor

        /**
         * Create a new ChangeLevelAction.
         *
         * @param       level   The level to set when the action is invoked.
         */
        public ChangeLevelAction(int level)
        {
            this.level = level;
        }

        // }}}

        // {{{ actionPerformed

        /**
         * Set the level on the selected nodes.
         *
         * @param       e       Ignored.
         */
        public void actionPerformed (ActionEvent e)
        {
            if (logging && claLogger.level >= Log.MTD) {
                claLogger.log(Log.MTD, "actionPerformed(ActionEvent)",
                           claLogger.level < Log.MTDDETAIL ? "" :
                           "level: " + level);
            }
            TreePath[] paths = tree.getSelectionPaths();

            if (paths != null) {
                for (int i = 0; i < paths.length; i++) {
                    model.setLevel(paths[i], level);
                }
            }
        }

        // }}}
    }

    // }}}
    // {{{ class RemoveLevelAction

    // {{{ static RemoveLevelAction variables

    /**
     * The Logger used for logging RemoveLevelAction methods.
     */
    private final static Logger rlaLogger =
        new Logger (RemoveLevelAction.class);

    // }}}

    /**
     * Remove the levels from the selected filter tree nodes.
     */
    private class RemoveLevelAction
        extends AbstractAction
    {
        // {{{ variables

        /**
         * Whether to recursively remove levels.
         */
        private boolean recurse;

        // }}}

        // {{{ constructor

        /**
         * Create a new RemoveLevelAction.
         *
         * @param       recurse Whether to recursively remove levels.
         */
        public RemoveLevelAction(boolean recurse)
        {
            this.recurse = recurse;
        }

        // }}}

        // {{{ actionPerformed

        /**
         * Remove the levels from the selected filter tree nodes.
         *
         * @param       e       Ignored.
         */
        public void actionPerformed (ActionEvent e)
        {
            if (logging && rlaLogger.level >= Log.MTD) {
                rlaLogger.log(Log.MTD, "actionPerformed(ActionEvent)",
                           rlaLogger.level < Log.MTDDETAIL ? "" :
                           "recurse: " + recurse);
            }
            TreePath[] paths = tree.getSelectionPaths();

            if (paths != null) {
                for (int i = 0; i < paths.length; i++) {
                    model.removeLevel(paths[i], recurse);
                }
            }
        }

        // }}}
    }

    // }}}
    // {{{ class ShowSetLevelsAction

    // {{{ static ShowSetLevelsAction variables

    /**
     * The Logger used for logging ShowSetLevelsAction methods.
     */
    private final static Logger sslaLogger =
        new Logger (ShowSetLevelsAction.class);

    // }}}
    /**
     * Expand or collapse tree nodes so that all nodes with a non-null level
     * are displayed.
     */
    private class ShowSetLevelsAction
        extends AbstractAction
    {
        // {{{ constructor

        /**
         * Create a new ShowSetLevelsAction.
         */
        public ShowSetLevelsAction()
        {
        }

        // }}}

        // {{{ actionPerformed

        /**
         * Expand/collapse tree nodes according to the levels.
         *
         * @param       e       Ignored.
         */
        public void actionPerformed (ActionEvent e)
        {
            boolean excluded = !logging;
            if (excluded) {
                Log.excludeThread();
            }
            try {
                maybeExpandPath(new TreePath (new Object[] {model.getRoot()}));
            } finally {
                if (excluded) {
                    Log.includeThread();
                }
            }
        }

        // }}}
        // {{{ maybeExpandPath

        /**
         * Recursively expand/collapse the child nodes or a node while
         * checking whether the node has any direct or indirect children with
         * a non-null level.
         *
         * @param       path    Path to the start node.
         *
         * @return      True if the node has children wit a non-null level.
         */
        private boolean maybeExpandPath(TreePath path)
        {
            if (logging && sslaLogger.level >= Log.MTD) {
                sslaLogger.log(Log.MTD, "maybeExpandPath(TreePath)",
                           sslaLogger.level < Log.MTDDETAIL ? "" :
                           "path: " + path);
            }
            Object node = path.getLastPathComponent();
            boolean isLevel = model.getLevel(node) != null;
            boolean expand = false;
            if (model.isLeaf(node)) {
                return isLevel;
            }

            int num = model.getChildCount(node);
            for (int i = 0; i < num; i++) {
                Object child = model.getChild(node, i);
                if (maybeExpandPath(path.pathByAddingChild(child))) {
                    expand = true;
                }
            }
            if (sslaLogger.level >= Log.DBG) {
                sslaLogger.log(Log.DBG, "maybeExpandPath(TreePath)",
                           "expand: " + expand);
            }

            if (expand) {
                tree.expandPath(path);
            } else {
                tree.collapsePath(path);
            }
            return expand || isLevel;
        }

        // }}}
    }

    // }}}

    //----------------------------------------------------------------------
    // Popup menu listener
    //----------------------------------------------------------------------
    // {{{ class PopupListener

    /**
     * Listener for mouseclicks that brings up a popup menu.
     */
    class PopupListener
        extends MouseAdapter
    {
        /**
         * Notify the listener that a mouse button has been pressed.
         *
         * @param       event   Details aobut the event.
         */
        public void mousePressed(MouseEvent event)
        {
            maybePopup(event);
        }

        /**
         * Notify the listener that a mouse button has been released.
         *
         * @param       event   Details aobut the event.
         */
        public void mouseReleased(MouseEvent event)
        {
            maybePopup(event);
        }

        /**
         * Check if the event is a popup trigger and bring up a popup menu if
         * required. If the click happens on a tree node that is selected, the
         * selection doesn't change. If the node is not selected, it will be
         * made the only selected node.
         *
         * @param       event   The MouseEvent containing the details.
         */
        private void maybePopup(MouseEvent event)
        {
            boolean excluded = !logging;
            if (excluded) {
                Log.excludeThread();
            }
            try {
                if (event.isPopupTrigger()) {
                    int row =
                        tree.getRowForLocation(event.getX(), event.getY());
                    if (row >= 0) {
                        if (! tree.isRowSelected(row)) {
                            tree.setSelectionRow(row);
                        }
                    }
                    SwingUtil.showPopup(popup, tree,
                                        event.getX(), event.getY(),
                                        false);
                    popup.requestFocus();
                }
            } finally {
                if (excluded) {
                    Log.includeThread();
                }
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Renderer for the tree
    //----------------------------------------------------------------------
    // {{{ class Renderer

    // {{{ static Renderer variables

    /**
     * Index of the method icon.
     */
    private final static int METHOD = 10;

    /**
     * Index of the class icon.
     */
    private final static int CLASS = 11;

    /**
     * Index of the open class icon.
     */
    private final static int CLASS_OPEN = 12;

    /**
     * Index of the package icon.
     */
    private final static int PACKAGE = 13;

    /**
     * Index of the open package icon.
     */
    private final static int PACKAGE_OPEN = 14;

    /**
     * The names of the icons resources.
     */
    private final static String iconNames[] = {
        "logFilterView.icon.err", "logFilterView.icon.errDetail",
        "logFilterView.icon.wrn", "logFilterView.icon.wrnDetail",
        "logFilterView.icon.msg", "logFilterView.icon.msgDetail",
        "logFilterView.icon.mtd", "logFilterView.icon.mtdDetail",
        "logFilterView.icon.dbg", "logFilterView.icon.dbgDetail",
        "logFilterView.icon.method",
        "logFilterView.icon.class", "logFilterView.icon.classOpen",
        "logFilterView.icon.package", "logFilterView.icon.packageOpen"
    };

    /**
     * The icons to render.
     */
    private final static Icon[] icons = new Icon[iconNames.length];

    // }}}

    /**
     * Renderer for the tree nodes. Sets package, class or method icons if no
     * level is set, or the level icon for the node's level. <p>
     *
     * If the node is used as an extra filter, it is displayed in red.
     */
    private class Renderer
        extends WindowsTreeCellRenderer
    {
        // {{{ constructor

        /**
         * Create a new Renderer.
         */
        public Renderer ()
        {
            initIcons();
        }

        // }}}

        // {{{ getTreeCellRendererComponent

       /**
         * Get the renderer component for the tree node.
         *
         * @param       tree            The tree being drawn.
         * @param       obj             The tree node.
         * @param       selected        Whether the node is selected.
         * @param       expanded        Whether the node is expanded.
         * @param       leaf            Whether the node is a leaf.
         * @param       row             The node's row index.
         * @param       hasFocus        Whether the node has the focus.
         *
         * @return      A label with the node's name and correct icon.
         */
         public Component getTreeCellRendererComponent
        (JTree tree, Object obj, boolean selected, boolean expanded,
         boolean leaf, int row, boolean hasFocus)
        {
            JLabel com = (JLabel) super.getTreeCellRendererComponent
                (tree, obj, selected, expanded, leaf, row, hasFocus);

            com.setForeground
                (((FilterNode) obj).isExtra()
                 ? Color.red :
                 (selected ? getTextSelectionColor()
                           : getTextNonSelectionColor()));
            Integer level = model.getLevel(obj);

            Icon outer;
            if (leaf) {
                outer = icons[METHOD];
            } else if ((obj instanceof FilterRootNode)
                       || ! ((FilterNode) obj).getData().getClazz()
                       .endsWith(".")) {
                outer = icons[expanded ? CLASS_OPEN : CLASS];
            } else {
                outer = icons[expanded ? PACKAGE_OPEN : PACKAGE];
            }
            if (level == null) {
                com.setIcon(outer);
                return com;
            }
            int lvl = level.intValue();
            if (lvl < Log.ERR || lvl > Log.DBGDETAIL) {
                com.setIcon(outer);
            } else {
                com.setIcon(icons[level.intValue() - 1]);
            }
            return com;
        }

        // }}}
        // {{{ initIcons

        /**
         * Fetch the icons.
         */
        private void initIcons()
        {
            if (icons[0] == null) {
                for (int i = 0; i < icons.length; i++) {
                    icons[i] = resources.getIcon(iconNames[i], null);
                }
            }
        }

        // }}}
    }

    // }}}
}
