// {{{ copyright

/********************************************************************
 *
 * $Id: LogFrame.java,v 1.22 2000/07/05 14:07:44 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog.logview;

// {{{ imports

import java.awt.Event;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;

import java.util.Observable;
import java.util.Observer;

import javax.swing.ImageIcon;
import javax.swing.Action;
import javax.swing.AbstractAction;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.KeyStroke;

import de.qfs.lib.command.CommandDistributor;
import de.qfs.lib.gui.FocusFixer;
import de.qfs.lib.gui.Message;
import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;
import de.qfs.lib.config.Configurator;
import de.qfs.lib.config.ConfigWrapper;
import de.qfs.lib.option.Option;
import de.qfs.lib.option.OptionDialog;
import de.qfs.lib.option.OptionSet;
import de.qfs.lib.util.MapResourceBundle;

import de.qfs.apps.qflog.logview.command.CommandSaveConfiguration;
import de.qfs.apps.qflog.logview.command.CommandSaveLog;

// }}}

// {{{ doc

/**
 * This is the Frame for a LogView. <p>
 *
 * The LogFrame must be initialized by a call to {@link #init init} after
 * creation. It uses the LogView's resources for localization, which can be
 * modified via {@link LogView#setResources setResources}. <p>
 *
 * The LogFrame has its own {@link de.qfs.lib.config.Configurator
 * Configurator}, that must be registered separately.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.22 $
 */

// }}}
public class LogFrame
    extends JFrame
    implements Observer
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private static Logger logger = new Logger (LogFrame.class);

    /**
     * The frame's menu bar.
     */
    private transient JMenuBar menuBar;

    /**
     * The menu for the filter.
     */
    private transient JMenu filterMenu;

    /**
     * The menu for the levels.
     */
    private transient JMenu levelMenu;

    /**
     * The LogView contained in the Frame.
     */
    protected transient LogView view;

    /**
     * The Configurator for the Frame.
     */
    protected transient Configurator cfg;

    /**
     * The dialog for the options.
     */
    private transient OptionDialog optionDialog;

    /**
     * Whether the view has been initialized.
     */
    private transient boolean initialized;

    /**
     * Whether logging is enabled (true) or disabled (false);
     */
    private transient boolean logging = false;

    //----------------------------------------------------------------------
    // actions
    //----------------------------------------------------------------------
    /**
     * Save the log messages to a file.
     */
    private transient Action actionFileSave;

    /**
     * Show keyboard help.
     */
    private transient Action actionHelpKeys;

    /**
     * Edit the options.
     */
    private transient Action actionEditOptions;

    /**
     * Save all log messages.
     */
    private transient Action actionSaveAll;

    /**
     * Save the visible messages.
     */
    private transient Action actionSaveVisible;

    /**
     * Load some log messages.
     */
    private transient Action actionLoadLog;

    /**
     * Save the configuration.
     */
    private transient Action actionSaveCfg;

    /**
     * Restore the configuration.
     */
    private transient Action actionLoadCfg;

    // }}}

    // {{{ constructors

    /**
     * Create a new LogFrame with a new LogView.
     */
    LogFrame ()
    {
        this (new LogView ());
        view.setLoggingEnabled(logging);
    }

    /**
     * Create a new LogFrame for a LogView.
     *
     * @param   view    The frame's LogView.
     */
    LogFrame (LogView view)
    {
        boolean excluded = ! logging;
        if (excluded) {
            Log.excludeThread();
        }
        try {
            this.view = view;
            cfg = new Configurator ("LogFrame");
        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Initialization
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the frame and its view. This method is not called by the
     * constructor to enable further configuration like setting the resources,
     * before the view is initialized.
     */
    public void init ()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "init()", "");
        }

        // init only once
        if (initialized) {
            return;
        }
        initialized = true;

        boolean excluded = ! logging;
        if (excluded) {
            Log.excludeThread();
        }
        try {
            //  ImageIcon icon = (ImageIcon)
                //  view.getResources().getIcon("logFrame.icon", null);
            //  if (icon != null && icon.getImage() != null) {
                //  setIconImage(icon.getImage());
            //  }
            view.init();
            setupActions();
            initMenu();

            OptionSet options = view.getOptions();
            options.getOption("ShowFilterView").addObserver(this);
            options.getOption("ShowLevelView").addObserver(this);
            filterMenu.setEnabled(options.getBoolean("ShowFilterView", false));
            levelMenu.setEnabled(options.getBoolean("ShowLevelView", false));

            getContentPane().add((JComponent) view);
            pack();
            setLocation(new Point (0, 10));
            setSize(new Dimension(1020, 750));

            cfg.add(ConfigWrapper.makeWrapper(this, "frame"));

            addKeyListener(new java.awt.event.KeyAdapter() {
                public void keyPressed(KeyEvent e) {
                    if (e.getKeyCode() == e.VK_F12 && e.getModifiers() == 0) {
                        System.err.println("" + javax.swing.SwingUtilities
                                           .findFocusOwner(LogFrame.this));
                    }
                }
            });

            FocusFixer.instance().addWindow(this);
        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}
    // {{{ setVisible

    /**
     * Make the frame visible/invisible. Ensures that the frame and view have
     * been initialized.
     *
     * @param   visible True to make the frame visible.
     */
    public void setVisible(boolean visible)
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "setVisible(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "visible: " + visible);
        }
        if (visible) {
            init();
        }
        super.setVisible(visible);
    }


    // }}}
    // {{{ setVisible

    /**
     * Display the frame. Ensures that the frame and view have been
     * initialized.
     */
    public void show()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "show()", "");
        }
        init();
        super.show();
    }


    // }}}

    //----------------------------------------------------------------------
    // Termination
    //----------------------------------------------------------------------
    // {{{ dispose

    /**
     * Dispose the LogFrame.
     */
    public void dispose()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "dispose()", "");
        }
        boolean excluded = ! logging;
        if (excluded) {
            Log.excludeThread();
        }

        try {
            if (cfg.isRegistered()) {
                cfg.unregister(true);
                cfg = null;
            }

            FocusFixer.instance().removeWindow(this);

            OptionSet options = view.getOptions();
            options.getOption("ShowFilterView").deleteObserver(this);
            options.getOption("ShowLevelView").deleteObserver(this);

            SwingUtil.cleanup(this);
            view.cleanup();
            view = null;

            super.dispose();
            removeNotify();


            //  javax.swing.RepaintManager rm =
                //  javax.swing.RepaintManager.currentManager(this);
            //  Dimension d = rm.getDoubleBufferMaximumSize();
            //  rm.setDoubleBufferMaximumSize(new Dimension (0, 0));
            //  rm.setDoubleBufferMaximumSize(d);
        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}
    // {{{ finalize

    /**
     * Finalize the LogFrame.
     */
    public void finalize()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "finalize()", "");
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Various getters and setters
    //----------------------------------------------------------------------
    // {{{ setLoggingEnabled

    /**
     * Enable or disable logging output caused directly or indirectly by
     * LogFrame methods. If logging is disabled, all calls to methods that
     * might create log messages will be protected with {@link
     * de.qfs.lib.log.Log#excludeThread de.qfs.lib.log.Log.excludeThread}.
     * This is necessary, if the LogFrame is used inside the application whose
     * logs it is supposed to filter. <p>
     *
     * The default value is false.
     *
     * @param   enable  True to enable logging, false to disable it.
     */
    public final void setLoggingEnabled(boolean enable)
    {
        logging = enable;
    }

    // }}}
    // {{{ getView

    /**
     * Get the LogView of the LogFrame.
     *
     * @return  The view of the LogFrame.
     */
    public final LogView getView()
    {
        return view;
    }

    // }}}
    // {{{ getConfigurator

    /**
     * Get the configurator of the LogFrame.
     *
     * @return  The configurator of the LogFrame.
     */
    public final Configurator getConfigurator()
    {
        return cfg;
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ setupActions

    /**
     * Set up the actions for the LogFrame.
     */
    private void setupActions()
    {
        actionFileSave = new AbstractAction () {
            public void actionPerformed (ActionEvent e)
            {
            }
        };

        actionHelpKeys = new AbstractAction () {
            public void actionPerformed (ActionEvent e)
            {
                showKeyBindings();
            }
        };

        actionEditOptions = new AbstractAction () {
            public void actionPerformed (ActionEvent e)
            {
                editOptions();
            }
        };

        actionSaveAll = new AbstractAction () {
            public void actionPerformed (ActionEvent e)
            {
                CommandDistributor.getDefaultDistributor().assignCommand
                (new CommandSaveLog
                 (LogFrame.this, LogFrame.this, null, true, true));
            }
        };

        actionSaveVisible = new AbstractAction () {
            public void actionPerformed (ActionEvent e)
            {
                CommandDistributor.getDefaultDistributor().assignCommand
                (new CommandSaveLog
                 (LogFrame.this, LogFrame.this, null, true, false));
            }
        };

        actionLoadLog = new AbstractAction () {
            public void actionPerformed (ActionEvent e)
            {
                CommandDistributor.getDefaultDistributor().assignCommand
                (new CommandSaveLog
                 (LogFrame.this, LogFrame.this, null, false, false));
            }
        };

        actionSaveCfg = new AbstractAction () {
            public void actionPerformed (ActionEvent e)
            {
                CommandDistributor.getDefaultDistributor().assignCommand
                (new CommandSaveConfiguration
                 (LogFrame.this, LogFrame.this, true));
            }
        };

        actionLoadCfg = new AbstractAction () {
            public void actionPerformed (ActionEvent e)
            {
                CommandDistributor.getDefaultDistributor().assignCommand
                (new CommandSaveConfiguration
                 (LogFrame.this, LogFrame.this, false));
            }
        };
    }

    // }}}
    // {{{ initMenu

    /**
     * Initialize the menus of the frame.
     */
    protected void initMenu ()
    {
        boolean excluded = ! logging;
        if (excluded) {
            Log.excludeThread();
        }
        try {
            menuBar = new JMenuBar ();
            setJMenuBar(menuBar);

            // File
            JMenu mFile = makeMenu("file");
            menuBar.add(mFile);

            //  JMenuItem mi = makeMenuItem("fileLoadLog", actionLoadLog);
            //  mFile.add(mi);
            JMenuItem mi = makeMenuItem("fileSaveAll", actionSaveAll);
            mFile.add(mi);
            mi = makeMenuItem("fileSaveVis", actionSaveVisible);
            mFile.add(mi);

            mFile.addSeparator();

            mi = makeMenuItem("fileLoadCfg", actionLoadCfg);
            mFile.add(mi);
            mi = makeMenuItem("fileSaveCfg", actionSaveCfg);
            mFile.add(mi);

            // Edit
            //  JMenu mEdit = makeMenu("edit");
            //  menuBar.add(mEdit);
            //  mi = makeMenuItem("editOptions", actionEditOptions);
            //  mEdit.add(mi);

            // Messages
            JMenu mMessages = makeMenu("messages");
            menuBar.add(mMessages);
            view.getTableView().fillMenu(mMessages);

            // Filter
            filterMenu = makeMenu("filter");
            menuBar.add(filterMenu);
            view.getFilterView().fillMenu(filterMenu);

            // Level
            levelMenu = makeMenu("level");
            menuBar.add(levelMenu);
            if (view.getLevelView() != null) {
                view.getLevelView().fillMenu(levelMenu);
            } else {
                levelMenu.setEnabled(false);
            }

            // View
            menuBar.add(view.makeViewMenu());

            // Help
            JMenu mHelp = makeMenu("help");
            menuBar.add(mHelp);

            mi = makeMenuItem("helpKeys", actionHelpKeys);
            mHelp.add(mi);
        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}
    // {{{ makeMenu

    /**
     * Make a menu.
     *
     * @param   resource        Part of the menu's resource id.
     *
     * @return  The new menu.
     */
    protected JMenu makeMenu (String resource)
    {
        JMenu menu = new JMenu
            (view.getResources().getString("logView.menu." + resource +
                                           ".name",
                                           "MISSING: " + resource));
        String mnemo =
            view.getResources().getString("logView.menu." + resource +
                                          ".mnemonic",
                                          null);
        if (mnemo != null) {
            menu.setMnemonic(mnemo.charAt(0));
        }
        return menu;
    }

    // }}}
    // {{{ makeMenuItem

    /**
     * Make a menu item.
     *
     * @param   resource        Part of the menu item's resource id.
     * @param   action          An action listener for the menu item.
     *
     * @return  The new menu.
     */
    protected JMenuItem makeMenuItem (String resource, ActionListener action)
    {
        MapResourceBundle rb = view.getResources();

        JMenuItem mi = new JMenuItem
            (rb.getString("logView.action." + resource + ".name",
                          "MISSING NAME: " + resource));
        String mnemo = rb.getString("logView.action." + resource +
                                    ".mnemonic", null);
        if (mnemo != null) {
            mi.setMnemonic(mnemo.charAt(0));
        }
        mi.setActionCommand
            (rb.getString("logView.action." + resource + ".command",
                          "MISSING COMMAND: " + resource));
        mi.addActionListener(action);

        return mi;

    }

    // }}}
    // {{{ showKeyBindings

    /**
     * Display the key bingings for a LogView. This is a temporary hack until
     * a true help is available.
     */
    private void showKeyBindings()
    {
        boolean excluded = ! logging;
        if (excluded) {
            Log.excludeThread();
        }
        try {
            Message.showMessage(this, "keyBindings");
        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}
    // {{{ editOptions

    /**
     * Bring up a dialog to edit the options for the LogFrame and LogView.
     */
    private void editOptions()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "editOptions()", "");
        }

        boolean excluded = ! logging;
        if (excluded) {
            Log.excludeThread();
        }
        try {
            if (optionDialog == null) {
                optionDialog = new OptionDialog
                    (this,
                     view.getResources().getString
                     ("logView.optionDialog.title"),
                     view.getOptions(),
                     view.getLevelView() == null ? "logview" : "logviewplus");
            } else {
                optionDialog.update(view.getOptions());
            }
            optionDialog.doModal();
        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Observer interface
    //----------------------------------------------------------------------
    // {{{ update

    /**
     * Notify the LogFrame that one of its Observables has changed.
     *
     * @param   obj     The Observable that changed.
     * @param   arg     Details about the change.
     */
    public void update(Observable obj, Object arg)
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "update(Observable,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "obj: " + obj + ", " +
                       "arg: " + arg);
        }
        if (obj instanceof Option) {
            boolean excluded = ! logging;
            if (excluded) {
                Log.excludeThread();
            }
            try {
                Option option = (Option) obj;
                OptionSet options = view.getOptions();
                if (option.getName().equals("ShowFilterView")) {
                    filterMenu.setEnabled(options.getBoolean
                                          ("ShowFilterView", false));
                } else if (option.getName().equals("ShowLevelView")) {
                    if (options.getBoolean("ShowLevelView", false)
                        && view.getLevelView() != null
                        && levelMenu.getItemCount() == 0) {
                        view.getLevelView().fillMenu(levelMenu);
                    }
                    levelMenu.setEnabled(options.getBoolean
                                         ("ShowLevelView", false));
                }
            } finally {
                if (excluded) {
                    Log.includeThread();
                }
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Main method for logging test purposes
    //----------------------------------------------------------------------
    // {{{ main

    /**
     * Bring up a LogFrame to let the user fool around a bit and test whether
     * logging is completely turned off.
     *
     * @param   args    Ignored.
     */
    public static void main(String[] args)
    {
        Log.setQueueing(false);
        Log.setOutputLevel(10);
        Log.setPreQueueLevel(10);
        Logger.setDefaultLogLevel(10);

        final LogFrame logFrame = new LogFrame ();

        SwingUtil.invokeAndWait(new Runnable () {
            public void run() {
                //  logFrame.getView().startLogging();
                logFrame.init();
                Log.excludeThread();
                try {
                    logFrame.getConfigurator().register();
                    logFrame.getView().getConfigurator().register();
                } finally {
                    Log.includeThread();
                }
                LogLevelTreeModel levels = new LogLevelTreeModel ();
                Logger.addLogLevelListener(levels);
                levels.setLogLevelCallback
                    (new InternalLogLevelCallback (levels));
                logFrame.getView().setLogLevels(levels);

                logFrame.setTitle("Logging selftest");

                logFrame.addWindowListener
                    (new java.awt.event.WindowAdapter () {
                        public void windowClosing
                            (java.awt.event.WindowEvent e) {
                            try {
                                java.io.FileOutputStream fos =
                                    new java.io.FileOutputStream
                                    ("/tmp/logtest.cfg");
                                de.qfs.lib.config.Configuration
                                    .instance().save(fos);
                                fos.close();
                            } catch (Exception ex) {
                                ex.printStackTrace();
                            }
                            System.exit(0);
                        }
                    });

                Log.excludeThread();
                try {
                    de.qfs.lib.option.Option.setResources
                        (logFrame.view.getResources());
                    Message.setResources(logFrame.view.getResources());
                    java.io.FileInputStream fis =
                        new java.io.FileInputStream("/tmp/logtest.cfg");
                    de.qfs.lib.config.Configuration.instance().restore(fis);
                    fis.close();
                } catch (Exception ex) {
                    ex.printStackTrace();
                } finally {
                    Log.includeThread();
                }

                logFrame.show();
            }
        });

        for (int i = 0; i < 100; i++) {
            de.qfs.lib.log.LogEntry entry = new de.qfs.lib.log.LogEntry
                (i % 10 + 1, i, "Thread" + i % 3, "Class" + i % 7,
                 "Method" + i % 11, "message");
            logFrame.view.getTableView().getModel().filter(entry);
        }

        while (true) {
            try {
                Thread.currentThread().sleep(10000);
            } catch (InterruptedException ex) {
            }
        }
    }

    // }}}
}
