// {{{ copyright

/********************************************************************
 *
 * $Id: LogTableFilter.java,v 1.7 2000/07/05 14:07:44 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog.logview;

// {{{ imports

import javax.swing.table.TableModel;

import java.util.Vector;

import de.qfs.lib.gui.TableModelFilter;
import de.qfs.lib.gui.TableModelFilterEvent;
import de.qfs.lib.gui.TableModelFilterListener;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.LogEntry;
import de.qfs.lib.log.Logger;

// }}}

/**
 * This class is modeled after the adapter pattern. It can wrap a
 * TableModelFilter and extend its filtering mechanism with its own.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.7 $
 */
class LogTableFilter
    implements TableModelFilter, TableModelFilterListener
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (LogTableFilter.class);

    /**
     * The nested filter.
     */
    private TableModelFilter filter;

    /**
     * The level to filter on.
     */
    private int filterLevel = -1;

    /**
     * The Thread to filter on.
     */
    private String filterThread = null;

    /**
     * The Class to filter on.
     */
    private String filterClass = null;

    /**
     * The Method to filter on.
     */
    private String filterMethod = null;

    /**
     * The TreeModelFilterListeners.
     */
    private Vector listeners = new Vector ();

    /**
     * Whether logging is enabled (true) or disabled (false);
     */
    private boolean logging;

    // }}}

    // {{{ constructors

    /**
     * Create a new LogTableFilter.
     */
    public LogTableFilter()
    {
    }

    /**
     * Create a new LogTableFilter.
     *
     * @param   filter  The filter to adapt.
     */
    public LogTableFilter(TableModelFilter filter)
    {
        this.filter = filter;
        if (filter != null) {
            filter.addTableModelFilterListener(this);
        }
    }

    // }}}

    // {{{ setLoggingEnabled

    /**
     * Enable or disable logging output caused directly or indirectly by
     * LogTableFilter methods. If logging is disabled, all calls to methods
     * that might create log messages will be protected with {@link
     * de.qfs.lib.log.Log#excludeThread de.qfs.lib.log.Log.excludeThread}.
     * This is necessary, if the LogFilterTreeModel is used inside the
     * application whose logs it is supposed to filter. <p>
     *
     * The default value is false.
     *
     * @param   enable  True to enable logging, false to disable it.
     */
    public final void setLoggingEnabled(boolean enable)
    {
        logging = enable;
    }

    // }}}

    // {{{ setFilter

    /**
     * Set the nested filter.
     *
     * @param   filter  The filter to adapt.
     */
    public void setFilter(TableModelFilter filter)
    {
        if (this.filter != null) {
            this.filter.removeTableModelFilterListener(this);
        }
        this.filter = filter;
        if (filter != null) {
            filter.addTableModelFilterListener(this);
        }
        fireFilterChanged();
    }

    // }}}
    // {{{ setFilterLevel

    /**
     * Set the level to filter on.
     *
     * @param   level   The level to filter on.
     */
    public void setFilterLevel(int level)
    {
        if (filterLevel == -1) {
            filterLevel = level;
        } else {
            filterLevel = -1;
        }
        fireFilterChanged();
    }

    // }}}
    // {{{ setFilterThread

    /**
     * Set the Thread to filter on.
     *
     * @param   thread  The Thread to filter on.
     */
    public void setFilterThread(String thread)
    {
        if (filterThread == null) {
            filterThread = thread;
        } else {
            filterThread = null;
        }
        fireFilterChanged();
    }

    // }}}
    // {{{ setFilterClass

    /**
     * Set the Class to filter on.
     *
     * @param   cl      The Class to filter on.
     */
    public void setFilterClass(String cl)
    {
        if (filterClass == null) {
            filterClass = cl;
        } else {
            filterClass = null;
        }
        fireFilterChanged();
    }

    // }}}
    // {{{ setFilterMethod

    /**
     * Set the Method to filter on.
     *
     * @param   method  The Method to filter on.
     */
    public void setFilterMethod(String method)
    {
        if (filterMethod == null) {
            filterMethod = method;
        } else {
            filterMethod = null;
        }
        fireFilterChanged();
    }

    // }}}
    // {{{ clearFilters

    /**
     * Clear all filters.
     */
    public void clearFilters()
    {
        filterLevel = -1;
        filterThread = null;
        filterClass = null;
        filterMethod = null;
        fireFilterChanged();
    }

    // }}}

    //----------------------------------------------------------------------
    // The TableModelFilter interface
    //----------------------------------------------------------------------
    // {{{ filter

    /**
     * Filter rows of a TableModel.
     *
     * @param   model   The TableModel.
     * @param   row     The row to filter.
     *
     * @return  True if the row should be displayed.
     */
    public boolean filter(TableModel model, int row)
    {
        if (filter != null
            && ! filter.filter(model, row)) {
            return false;
        }
        LogEntry entry = ((LogTableModel)model).getEntry(row);
        if (filterLevel >= 0 && entry.getLevel() != filterLevel) {
            return false;
        }
        if (filterThread != null && !filterThread.equals(entry.getThread())) {
            return false;
        }
        if (filterClass != null && !filterClass.equals(entry.getClazz())) {
            return false;
        }
        if (filterMethod != null && !filterMethod.equals(entry.getMethod())) {
            return false;
        }
        return true;
    }

    // }}}
    // {{{ addTableModelFilterListener

    /**
     * Add a TableModelFilterListener to the LogFilterTreeModel.
     *
     * @param   listener        The TableModelFilterListener to add.
     */
    public void addTableModelFilterListener(TableModelFilterListener listener)
    {
        listeners.addElement(listener);
    }

    // }}}
    // {{{ removeTableModelFilterListener

    /**
     * Remove a TableModelFilterListener from the LogFilterTreeModel.
     *
     * @param   listener        The TableModelFilterListener to remove.
     */
    public void removeTableModelFilterListener
        (TableModelFilterListener listener)
    {
        listeners.removeElement(listener);
    }

    // }}}
    // {{{ fireFilterChanged

    /**
     * Notify all TableModelFilterListeners of a change in the filter.
     */
    private void fireFilterChanged()
    {
        final Vector tmp = (Vector) listeners.clone();
        TableModelFilterEvent event = new TableModelFilterEvent (this);
        boolean excluded = ! logging;
        if (excluded) {
            Log.excludeThread();
        }
        try {
            for (int i = 0; i < tmp.size(); i++) {
                ((TableModelFilterListener) tmp.elementAt(i))
                    .filterChanged(event);
            }
        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The TableModelFilter interface
    //----------------------------------------------------------------------
    // {{{ filterChanged

    /**
     * Called, when the adapted filter has changed. The call is forwarded
     * to the listeners of the LogTableFilter.
     *
     * @param   event   Details bout the change - ignored.
     */
    public void filterChanged(TableModelFilterEvent event)
    {
        fireFilterChanged();
    }

    // }}}
}
