// {{{ copyright

/********************************************************************
 *
 * $Id: LogTableView.java,v 1.21 2000/07/05 14:07:44 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog.logview;

// {{{ imports

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Frame;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Observable;
import java.util.Observer;
import java.util.ResourceBundle;

import javax.swing.Action;
import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.ButtonModel;
import javax.swing.Icon;
import javax.swing.KeyStroke;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.JViewport;
import javax.swing.SwingUtilities;
import javax.swing.border.Border;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;
import java.awt.datatransfer.StringSelection;

import de.qfs.lib.command.CommandDistributor;
import de.qfs.lib.config.Configurator;
import de.qfs.lib.config.ConfigWrapper;
import de.qfs.lib.config.JSplitPaneConfigWrapper;
import de.qfs.lib.gui.DelayedListSelectionListener;
import de.qfs.lib.gui.FilteredAndSortedTableModel;
import de.qfs.lib.gui.Message;
import de.qfs.lib.gui.MultiPane;
import de.qfs.lib.gui.MultiPaneEvent;
import de.qfs.lib.gui.MultiPaneListener;
import de.qfs.lib.gui.SortedTableHelper;
import de.qfs.lib.gui.SortedTableHeaderCellRenderer;
import de.qfs.lib.gui.StatusLine;
import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.gui.TableModelFilter;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.LogEntry;
import de.qfs.lib.log.LogFilter;
import de.qfs.lib.log.Logger;
import de.qfs.lib.option.BooleanOption;
import de.qfs.lib.option.IntegerOption;
import de.qfs.lib.option.Option;
import de.qfs.lib.option.OptionDialog;
import de.qfs.lib.option.OptionSet;
import de.qfs.lib.util.DelayedAction;
import de.qfs.lib.util.MapResourceBundle;

// }}}

/**
 * This is a view that displays log messages in a table. The table can be
 * filtered and sorted in various ways. Also included is a detail view for the
 * currently selected message. A LogTableView is usually used as part of the
 * {@link LogView LogView} component, that integrates the LogTableView with
 * other logging related views. <p>
 *
 * The LogTableView must be initialized by a call to {@link #init init} after
 * creation. It uses its own resources for localization, which can be modified
 * via {@link #setResources setResources}. <p>
 *
 * The {@link Configurable Configurable} members of the LogTableView are
 * collected in a {@link Configurator Configurator}. They include the table
 * layout, the sort order and the state and location of the detail view. The
 * Configurator can be retrieved via {@link #getConfigurator getConfigurator}
 * and must be registered by the caller, unless the LogTableView is part of a
 * LogView.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.21 $
 */
public class LogTableView
    extends JPanel
    implements Observer
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger = new Logger (LogTableView.class);

    /**
     * The names of the icon resources.
     */
    private final static String iconNames[] = {
        "logTableView.icon.err", "logTableView.icon.errDetail",
        "logTableView.icon.wrn", "logTableView.icon.wrnDetail",
        "logTableView.icon.msg", "logTableView.icon.msgDetail",
        "logTableView.icon.mtd", "logTableView.icon.mtdDetail",
        "logTableView.icon.dbg", "logTableView.icon.dbgDetail"};

    /**
     * The icons to render.
     */
    private final static Icon[] icons = new Icon[iconNames.length];

    /**
     * The model of log entries.
     */
    private transient LogTableModel model;

    /**
     * Whether the view has been initialized.
     */
    private transient boolean initialized;

    /**
     * Title label for the view.
     */
    private JLabel titleLabel;

    /**
     * The base name for the title.
     */
    private String titleBase;

    /**
     * The table component.
     */
    private transient JTable table;

    /**
     * The scroll pane for the table.
     */
    private transient JScrollPane tablePane;

    /**
     * The sorter for the table.
     */
    private transient LogSorter sorter;

    /**
     * A filter for the table.
     */
    private transient LogTableFilter filter;

    /**
     * Flags for the filters.
     */
    private transient boolean filterEnabled[] =
        new boolean [LogTableModel.NUM_COLUMNS];

    /**
     * The helper for sorting the table.
     */
    private transient SortedTableHelper tableHelper;

    /**
     * Listener for table selection and changes.
     */
    private transient CellSelectionListener selListener;

    /**
     * Focus listener for initial row selection.
     */
    private transient InitialFocusListener focusListener;

    /**
     * Key listener for fast scrolling.
     */
    private transient TableKeyListener tableKeyListener;

    /**
     * Popup menu mouse listener.
     */
    private transient PopupListener popupListener;

    /**
     * The markers for the table rows.
     */
    private Markers markers;

    /**
     * The view for the entry details.
     */
    private transient LogDetailView detailView;

    /**
     * Multi pane used to toggle the detail view.
     */
    private transient MultiPane multiPane;

    /**
     * Status line to use for search state display. If null, use the
     * table view's title.
     */
    private transient StatusLine statusLine;

    /**
     * The search object for the table.
     */
    private transient Search search;

    /**
     * The Configurator used by the LogTableView.
     */
    private transient Configurator configurator =
        new Configurator ("LogTable");

    /**
     * The CommandDistributor to use for command dispatch (optional).
     */
    private transient CommandDistributor commandDistributor;

    /**
     * The resources to use.
     */
    private transient MapResourceBundle resources;

    /**
     * The options of the LogTableView.
     */
    private OptionSet options = new OptionSet ("Options");

    /**
     * Popup menu for context menu.
     */
    private transient JPopupMenu popup;

    /**
     * Popup sub menu set markers.
     */
    private transient JMenu setMenu;

    /**
     * Popup sub menu goto markers.
     */
    private transient JMenu goMenu;

    /**
     * Dialog to edit the options.
     */
    private OptionDialog optionDialog;

    /**
     * Whether logging is enabled (true) or disabled (false);
     */
    private transient boolean logging;

    //----------------------------------------------------------------------
    // Actions
    //----------------------------------------------------------------------

    /**
     * Copy selected rows to the clipboard.
     */
    private Action copySelectedAction;

    /**
     * Copy visible rows to the clipboard.
     */
    private Action copyVisibleAction;

    /**
     * Copy all rows to the clipboard.
     */
    private Action copyAllAction;

    /**
     * Clear all rows.
     */
    private Action clearAction;

    /**
     * Clear all invisible rows.
     */
    private Action reduceToVisibleAction;

    /**
     * Clear all visible rows.
     */
    private Action removeVisibleAction;

    /**
     * Toggle filter on column.
     */
    private Action toggleColumnFilterAction;

    /**
     * Clear column filters.
     */
    private Action clearColumnFiltersAction;

    /**
     * Set the sort column.
     */
    private Action setSortColumnAction;

    /**
     * Set a marker.
     */
    private Action[] setMarkerActions;

    /**
     * Goto a marker.
     */
    private Action[] gotoMarkerActions;

    /**
     * Edit the options.
     */
    private Action editOptionsAction;

    // }}}

    // {{{ constructor

    /**
     * Create a new LogTableView.
     *
     * @param   model   The mode to use for the tree.
     */
    public LogTableView(LogTableModel model)
    {
        this.model = model;
        resources = new MapResourceBundle ();
        resources.fetchProperties
            ("/de/qfs/apps/qflog/logview/resources/properties/logtableview",
             LogTableView.class);
    }

    // }}}

    //----------------------------------------------------------------------
    // Initialization
    //----------------------------------------------------------------------
    // {{{ init

    /**
     * Initialize the components of the view. This method is not called by the
     * constructor to enable further configuration like setting the resources,
     * before the view is initialized.
     */
    public void init()
    {
        if (initialized) {
            return;
        }
        initialized = true;

        Option option =
            new IntegerOption ("MaxRows", 3000, true, false);
        option.addObserver(this);
        options.add(option);

        option = new BooleanOption ("ShowLines", true);
        option.addObserver(this);
        options.add(option);

        initIcons();

        setLayout(new BorderLayout ());
        setBorder(null);

        JPanel top = new JPanel ();
        top.setLayout(new BorderLayout ());
        top.setBorder(null);

        titleBase = resources.getString("logTableView.title", "Messages");
        titleLabel = new JLabel (titleBase);
        titleLabel.setBorder(BorderFactory.createEtchedBorder());
        titleLabel.setHorizontalTextPosition(titleLabel.LEADING);
        top.add(titleLabel, BorderLayout.NORTH);

        String[] names = {
            resources.getString("logTableView.table.level.title", "Level"),
            resources.getString("logTableView.table.time.title", "Time"),
            resources.getString("logTableView.table.thread.title", "Thread"),
            resources.getString("logTableView.table.class.title", "Class"),
            resources.getString("logTableView.table.method.title", "Method"),
            resources.getString("logTableView.table.message.title", "Message")
            };
        model.setColumnNames(names);

        table = new JTable (model) {
            public boolean isManagingFocus()
            {
                return false;
            }
        };
        JTableHeader header = new JTableHeader (table.getColumnModel()) {
            public boolean isFocusTraversable()
            {
                return false;
            }
        };
        header.setRequestFocusEnabled(false);
        table.setTableHeader(header);
        table.setBorder(null);
        setBackground(table.getBackground());
        table.setColumnSelectionAllowed(false);
        table.setRowMargin(1);
        table.setRowHeight(16);
        sorter = new LogSorter ("Sorter");
        configurator.add(sorter);

        // The renderer for the level icons.
        table.setDefaultRenderer(Integer.class, new LevelRenderer (true));
        table.setDefaultRenderer(Object.class, new LevelRenderer (false));

        if (filter == null) {
            filter = new LogTableFilter ();
            filter.setLoggingEnabled(logging);
        }
        tableHelper = new SortedTableHelper (table, filter, sorter);
        tableHelper.prepareTable();

        // table Columns
        table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        TableColumnModel tcm = table.getColumnModel();
        tcm.getColumn(model.COL_LEVEL).setIdentifier("Level");
        tcm.getColumn(model.COL_LEVEL).setPreferredWidth(24);
        tcm.getColumn(model.COL_TIME).setIdentifier("Time");
        tcm.getColumn(model.COL_TIME).setPreferredWidth(94);
        tcm.getColumn(model.COL_THREAD).setIdentifier("Thread");
        tcm.getColumn(model.COL_THREAD).setPreferredWidth(84);
        tcm.getColumn(model.COL_CLASS).setIdentifier("Class");
        tcm.getColumn(model.COL_CLASS).setPreferredWidth(100);
        tcm.getColumn(model.COL_METHOD).setIdentifier("Method");
        tcm.getColumn(model.COL_METHOD).setPreferredWidth(100);
        tcm.getColumn(model.COL_DETAIL).setIdentifier("Message");
        tcm.getColumn(model.COL_DETAIL).setPreferredWidth(380);

        configurator.add(ConfigWrapper.makeWrapper(table, "Table"));

        // Selection is difficult. We don't know anything about the
        // filter settings and sort order after the configuration is
        // restored, so we cannot select the first row here.
        table.getSelectionModel().setAnchorSelectionIndex(-1);
        focusListener = new InitialFocusListener ();
        table.addFocusListener(focusListener);
        tableHelper.saveSelection(true);
        tableHelper.setHeaderCellRenderer (new HeaderRenderer());
        tablePane = new JScrollPane (table);
        SwingUtil.constrainScroll(tablePane);
        tablePane.setBorder(BorderFactory.createLoweredBevelBorder());
        tablePane.getViewport().putClientProperty("EnableWindowBlit", "true");

        top.add(tablePane, BorderLayout.CENTER);

        // The detail view
        detailView = new LogDetailView ();
        detailView.setLoggingEnabled(logging);
        detailView.setResources(resources);
        detailView.init();

        // Updater for the detail view
        selListener = new CellSelectionListener ();
        table.getSelectionModel().addListSelectionListener(selListener);
        table.getColumnModel().getSelectionModel()
            .addListSelectionListener(selListener);
        table.getModel().addTableModelListener(selListener);
        model.addTableModelListener(selListener);

        // The markers
        markers = new Markers ();
        model.addTableModelListener(markers);

        // The split pane
        multiPane = new MultiPane
            ("MultiPane", MultiPane.VERTICAL,
             top, detailView, 0.8, MultiPane.SHOW_BOTH);
        configurator.add(multiPane);
        add(multiPane, BorderLayout.CENTER);

        // key listeners
        tableKeyListener = new TableKeyListener ();
        table.addKeyListener(tableKeyListener);
        search = new Search ();
        table.addKeyListener(search);
        table.registerKeyboardAction
            (search, "backsearch",
             KeyStroke.getKeyStroke(KeyEvent.VK_BACK_SPACE, 0),
             table.WHEN_FOCUSED);
        table.registerKeyboardAction
            (search, "copysearch",
             KeyStroke.getKeyStroke(KeyEvent.VK_F2, 0),
             table.WHEN_FOCUSED);
        table.registerKeyboardAction
            (search, "searchagain",
             KeyStroke.getKeyStroke(KeyEvent.VK_F3, 0),
             table.WHEN_FOCUSED);
        table.registerKeyboardAction
            (search, "clearsearch",
             KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
             table.WHEN_FOCUSED);
        table.registerKeyboardAction
            (search, "revertsearch",
             KeyStroke.getKeyStroke(KeyEvent.VK_R,  KeyEvent.CTRL_MASK),
             table.WHEN_FOCUSED);

        // setup and bind actions
        setupActions();
        table.registerKeyboardAction
            (toggleColumnFilterAction, "toggleFilter",
             KeyStroke.getKeyStroke(KeyEvent.VK_F,  KeyEvent.CTRL_MASK),
             table.WHEN_FOCUSED);

        table.registerKeyboardAction
            (clearColumnFiltersAction, "clearFilters",
             KeyStroke.getKeyStroke(KeyEvent.VK_K,  KeyEvent.CTRL_MASK),
             table.WHEN_FOCUSED);

        table.registerKeyboardAction
            (setSortColumnAction, "setSortColumn",
             KeyStroke.getKeyStroke(KeyEvent.VK_S,  KeyEvent.CTRL_MASK),
             table.WHEN_FOCUSED);

        // popup menu
        popup = new JPopupMenu ();
        fillMenu(popup);
        popupListener = new PopupListener ();
        table.addMouseListener(popupListener);

        configurator.add(options);
    }

    // }}}
    // {{{ cleanup

    /**
     * Clear everything that might prevent garbage collection.
     */
    public void cleanup()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "cleanup()", "");
        }

        SwingUtil.cleanup(this);

        statusLine = null;

        table.removeFocusListener(focusListener);
        focusListener = null;

        table.getSelectionModel().removeListSelectionListener(selListener);
        table.getColumnModel().getSelectionModel()
            .removeListSelectionListener(selListener);
        table.getModel().removeTableModelListener(selListener);
        model.removeTableModelListener(selListener);
        selListener = null;

        model.removeTableModelListener(markers);
        markers = null;

        table.removeKeyListener(tableKeyListener);
        tableKeyListener = null;
        table.removeKeyListener(search);
        search = null;

        SwingUtil.cleanup(popup);
        popup = null;
        setMenu = null;
        goMenu = null;
        table.removeMouseListener(popupListener);
        popupListener = null;

        detailView.cleanup();
        detailView = null;

        model.cleanup();
        model = null;

        tableHelper = null;
        table.resetKeyboardActions();
        table = null;
        Option[] opts = options.getOptions();
        for (int i = 0; i < opts.length; i++) {
            opts[i].deleteObservers();
        }
        options = null;
        resources = null;
    }

    // }}}
    // {{{ finalize

    /**
     * Finalize the LogTableView.
     */
    public void finalize()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "finalize()", "");
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Various getters and setters
    //----------------------------------------------------------------------
    // {{{ getModel

    /**
     * Get the view's associated LogTableModel.
     *
     * @return  The model of the view.
     */
    public final LogTableModel getModel()
    {
        return model;
    }

    // }}}
    // {{{ getTable

    /**
     * Get the JTable component of the view.
     *
     * @return  The table.
     */
    public final JTable getTable()
    {
        return table;
    }

    // }}}
    // {{{ getMultiPane

    /**
     * Get the MultiPane used for the display of the table and the detail view.
     *
     * @return  The MultiPane used by the LogTableView.
     */
    public final MultiPane getMultiPane()
    {
        return multiPane;
    }

    // }}}
    // {{{ fillMenu

    /**
     * Fill a JMenu with menu items for the LogTableView.
     *
     * @param   menu    The JMenu to fill.
     */
    public void fillMenu(JMenu menu)
    {
        JMenuItem mi = menu.add(copySelectedAction);
        setMnemonic(mi, "copySelected");
        mi.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C,
                                                 KeyEvent.CTRL_MASK));
        mi = menu.add(copyVisibleAction);
        setMnemonic(mi, "copyVisible");
        mi = menu.add(copyAllAction);
        setMnemonic(mi, "copyAll");

        mi = menu.add(clearAction);
        setMnemonic(mi, "clear");
        mi.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_L,
                                                 KeyEvent.CTRL_MASK));
        mi = menu.add(reduceToVisibleAction);
        setMnemonic(mi, "reduceToVisible");
        mi.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_I,
                                                 KeyEvent.CTRL_MASK));
        mi = menu.add(removeVisibleAction);
        setMnemonic(mi, "removeVisible");
        menu.addSeparator();

        mi = menu.add(setSortColumnAction);
        setMnemonic(mi, "setSortColumn");
        mi = menu.add(toggleColumnFilterAction);
        setMnemonic(mi, "toggleColumnFilter");
        mi = menu.add(clearColumnFiltersAction);
        setMnemonic(mi, "clearColumnFilters");

        menu.addSeparator();

        JMenu set = new JMenu (resources.getString
                               ("logTableView.action.setMarker.name",
                                "Set Mark"));
        String mnemo = resources.getString
            ("logTableView.action.setMarker.name", null);
        if (mnemo != null) {
            set.setMnemonic(mnemo.charAt(0));
        }
        for (int i = 0; i <= 9; i++) {
            mi = set.add(setMarkerActions[i]);
            setMnemonic(mi, "setMarker" + i);
        }
        menu.add(set);

        JMenu go = new JMenu (resources.getString
                               ("logTableView.action.gotoMarker.name",
                                "Goto Mark"));
        mnemo = resources.getString
            ("logTableView.action.gotoMarker.name", null);
        if (mnemo != null) {
            go.setMnemonic(mnemo.charAt(0));
        }
        for (int i = 0; i <= 9; i++) {
            mi = go.add(gotoMarkerActions[i]);
            setMnemonic(mi, "gotoMarker" + i);
        }
        menu.add(go);

        menu.addSeparator();

        mi = menu.add(editOptionsAction);
        setMnemonic(mi, "editOptions");
    }

    /**
     * Fill a JPopupMenu with menu items for the LogTableView.
     *
     * @param   menu    The JPopupMenu to fill.
     */
    public void fillMenu(JPopupMenu menu)
    {
        menu.add(copySelectedAction);
        menu.add(copyVisibleAction);
        menu.add(copyAllAction);

        menu.add(clearAction);
        menu.add(reduceToVisibleAction);
        menu.add(removeVisibleAction);

        menu.addSeparator();

        menu.add(setSortColumnAction);
        menu.add(toggleColumnFilterAction);
        menu.add(clearColumnFiltersAction);

        menu.addSeparator();

        setMenu = new JMenu (resources.getString
                               ("logTableView.action.setMarker.name",
                                "Set Mark"));
        for (int i = 0; i <= 9; i++) {
            setMenu.add(setMarkerActions[i]);
        }
        menu.add(setMenu);

        goMenu = new JMenu (resources.getString
                            ("logTableView.action.gotoMarker.name",
                             "Goto Mark"));
        for (int i = 0; i <= 9; i++) {
            goMenu.add(gotoMarkerActions[i]);
        }
        menu.add(goMenu);

        menu.addSeparator();

        menu.add(editOptionsAction);
    }

    // }}}
    // {{{ getConfigurator

    /**
     * Get a {@link de.qfs.lib.config.Configurator Configurator} for the
     * LogTableView.
     *
     * @return  The Configurator for the LogTableView.
     */
    public Configurator getConfigurator()
    {
        return configurator;
    }

    // }}}
    // {{{ getOptions

    /**
     * Get the options of the LogTableView.
     *
     * @return  The options of the LogTableView.
     */
    public final OptionSet getOptions()
    {
        return options;
    }


    // }}}
    // {{{ setLoggingEnabled

    /**
     * Enable or disable logging output caused directly or indirectly by
     * LogTableView methods. If logging is disabled, all calls to methods that
     * might create log messages will be protected with {@link
     * de.qfs.lib.log.Log#excludeThread de.qfs.lib.log.Log.excludeThread}.
     * This is necessary, if the LogFilterTreeModel is used inside the
     * application whose logs it is supposed to filter. <p>
     *
     * The default value is false.
     *
     * @param   enable  True to enable logging, false to disable it.
     */
    public final void setLoggingEnabled(boolean enable)
    {
        logging = enable;
        model.setLoggingEnabled(enable);
        if (detailView != null) {
            detailView.setLoggingEnabled(logging);
        }
        if (filter != null) {
            filter.setLoggingEnabled(logging);
        }
    }

    // }}}
    // {{{ setStatusLine

    /**
     * Set the StatusLine to use for informative display.
     *
     * @param   statusLine
     */
    public void setStatusLine(StatusLine statusLine)
    {
        this.statusLine = statusLine;
        if (search != null) {
            search.setTitle();
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Configuration
    //----------------------------------------------------------------------
    // {{{ setCommandDistributor

    /**
     * Set the CommandDistributor to use for command dispatch.
     *
     * @param   cd      The CommandDistributor to use.
     */
    public void setCommandDistributor(CommandDistributor cd)
    {
        commandDistributor = cd;
    }

    // }}}
    // {{{ getCommandDistributor

    /**
     * Get the CommandDistributor to use for command dispatch.
     *
     * @return  The CommandDistributor to use.
     */
    private CommandDistributor getCommandDistributor()
    {
        return commandDistributor == null
            ? CommandDistributor.getDefaultDistributor()
            : commandDistributor;
    }

    // }}}
    // {{{ setResources

    /**
     * Override the default resources used by the LogTableView. If this method
     * is not called, the LogTableView will use its default resource settings.
     *
     * @param   rb      ResourceBundle containing the resources to use.
     */
    public void setResources(ResourceBundle rb)
    {
        resources.addResources(rb);
    }


    // }}}
    // {{{ setFilter

    /**
     * Set the filter for the table.
     *
     * @param   filter  The new filter.
     */
    public final void setFilter(TableModelFilter filter)
    {
        if (this.filter == null) {
            this.filter = new LogTableFilter (filter);
            this.filter.setLoggingEnabled(logging);
        } else {
            this.filter.setFilter(filter);
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The clipboard
    //----------------------------------------------------------------------
    // {{{ copySelectedToClipboard

    /**
     * Copy the messages in the selected rows of the table to the clipboard.
     */
    public void copySelectedToClipboard()
    {
        int[] rows = table.getSelectedRows();

        if (rows.length > 0) {
            TableModel tm = table.getModel();
            StringBuffer sb = new StringBuffer ();
            for (int i = 0; i < rows.length; i++) {
                sb.append(((LogEntry) tm.getValueAt(rows[i], model.COL_ENTRY))
                          .toString());
                sb.append("\n");
            }
            StringSelection ss = new StringSelection (sb.toString());
            Toolkit.getDefaultToolkit().getSystemClipboard()
                .setContents(ss, ss);
        }
    }

    // }}}
    // {{{ copyVisibleToClipboard

    /**
     * Copy the messages from all visible rows of the table to the clipboard.
     */
    public void copyVisibleToClipboard()
    {
        if (table.getRowCount() > 0) {
            TableModel tm = table.getModel();
            StringBuffer sb = new StringBuffer ();
            for (int i = 0; i < table.getRowCount(); i++) {
                sb.append(((LogEntry) tm.getValueAt(i, model.COL_ENTRY))
                          .toString());
                sb.append("\n");
            }
            StringSelection ss = new StringSelection (sb.toString());
            Toolkit.getDefaultToolkit().getSystemClipboard()
                .setContents(ss, ss);
        }
    }

    // }}}
    // {{{ copyAllToClipboard

    /**
     * Copy the messages from all visible or invisible rows of the table to
     * the clipboard.
     */
    public void copyAllToClipboard()
    {
        LogEntry[] data = model.getData();
        if (data.length > 0) {
            StringBuffer sb = new StringBuffer ();
            for (int i = 0; i < data.length; i++) {
                sb.append(data[i].toString());
                sb.append("\n");
            }
            StringSelection ss = new StringSelection (sb.toString());
            Toolkit.getDefaultToolkit().getSystemClipboard()
                .setContents(ss, ss);
        }
    }



    // }}}

    //----------------------------------------------------------------------
    // Removing rows
    //----------------------------------------------------------------------
    // {{{ clearAll

    /**
     * Remove all rows from the table model.
     */
    public void clearAll()
    {
        selListener.forgetSavedRow();
        model.setData(new LogEntry[0]);
    }

    // }}}
    // {{{ reduceToVisible

    /**
     * Remove all invisible rows from the table model.
     */
    public void reduceToVisible()
    {
        selListener.forgetSavedRow();
        model.reduceToVisible(filter);
    }

    // }}}
    // {{{ removeVisible

    /**
     * Remove all visible rows from the table model.
     */
    public void removeVisible()
    {
        selListener.forgetSavedRow();
        model.removeVisible(filter);
    }

    // }}}

    //----------------------------------------------------------------------
    // The detail view
    //----------------------------------------------------------------------
    // {{{ showDetail

    /**
     * Tell the LogTableView whether the detail view should be displayed.
     *
     * @param   show    True to display the details.
     */
    public void showDetail(boolean show)
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "showDetail(boolean)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "show: " + show);
        }
        multiPane.setShowing(show
                             ? multiPane.SHOW_BOTH
                             : multiPane.SHOW_FIRST);
    }

    // }}}
    // {{{ toggleDetail

    /**
     * Toggle display of the detail view.
     */
    private void toggleDetail()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "showDetail()", "");
        }
        showDetail(multiPane.getShowing() != multiPane.SHOW_BOTH);
    }

    // }}}

    //----------------------------------------------------------------------
    // Observer interface
    //----------------------------------------------------------------------
    // {{{ update

    /**
     * Notify the LogTableView that one of its Observables has changed.
     *
     * @param   obj     The Observable that changed.
     * @param   arg     Details about the change.
     */
    public void update(Observable obj, Object arg)
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "update(Observable,Object)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "obj: " + obj + ", " +
                       "arg: " + arg);
        }
        if (obj instanceof Option) {
            boolean excluded = ! logging;
            if (excluded) {
                Log.excludeThread();
            }
            try {
                Option option = (Option) obj;
                if (option.getName().equals("MaxRows")) {
                    model.setMaxRows(options.getInt("MaxRows", 0));
                } else if (option.getName().equals("ShowLines")) {
                    if (options.getBoolean("ShowLines", true)) {
                        table.getColumnModel().setColumnMargin(1);
                        table.setRowMargin(1);
                        table.setRowHeight(16);
                        table.setShowVerticalLines(true);
                        table.setShowHorizontalLines(true);
                    } else {
                        table.getColumnModel().setColumnMargin(0);
                        table.setRowMargin(0);
                        table.setRowHeight(17);
                        table.setShowVerticalLines(false);
                        table.setShowHorizontalLines(false);
                    }
                }
            } finally {
                if (excluded) {
                    Log.includeThread();
                }
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Helper methods
    //----------------------------------------------------------------------
    // {{{ initIcons

    /**
     * Fetch the icons.
     */
    private void initIcons()
    {
        if (icons[0] == null) {
            for (int i = 0; i < icons.length; i++) {
                icons[i] = resources.getIcon(iconNames[i], null);
            }
        }
    }

    // }}}
    // {{{ setupActions

    /**
     * Set up the actions for the LogTableView.
     */
    private void setupActions()
    {
        // do this only once
        if (copySelectedAction == null) {
            copySelectedAction = new AbstractAction () {
                public void actionPerformed (ActionEvent e) {
                    copySelectedToClipboard();
                }
            };
            setActionProperties(copySelectedAction, "copySelected");

            copyVisibleAction = new AbstractAction () {
                public void actionPerformed (ActionEvent e) {
                    copyVisibleToClipboard();
                }
            };
            setActionProperties(copyVisibleAction, "copyVisible");

            copyAllAction = new AbstractAction () {
                public void actionPerformed (ActionEvent e) {
                    copyAllToClipboard();
                }
            };
            setActionProperties(copyAllAction, "copyAll");

            clearAction = new AbstractAction () {
                public void actionPerformed (ActionEvent e) {
                    clearAll();
                }
            };
            setActionProperties(clearAction, "clear");

            reduceToVisibleAction = new AbstractAction () {
                public void actionPerformed (ActionEvent e) {
                    reduceToVisible();
                }
            };
            setActionProperties(reduceToVisibleAction, "reduceToVisible");

            removeVisibleAction = new AbstractAction () {
                public void actionPerformed (ActionEvent e) {
                    removeVisible();
                }
            };
            setActionProperties(removeVisibleAction, "removeVisible");

            toggleColumnFilterAction = new AbstractAction () {
                public void actionPerformed (ActionEvent e) {
                    int row = table.getSelectionModel()
                        .getAnchorSelectionIndex();
                    int column = table.getColumnModel().getSelectionModel()
                        .getAnchorSelectionIndex();
                    if (row == -1 || column == -1) {
                        return;
                    }
                    int ccol = table.convertColumnIndexToModel(column);
                    filterCell(row, ccol);
                }
            };
            setActionProperties(toggleColumnFilterAction,
                                "toggleColumnFilter");

            clearColumnFiltersAction = new AbstractAction () {
                public void actionPerformed (ActionEvent e) {
                    filter.clearFilters();
                    for (int i = 0; i < filterEnabled.length; i++) {
                        filterEnabled[i] = false;
                        table.getTableHeader().repaint();
                    }
                }
            };
            setActionProperties(clearColumnFiltersAction,
                                "clearColumnFilters");

            setSortColumnAction = new AbstractAction () {
                public void actionPerformed (ActionEvent e) {
                    boolean excluded = ! logging;
                    if (excluded) {
                        Log.excludeThread();
                    }
                    try {
                        int column = table.getColumnModel().getSelectionModel()
                        .getAnchorSelectionIndex();
                        if (column == -1) {
                            return;
                        }
                        int ccol = table.convertColumnIndexToModel(column);
                        sorter.setSortColumn(ccol);
                    } finally {
                        if (excluded) {
                            Log.includeThread();
                        }
                    }
                }
            };
            setActionProperties(setSortColumnAction,
                                "setSortColumn");

            setMarkerActions = new Action[10];
            gotoMarkerActions = new Action[10];
            for (int i = 0; i <= 9; i++) {
                setMarkerActions[i] = new SetMarkerAction ("" + i);
                setActionProperties(setMarkerActions[i], "setMarker" + i);
                gotoMarkerActions[i] = new GotoMarkerAction ("" + i);
                setActionProperties(gotoMarkerActions[i], "gotoMarker" + i);
            }

            editOptionsAction = new AbstractAction () {
                public void actionPerformed (ActionEvent e) {
                    editOptions();
                }
            };
            setActionProperties(editOptionsAction,
                                "editOptions");

        }
    }

    // }}}
    // {{{ setActionProperties

    /**
     * Set name and icon for an action from the LogTableView's resources.
     *
     * @param   action   The action to prepare.
     * @param   resource Identifier for the action's resources.
     */
    private void setActionProperties(Action action, String resource)
    {
        String name = resources.getString
            ("logTableView.action." + resource + ".name", null);
        if (name != null) {
            action.putValue(Action.NAME, name);
        }

        String iconName = resources.getString
            ("logTableView.action." + resource + ".icon", null);
        if (iconName != null) {
            Icon icon =  resources.getIcon (iconName, null);
            if (icon != null) {
                action.putValue(Action.SMALL_ICON, icon);
            }
        }
    }

    // }}}
    // {{{ setMnemonic

    /**
     * Set the mnemonic shortcut for a menu item.
     *
     * @param   mi              The menu item.
     * @param   resource        The identifier to look up the mnemonic
     *                          resource.
     */
    private void setMnemonic(JMenuItem mi, String resource)
    {
        String mnemo = resources.getString
            ("logTableView.action." + resource + ".mnemonic", null);
        if (mnemo != null) {
            mi.setMnemonic(mnemo.charAt(0));
        }
    }

    // }}}
    // {{{ editOptions

    /**
     * Bring up a dialog to edit the options for the LogTableView.
     */
    private void editOptions()
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "editOptions()", "");
        }

        boolean excluded = ! logging;
        if (excluded) {
            Log.excludeThread();
        }
        try {
            Frame frame;
            Window win = SwingUtilities.windowForComponent(this);
            if (win instanceof Frame) {
                frame = (Frame) win;
            } else {
                // create a dummy parent
                frame = new Frame();
            }
            if (optionDialog == null) {
                optionDialog = new OptionDialog
                    (frame, resources.getString
                         ("logTableView.optionDialog.title", "MISSING"),
                     options, "tableview");
            } else {
                optionDialog.update(options);
            }
            optionDialog.setLocationRelativeTo(this);
            optionDialog.doModal();
        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}

    // {{{ restoreRow

    /**
     * Restore the last row known to be selected. If the row is invisible find
     * a row as close as possible.
     *
     * @param   row     The row to search for.
     * @param   column  The column to select.
     */
    private void restoreRow(int row, int column)
    {
        if (logging && logger.level >= Log.MTD) {
            logger.log(Log.MTD, "restoreRow(int,int)",
                       logger.level < Log.MTDDETAIL ? "" :
                       "row: " + row + ", " +
                       "column: " + column);
        }
        if (row >= 0) {
            int i;
            FilteredAndSortedTableModel filteredTableModel =
                (FilteredAndSortedTableModel) table.getModel();
            // first try: exact match
            for (i = 0; i < filteredTableModel.getRowCount(); i++) {
                if (row == filteredTableModel.getMappedRow(i)) {
                    break;
                }
            }
            // second try: nearest match
            if (i == filteredTableModel.getRowCount()) {
                for (i = 0; i < filteredTableModel.getRowCount(); i++) {
                    if (sorter.compare
                        (model, row,
                         filteredTableModel.getMappedRow(i)) <= 0) {
                        break;
                    }
                }
            }

            if (i == filteredTableModel.getRowCount()) {
                i--;
            }

            jumpToRow(i, false);

            if (column >= 0) {
                table.addColumnSelectionInterval(column, column);
            }
        }
    }

    // }}}
    // {{{ jumpToRow

    /**
     * Select a row and center it on the table.
     *
     * @param   row     The row to select.
     * @param   always  If true, always center the row, if false, recenter only
     *                  if the row is not currently visible.
     */
    private void jumpToRow(int row, boolean always)
    {
            // scroll so selected row is centered vertically
            Rectangle rect = table.getCellRect(row, 0, false);
            JViewport vp = (JViewport) table.getParent();
            Dimension vs = vp.getSize();
            Point pos = vp.getViewPosition();
            if (!always) {
                if (pos.y <= rect.y
                    && pos.y + vs.height >= rect.y + rect.height) {
                    // already visible
                    table.addRowSelectionInterval(row, row);
                    return;
                }
            }
            Dimension size = table.getSize();
            pos.y = rect.y - (vs.height - rect.height) / 2;
            if (pos.y + vs.height > size.height) {
                pos.y = size.height - vs.height;
            }
            if (pos.y < 0) {
                pos.y = 0;
            }
            ((JViewport) table.getParent()).setViewPosition(pos);
            table.getParent().repaint();

            table.addRowSelectionInterval(row, row);
            table.getSelectionModel().setAnchorSelectionIndex(row);
    }

    // }}}
    // {{{ filterCell

    /**
     * Toggle a filter for a column.
     *
     * @param   row     The selected row that determines the filter value.
     * @param   column  The column whose filter to toggle.
     */
    private void filterCell(int row, int column)
    {
        LogEntry entry = (LogEntry) table.getValueAt(row, model.COL_ENTRY);
        switch (column) {
        case model.COL_LEVEL:
            filter.setFilterLevel(entry.getLevel());
            filterEnabled[model.COL_LEVEL] = ! filterEnabled[model.COL_LEVEL];
            table.getTableHeader().repaint();
            break;
        case model.COL_THREAD:
            filter.setFilterThread(entry.getThread());
            filterEnabled[model.COL_THREAD] =
                ! filterEnabled[model.COL_THREAD];
            table.getTableHeader().repaint();
            break;
        case model.COL_CLASS:
            filter.setFilterClass(entry.getClazz());
            filterEnabled[model.COL_CLASS] = ! filterEnabled[model.COL_CLASS];
            table.getTableHeader().repaint();
            break;
        case model.COL_METHOD:
            filter.setFilterMethod(entry.getMethod());
            filterEnabled[model.COL_METHOD] =
                ! filterEnabled[model.COL_METHOD];
            table.getTableHeader().repaint();
            break;
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Markers
    //----------------------------------------------------------------------
    // {{{ class Markers

    // {{{ static Markers variables

    /**
     * The Logger used for logging Markers methods.
     */
    private final static Logger mLogger = new Logger (Markers.class);

    // }}}

    /**
     * Markers for the table. Markers can be added, removed, cleared and
     * jumped to. They are updated according to changes in the table model.
     */
    private class Markers
        implements TableModelListener
    {
        // {{{ variables

        /**
         * Hashtable mapping from identifier to row.
         */
        private Hashtable markers = new Hashtable ();

        // }}}

        // {{{ constructor

        /**
         * Create a new Markers object.
         */
        public Markers ()
        {
        }

        // }}}

        // {{{ addMarker

        /**
         * Add a marker.
         *
         * @param       id      Identifier for the marker.
         * @param       row     The row to mark.
         */
        public void addMarker(String id, int row)
        {
            if (logging && mLogger.level >= Log.MTD) {
                mLogger.log(Log.MTD, "addMarker(String,int)",
                           mLogger.level < Log.MTDDETAIL ? "" :
                           "id: " + id + ", " +
                           "row: " + row);
            }
            markers.put(id, new Integer(row));
            if (statusLine != null) {
                statusLine.setTemporaryMessage
                    (Message.format("logTableView.setMark",
                                    new String[] {id}));
            }
        }


        // }}}
        // {{{ removeMarker

        /**
         * Remove a marker.
         *
         * @param       id      Identifier of the marker.
         */
        public void removeMarker(String id)
        {
            if (logging && mLogger.level >= Log.MTD) {
                mLogger.log(Log.MTD, "removeMarker(String)",
                           mLogger.level < Log.MTDDETAIL ? "" :
                           "id: " + id);
            }
            markers.remove(id);
        }

        // }}}
        // {{{ clearMarkers

        /**
         * Remove all markers.
         */
        public void clearMarkers()
        {
            if (logging && mLogger.level >= Log.MTD) {
                mLogger.log(Log.MTD, "clearMarkers()", "");
            }
            markers.clear();
        }

        // }}}
        // {{{ jumpToMarker

        /**
         * Jump to the row pointed to by a marker.
         *
         * @param       id      Identifier for the marker.
         */
        public void jumpToMarker(String id)
        {
            if (logging && mLogger.level >= Log.MTD) {
                mLogger.log(Log.MTD, "jumpToMarker(String)",
                           mLogger.level < Log.MTDDETAIL ? "" :
                           "id: " + id);
            }
            Integer row = (Integer) markers.get(id);
            if (row != null) {
                if (statusLine != null) {
                    statusLine.setTemporaryMessage
                        (Message.format("logTableView.jumpToMark",
                                        new String[] {id}));
                }
                table.clearSelection();
                restoreRow(row.intValue(), -1);
            }
        }

        // }}}

        // {{{ tableChanged

        /**
         * Notification of a change in the table model's contents.
         *
         * @param       event   The event containing the details.
         */
        public void tableChanged(TableModelEvent event)
        {
            int start = event.getFirstRow();
            int end = event.getLastRow();
            if (logging && mLogger.level >= Log.MTD) {
                mLogger.log(Log.MTD, "tableChanged(TableModelEvent)",
                           mLogger.level < Log.MTDDETAIL ? "" :
                           ", start: " + start +
                           ", end: " + end);
            }
            if (markers.size() >= 0) {
                int type = event.getType();
                int diff = end - start + 1;
                if (start == event.HEADER_ROW
                    || (start == 0 && end >= model.getRowCount()
                        && type == event.UPDATE)) {
                    // complete change
                    if (logging && mLogger.level >= Log.DBG) {
                        mLogger.log(Log.DBG,
                                   "tableChanged(TableModelEvent)",
                                   "markers cleared");
                    }
                    clearMarkers();
                } else if (type == event.INSERT) {
                    for (Enumeration e = markers.keys();
                         e.hasMoreElements(); ) {
                        String id = (String) e.nextElement();
                        Integer row = (Integer) markers.get(id);
                        if (start <= row.intValue()) {
                            markers.put(id,
                                        new Integer (row.intValue() + diff));
                            if (logging && mLogger.level >= Log.DBG) {
                                mLogger.log(Log.DBG,
                                           "tableChanged(TableModelEvent)",
                                           "marker " + id + " moved up");
                            }
                        }
                    }
                } else if (type == event.DELETE) {
                    for (Enumeration e = markers.keys();
                         e.hasMoreElements(); ) {
                        String id = (String) e.nextElement();
                        Integer row = (Integer) markers.get(id);
                        if (start <= row.intValue()) {
                            if (end >= row.intValue()) {
                                markers.remove(id);
                                if (logging && mLogger.level >= Log.DBG) {
                                    mLogger.log(Log.DBG,
                                               "tableChanged(TableModelEvent)",
                                               "marker " + id + " removed");
                                }
                            } else {
                                markers.put(id, new Integer
                                            (row.intValue() - diff));
                                if (logging && mLogger.level >= Log.DBG) {
                                    mLogger.log(Log.DBG,
                                               "tableChanged(TableModelEvent)",
                                               "marker " + id + " moved down");
                                }
                            }
                        }
                    }
                }
            }
        }

        // }}}
    }

    // }}}
    // {{{ class SetMarkerAction

    private class SetMarkerAction
        extends AbstractAction
    {
        /**
         * The marker identifier for this action.
         */
        private String marker;

        // {{{ constructor

        public SetMarkerAction (String marker)
        {
            this.marker = marker;
        }

        // }}}

        // {{{ actionPerformed

        /**
         * Perform the action.
         *
         * @param       e       Ignored.
         */
        public void actionPerformed(ActionEvent e)
        {
            int row = table.getSelectionModel().getAnchorSelectionIndex();
            if (row >= 0) {
                row = ((FilteredAndSortedTableModel) table.getModel())
                    .getMappedRow(row);
                markers.addMarker(marker, row);
            }
        }

        // }}}
    }

    // }}}
    // {{{ class GotoMarkerAction

    private class GotoMarkerAction
        extends AbstractAction
    {
        /**
         * The marker identifier for this action.
         */
        private String marker;

        // {{{ constructor

        public GotoMarkerAction (String marker)
        {
            this.marker = marker;
        }

        // }}}

        // {{{ actionPerformed

        /**
         * Perform the action.
         *
         * @param       e       Ignored.
         */
        public void actionPerformed(ActionEvent e)
        {
            int row = table.getSelectionModel().getAnchorSelectionIndex();
            markers.jumpToMarker(marker);
        }

        // }}}
    }

    // }}}

    //----------------------------------------------------------------------
    // Key listeners
    //----------------------------------------------------------------------
    // {{{ class TableKeyListener

    // {{{ static variables for TableKeyListener

    /**
     * Modifier mask to filter unwanted detail.
     */
    private final static int KEY_MASK =
        KeyEvent.SHIFT_MASK | KeyEvent.CTRL_MASK | KeyEvent.ALT_MASK;

    // }}}
    /**
     * A listener for key presses on the table. Used for fast scrolling via
     * Alt-up/Alt-down.
     */
    protected class TableKeyListener
        extends KeyAdapter
    {
        // {{{ variables

        /**
         * The selected row during fast scrolling, restored after the key
         * is released.
         */
        private int row = -1;

        /**
         * Delayed action for restoring selection after fast scroll.
         */
        DelayedAction delayed = new DelayedAction (300);

        /**
         * Action for delayed selection restore.
         */
        private Runnable restoreSel = new Runnable () {
            public void run()
            {
                if (row >= 0) {
                    table.getSelectionModel().setSelectionInterval(row, row);
                    row = -1;
                }
            }
        };

        // }}}

        // {{{ keyPressed

        /**
         * Fast scrolling for Alt-Up, Alt-Down
         *
         * @param       e       KeyPressed event.
         */
        public void keyPressed(KeyEvent e) {
            if (e.getKeyCode() == e.VK_UP
                && (e.getModifiers() & KEY_MASK) == e.ALT_MASK) {
                e.consume();
                fastUp();
            } else if (e.getKeyCode() == e.VK_DOWN
                && (e.getModifiers() & KEY_MASK) == e.ALT_MASK) {
                e.consume();
                fastDown();
            } else if (e.getKeyCode() >= e.VK_0 && e.getKeyCode() <= e.VK_9
                       &&  (e.getModifiers() & KEY_MASK) == e.ALT_MASK) {
                // set mark
                e.consume();
                int row = table.getSelectionModel().getAnchorSelectionIndex();
                if (row >= 0) {
                    row = ((FilteredAndSortedTableModel) table.getModel())
                        .getMappedRow(row);
                    char c = (char) ('0' + e.getKeyCode() - e.VK_0);
                    markers.addMarker("" + c, row);
                }
            } else if (e.getKeyCode() >= e.VK_0 && e.getKeyCode() <= e.VK_9
                       &&  (e.getModifiers() & KEY_MASK) == e.CTRL_MASK) {
                e.consume();
                char c = (char) ('0' + e.getKeyCode() - e.VK_0);
                markers.jumpToMarker("" + c);
            }
        }

        // }}}

        // {{{ fastUp

        private void fastUp()
        {
            if (row < 0) {
                row = table.getSelectionModel().getAnchorSelectionIndex();
            }
            if (row == 0) {
                return;
            }
            row--;
            JViewport vp = tablePane.getViewport();
            Point pos = vp.getViewPosition();
            if (pos.y > 0) {
                pos.y -= table.getRowHeight() + table.getRowMargin();
                if (pos.y < 0) {
                    pos.y = 0;
                }
                vp.setViewPosition(pos);
                delayed.perform(restoreSel);
            } else {
                table.getSelectionModel()
                    .setSelectionInterval(row, row);
                row = -1;
            }
        }

        // }}}
        // {{{ fastDown

        private void fastDown()
        {
            if (row < 0) {
                row = table.getSelectionModel().getAnchorSelectionIndex();
            }
            if (row == table.getRowCount() - 1) {
                return;
            }
            row++;
            JViewport vp = tablePane.getViewport();
            Point pos = vp.getViewPosition();
            Dimension vs = vp.getViewSize();
            if (pos.y + vp.getSize().height < vs.height) {
                pos.y += table.getRowHeight() + table.getRowMargin();
                if (pos.y + vp.getSize().height > vs.height) {
                    pos.y = vs.height - vp.getSize().height;
                }
                vp.setViewPosition(pos);
                delayed.perform(restoreSel);
            } else {
                table.getSelectionModel()
                    .setSelectionInterval(row, row);
                row = -1;
            }
        }

        // }}}
    }

    // }}}
    // {{{ class Search

    /**
     * This class implements a simple search mechanism on the table.
     */
    private class Search
        extends KeyAdapter
        implements ActionListener
    {
        // {{{ variables

        /**
         * Search timestamp.
         */
        private long lastTimestamp = 0;

        /**
         * Search string.
         */
        private StringBuffer search = new StringBuffer ();

        /**
         * Search level.
         */
        private int lastLevel = -1;

        /**
         * Forward or reverse search.
         */
        private boolean reverse = false;

        /**
         * Whether last search wrapped around.
         */
        private boolean wrapped = false;

        /**
         * Whether last search was successful.
         */
        private boolean found = false;

        /**
         * Title string for search.
         */
        private String searchTitle =
            resources.getString("logTableView.search.title", "search");

        /**
         * Title string for reverse search.
         */
        private String rsearchTitle =
            resources.getString("logTableView.rsearch.title", "rsearch");

        /**
         * Title string for wrapped search.
         */
        private String wrappedTitle =
            resources.getString("logTableView.wrapped.title", "wrapped");

        /**
         * Title string for unsuccessful search.
         */
        private String notFoundTitle =
            resources.getString("logTableView.notFound.title", "not found");

        // }}}

        // {{{ keyPressed

        /**
         * Search in current column.
         *
         * @param       e       KeyPressed event.
         */
        public void keyPressed(KeyEvent e)
        {
            char c = e.getKeyChar();
            if ((e.getModifiers() & (e.ALT_MASK | e.CTRL_MASK)) == 0
                && c != e.CHAR_UNDEFINED
                && ! Character.isISOControl(c)) {
                e.consume();

                c = Character.toLowerCase(c);

                int row = table.getSelectionModel().getAnchorSelectionIndex();
                int column = table.getColumnModel().getSelectionModel()
                    .getAnchorSelectionIndex();
                if (row == -1 || column == -1) {
                    return;
                }

                int ccol = table.convertColumnIndexToModel(column);
                wrapped = false;
                if (ccol == model.COL_LEVEL) {
                    if (Character.isDigit(c)) {
                        lastLevel = c == '0' ? 10 : c - '0';
                        searchRow(row, lastLevel, reverse);
                    }
                    search.setLength(0);
                } else {
                    lastLevel = -1;
                    search.append(c);
                    // need unconverted column for getValueAt
                    searchRow(row, column, search.toString(), true, reverse);
                }
                lastTimestamp = e.getWhen();
                setTitle();
            }
        }

        // }}}
        // {{{ actionPerformed

        /**
         * Bindings for various KeyStrokes.
         *
         * @param       event   The ActionEvent.
         */
        public void actionPerformed (ActionEvent event)
        {
            int pos = event.paramString().lastIndexOf ("=");
            if (pos == -1) {
                return;
            }

            int row = table.getSelectionModel().getAnchorSelectionIndex();
            int column = table.getColumnModel().getSelectionModel()
                .getAnchorSelectionIndex();
            if (row == -1 || column == -1) {
                return;
            }
            int ccol = table.convertColumnIndexToModel(column);

            boolean excluded = ! logging;
            if (excluded) {
                Log.excludeThread();
            }
            try {
                String cmd = event.paramString().substring (pos + 1);
                if (cmd.equals("backsearch")) {
                    wrapped = false;
                    found = true;
                    if (ccol != model.COL_LEVEL && search.length() > 0) {
                        search.setLength(search.length() - 1);
                    }
                    setTitle();
                } else if (cmd.equals("copysearch")) {
                    if (ccol != model.COL_LEVEL) {
                        wrapped = false;
                        found = true;
                        search = new StringBuffer
                            (((String) table.getValueAt(row, column))
                                .toLowerCase());
                        setTitle();
                    }
                } else if (cmd.equals("searchagain")) {
                    if (ccol == model.COL_LEVEL) {
                        if (lastLevel >= 0) {
                            wrapped = false;
                            searchRow(row, lastLevel, reverse);
                            setTitle();
                        }
                    } else if (search.length() > 0) {
                        wrapped = false;
                        searchRow(row, column, search.toString(), false,
                                  reverse);
                        setTitle();
                    }
                } else if (cmd.equals("clearsearch")) {
                    wrapped = false;
                    search.setLength(0);
                    lastLevel = -1;
                    setTitle();
                } else if (cmd.equals("revertsearch")) {
                    wrapped = false;
                    found = true;
                    reverse = !reverse;
                    setTitle();
                }
            } finally {
                if (excluded) {
                    Log.includeThread();
                }
            }
        }

        // }}}

        // {{{ searchRow

        /**
         * Search for a row in the table matching certain criteria.
         *
         * @param       row     The row to start from.
         * @param       column  The column to search in.
         * @param       find    The string to find.
         * @param       expand  Whether the string was just expanded, in which
         *                      case tha start row is a valid target.
         * @param       reverse Whether the search direction is reversed.
         */
        private void searchRow(int row, int column, String find,
                               boolean expand, boolean reverse)
        {
            found = false;
            int len = find.length();
            if (reverse) {
                for (int i = row - (expand ? 0 : 1); i >= 0; i--) {
                    String val = (String) table.getValueAt(i, column);
                    if (val.length() >= len &&
                        val.toLowerCase().indexOf(find) >= 0) {
                        jumpToRow(i, false);
                        table.getSelectionModel().setSelectionInterval(i, i);
                        found = true;
                        break;
                    }
                }
                for (int i = table.getRowCount() - 1; !found && i > row; i--) {
                    String val = (String) table.getValueAt(i, column);
                    if (val.length() >= len &&
                        val.toLowerCase().indexOf(find) >= 0) {
                        jumpToRow(i, false);
                        table.getSelectionModel().setSelectionInterval(i, i);
                        found = true;
                        wrapped = true;
                        break;
                    }
                }
            } else {
                for (int i = row + (expand ? 0 : 1);
                     i < table.getRowCount(); i++) {
                    String val = (String) table.getValueAt(i, column);
                    if (val.length() >= len &&
                        val.toLowerCase().indexOf(find) >= 0) {
                        jumpToRow(i, false);
                        table.getSelectionModel().setSelectionInterval(i, i);
                        found = true;
                        break;
                    }
                }
                for (int i = 0; !found && i < row; i++) {
                    String val = (String) table.getValueAt(i, column);
                    if (val.length() >= len &&
                        val.toLowerCase().indexOf(find) >= 0) {
                        jumpToRow(i, false);
                        table.getSelectionModel().setSelectionInterval(i, i);
                        found = true;
                        wrapped = true;
                        break;
                    }
                }
            }
        }

        /**
         * Search for the next row in the table with a given level.
         *
         * @param       row     The start row.
         * @param       level   The level to look for.
         * @param       reverse Whether the search direction is reversed.
         */
        private void searchRow(int row, int level, boolean reverse)
        {
            found = false;
            if (reverse) {
                for (int i = row - 1; i >= 0; i--) {
                    if (((LogEntry) table.getValueAt(i, model.COL_ENTRY))
                        .getLevel() == level) {
                        jumpToRow(i, false);
                        table.getSelectionModel().setSelectionInterval(i, i);
                        found = true;
                        break;
                    }
                }
                for (int i = table.getRowCount() - 1; !found && i > row; i--) {
                    if (((LogEntry) table.getValueAt(i, model.COL_ENTRY))
                        .getLevel() == level) {
                        jumpToRow(i, false);
                        table.getSelectionModel().setSelectionInterval(i, i);
                        found = true;
                        wrapped = true;
                        break;
                    }
                }
            } else {
                for (int i = row + 1; i < table.getRowCount(); i++) {
                    if (((LogEntry) table.getValueAt(i, model.COL_ENTRY))
                        .getLevel() == level) {
                        jumpToRow(i, false);
                        table.getSelectionModel().setSelectionInterval(i, i);
                        found = true;
                        break;
                    }
                }
                for (int i = 0; !found && i < row; i++) {
                    if (((LogEntry) table.getValueAt(i, model.COL_ENTRY))
                        .getLevel() == level) {
                        jumpToRow(i, false);
                        table.getSelectionModel().setSelectionInterval(i, i);
                        found = true;
                        wrapped = true;
                        break;
                    }
                }
            }
        }

        // }}}
        // {{{ setTitle

        /**
         * Set the title of the table part of the LogTableView. Depends on
         * the current search string, direction and result.
         */
        private void setTitle()
        {
            StringBuffer sb = new StringBuffer ();
            StatusLine statusLine = null;
            if (statusLine == null) {
                sb.append(titleBase);
            }
            if (lastLevel >= 0
                || search.length() > 0) {
                sb.append("   ");
                sb.append(reverse ? rsearchTitle : searchTitle);
                if (wrapped) {
                    sb.append(" (");
                    sb.append(wrappedTitle);
                    sb.append(")");
                } else if (! found) {
                    sb.append(" (");
                    sb.append(notFoundTitle);
                    sb.append(")");
                }
                sb.append(": ");
                if (lastLevel > 0) {
                    if (statusLine == null) {
                        titleLabel.setIcon(icons[lastLevel - 1]);
                    } else {
                        statusLine.getMessageLabel()
                            .setIcon(icons[lastLevel - 1]);
                        statusLine.getMessageLabel()
                            .setHorizontalTextPosition(JLabel.LEADING);
                    }
                } else {
                    if (statusLine == null) {
                        titleLabel.setIcon(null);
                    } else {
                        statusLine.getMessageLabel().setIcon(null);
                    }
                    sb.append(search);
                }
            } else if (statusLine == null) {
                titleLabel.setIcon(null);
            }
            if (statusLine == null) {
                titleLabel.setText(sb.toString());
            } else {
                titleLabel.setText(titleBase);
                titleLabel.setIcon(null);
                if (sb.length() == 0) {
                    statusLine.popMessage();
                } else {
                    statusLine.pushMessage(sb.toString());
                }
            }
        }

        // }}}
    }

    // }}}

    //----------------------------------------------------------------------
    // Popup menu listener
    //----------------------------------------------------------------------
    // {{{ class PopupListener

    /**
     * Listener for mouseclicks that brings up a popup menu.
     */
    class PopupListener
        extends MouseAdapter
    {
        /**
         * Notify the listener that a mouse button has been pressed.
         *
         * @param       event   Details aobut the event.
         */
        public void mousePressed(MouseEvent event)
        {
            maybePopup(event);
        }

        /**
         * Notify the listener that a mouse button has been released.
         *
         * @param       event   Details aobut the event.
         */
        public void mouseReleased(MouseEvent event)
        {
            maybePopup(event);
        }

        /**
         * Check if the event is a popup trigger and bring up a popup menu if
         * required. If the click happens on a table row that is selected, the
         * selection doesn't change. If the row is not selected, it will be
         * made the only selected row.
         *
         * @param       event   The MouseEvent containing the details.
         */
        private void maybePopup(MouseEvent event)
        {
            if (event.isPopupTrigger()) {
                int row = table.rowAtPoint
                    (new Point (event.getX(), event.getY()));
                int column = table.columnAtPoint
                    (new Point (event.getX(), event.getY()));
                if (row >= 0) {
                    if (! table.isRowSelected(row)) {
                        table.setRowSelectionInterval(row, row);
                    }
                }
                if (column >= 0) {
                    table.setColumnSelectionInterval(column, column);
                }
                SwingUtil.showPopup(popup, table, event.getX(), event.getY(),
                                    false);
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // Selection and model listener for the table
    //----------------------------------------------------------------------
    // {{{ class CellSelectionListener

    // {{{ static variables for CellSelectionListener

    /**
     * Default dealy for selection and table actions.
     */
    private final static int DELAY = 200;

    /**
     * The Logger used for logging CellSelectionListener methods.
     */
    private final static Logger cslLogger =
        new Logger (CellSelectionListener.class);

    // }}}

    /**
     * Listener for changes in the table's selection or model contents that
     * updates the detail view accordingly.
     */
    protected class CellSelectionListener
        extends DelayedListSelectionListener
        implements TableModelListener, Runnable
    {
        // {{{ variables

        /**
         * Delayed action for table updates.
         */
        private DelayedAction delay;

        /**
         * The last known row to be selected (in the true model).
         */
        private int lastRow = -1;

        /**
         * The last known column to be selected.
         */
        private int lastColumn;

        // }}}

        // {{{ constructor

        /**
         * Create a new CellSelectionListener.
         */
        public CellSelectionListener()
        {
            super(DELAY);
            delay = new DelayedAction (DELAY);
        }

        // }}}

        // {{{ forgetSavedRow

        /**
         * Must be called when rows are deleted from the table model.
         */
        public void forgetSavedRow()
        {
            lastRow = -1;
        }

        // }}}

        // {{{ delayedValueChanged

        /**
         * Delayed notification of a change in the table's selection.
         *
         * @param       e       The event containing the details.
         */
        public void delayedValueChanged(ListSelectionEvent e)
        {
            if (logging && cslLogger.level >= Log.MTD) {
                cslLogger.log(Log.MTD, "delayedValueChanged(ListSelectionEvent)",
                           cslLogger.level < Log.MTDDETAIL ? "" :
                           "e: " + e);
            }
            if (e.getValueIsAdjusting()) {
                return;
            }
            int row = table.getSelectionModel().getAnchorSelectionIndex();
            lastColumn = table.getColumnModel().getSelectionModel()
                .getAnchorSelectionIndex();
            if (row >= 0 && row < table.getRowCount()) {
                detailView.setEntry((LogEntry) table.getModel().getValueAt
                                    (row, model.COL_ENTRY));
                lastRow = ((FilteredAndSortedTableModel) table.getModel())
                    .getMappedRow(row);
            } else {
                lastRow = -1;
            }
        }

        // }}}
        // {{{ tableChanged

        /**
         * Notification of a change in the table model's contents.
         *
         * @param       e       The event containing the details.
         */
        public void tableChanged(TableModelEvent e)
        {
            int start = e.getFirstRow();
            int end = e.getLastRow();
            if (logging && cslLogger.level >= Log.MTD) {
                cslLogger.log(Log.MTD, "tableChanged(TableModelEvent)",
                           cslLogger.level < Log.MTDDETAIL ? "" :
                           "source: " +
                           (e.getSource() == model
                            ? "model" : "filteredModel") +
                           ", start: " + start +
                           ", end: " + end +
                           ", lastRow: " + lastRow);
            }
            if (e.getSource() == model) {
                // A change in the original model - adapt lastRow value
                if (lastRow >= 0) {
                    int type = e.getType();
                    if (start == e.HEADER_ROW
                        || (start == 0 && end >= model.getRowCount()
                            && type == e.UPDATE)) {
                        // complete change
                        if (logging && cslLogger.level >= Log.DBG) {
                            cslLogger.log(Log.DBG,
                                       "tableChanged(TableModelEvent)",
                                       "lastRow invalidated");
                        }
                        lastRow = -1;
                    } else if (type == e.INSERT) {
                        if (start <= lastRow) {
                            lastRow += end - start + 1;
                            if (logging && cslLogger.level >= Log.DBG) {
                                cslLogger.log(Log.DBG,
                                           "tableChanged(TableModelEvent)",
                                           "lastRow moved up");
                            }
                        }
                    } else if (type == e.DELETE) {
                        if (start <= lastRow) {
                            if (end >= lastRow) {
                                lastRow = -1;
                                if (logging && cslLogger.level >= Log.DBG) {
                                    cslLogger.log(Log.DBG,
                                               "tableChanged(TableModelEvent)",
                                               "lastRow invalidated");
                                }
                            } else {
                                lastRow -= end - start + 1;
                                if (logging && cslLogger.level >= Log.DBG) {
                                    cslLogger.log(Log.DBG,
                                               "tableChanged(TableModelEvent)",
                                               "lastRow moved down");
                                }
                            }
                        }
                    }
                }
            } else {
                // A change in the FilteredAndSortedTableModel
                delay.perform(this);
            }
        }

        // }}}
        // {{{ run

        /**
         * Run method for the delayed table changed action.
         */
        public void run()
        {
            if (logging && cslLogger.level >= Log.DBG) {
                cslLogger.log(Log.DBG, "run()",
                           "lastRow: " + lastRow +
                           ", lastColumn: " + lastColumn);
            }
            if (lastRow >= 0) {
                restoreRow(lastRow, lastColumn);
            }
            int row = table.getSelectedRow();
            if (row >= 0 && row < table.getRowCount()) {
                detailView.setEntry
                    ((LogEntry) table.getModel().getValueAt
                     (row, model.COL_ENTRY));
            }
        }

        // }}}
    }

    // }}}

    //----------------------------------------------------------------------
    // Simple focus listener for the table
    //----------------------------------------------------------------------
    // {{{ class InitialFocusListener

    /**
     * Listen for focus gained and select the first row if the selection is
     * empty.
     */
    private class InitialFocusListener
        extends FocusAdapter
    {
        /**
         * Notify the listener that the table has gained focus.
         *
         * @param       e       Ignored.
         */
        public void focusGained(FocusEvent e) {
            if (table.getSelectionModel().getAnchorSelectionIndex() < 0
                && table.getRowCount() > 0) {
                table.setRowSelectionInterval(0, 0);
            }
        }
    }

    // }}}

    //----------------------------------------------------------------------
    // The CellRenderer for the table.
    //----------------------------------------------------------------------
    // {{{ class LevelRenderer

    /**
     * Renderer that displays the icons for the level column and draws a black
     * border around the selected cell instead of the default border.
     */
    public class LevelRenderer extends de.qfs.lib.gui.WindowsTableCellRenderer
    {
        /**
         * Whether the renderer is used for the level column.
         */
        private boolean isLevel;

        /**
         * Border for a selected focused cell.
         */
        private Border selectedFocusBorder;

        /**
         * Border for a non-selected focused cell.
         */
        private Border focusBorder;

        // {{{ constructor

        /**
         * Create a new LevelRenderer.
         *
         * @param       level   Whether the renderer is used for the level
                                column.
         */
        public LevelRenderer (boolean level)
        {
            isLevel = level;
            initIcons();
        }

        // }}}

        // {{{ getTableCellRendererComponent

        /**
         * Get the Component to render the cell.
         *
         * @param       table           The table being painted.
         * @param       value           The cell's value.
         * @param       isSelected      Whether the cell is selected.
         * @param       hasFocus        Whether the cell has the focus.
         * @param       row             The cell's row.
         * @param       column          The cell's column.
         *
         * @return      A label with either an icon for the row's level, or the
         *              text contents of the cell.
         */
        public Component getTableCellRendererComponent
                (JTable table, Object value, boolean isSelected,
                 boolean hasFocus, int row, int column)
        {
            JLabel label = (JLabel) super.getTableCellRendererComponent
                (table, value, isSelected, hasFocus, row, column);
            if (isLevel) {
                int level = ((Integer) value).intValue();
                Icon icon = null;
                if (level >= Log.ERR && level <= Log.DBGDETAIL) {
                    icon = icons[level - 1];
                }
                if (icon != null) {
                    label.setIcon(icon);
                    label.setText(null);
                }
            }
            //  if (hasFocus) {
                //  if (isSelected) {
                    //  if (selectedFocusBorder == null) {
                        //  selectedFocusBorder = BorderFactory.createLineBorder
                            //  (table.getSelectionForeground());
                    //  }
                    //  label.setBorder(selectedFocusBorder);
                //  } else {
                    //  if (focusBorder == null) {
                        //  focusBorder = BorderFactory.createLineBorder
                            //  (table.getForeground());
                    //  }
                    //  label.setBorder(focusBorder);
                //  }
            //  }
            return label;
        }

        // }}}
    }

    // }}}

    //----------------------------------------------------------------------
    // The HeaderCellRenderer for the table.
    //----------------------------------------------------------------------
    // {{{ class HeaderRenderer

    /**
     * Renderer that displays column headers in red, if the correspoding
     * filter is enabled.
     */
    private class HeaderRenderer
        extends SortedTableHeaderCellRenderer
    {
        // {{{ constructor

        /**
         * Create a new HeaderRenderer.
         */
        public HeaderRenderer()
        {
            super(LogTableView.this.table);
        }

        // }}}

        // {{{ getTableCellRendererComponent

        /**
         * Get the Component to render the header.
         *
         * @param       table           The table being painted.
         * @param       value           The cell's value.
         * @param       isSelected      Whether the cell is selected.
         * @param       hasFocus        Whether the cell has the focus.
         * @param       row             The cell's row.
         * @param       column          The cell's column.
         *
         * @return      A label with the text in red, if the column is
         *              currently used as a filter.
         */
        public Component getTableCellRendererComponent
                (JTable table, Object value, boolean isSelected,
                 boolean hasFocus, int row, int column)
        {
            JLabel label = (JLabel) super.getTableCellRendererComponent
                (table, value, isSelected, hasFocus, row, column);
            int idx = table.convertColumnIndexToModel(column);
            if (filterEnabled[idx]) {
                label.setForeground(Color.red);
            } else {
                label.setForeground(Color.black);
            }
            return label;
        }

        // }}}
    }

    // }}}
}
