// {{{ copyright

/********************************************************************
 *
 * $Id: NodeData.java,v 1.8 2000/07/05 14:07:45 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog.logview;

// {{{ imports

import java.io.Serializable;

import java.text.Collator;

// }}}

/**
 * Data structure that holds the class/package and method data for a {@link
 * FilterNode FilterNode} as well as the node's level. <p>
 */
public class NodeData implements Serializable
{
    // {{{ variables

    /**
     * The UID used for serialization.
     */
    static final long serialVersionUID = -2093208988174456189L;

    /**
     * The class/package of the node.
     */
    private String clazz;

    /**
     * The method of the node (may be null).
     */
    private String method;

    /**
     * The level of the node (may be null).
     */
    private Integer level;

    /**
     * Collator used for string comparisons.
     */
    private static Collator coll = Collator.getInstance();

    // }}}

    // {{{ constructor

    /**
     * Create a new NodeData.
     *
     * @param   clazz   The class/package of the node.
     * @param   method  The method of the node (may be null).
     * @param   level   The level of the node (may be null).
     */
    public NodeData (String clazz, String method, Integer level)
    {
        this.clazz = clazz;
        this.method = method;
        this.level = level;
    }

    // }}}

    // {{{ getClazz

    /**
     * Get the node's class/package.
     *
     * @return  The class/package of the node.
     */
    public String getClazz ()
    {
        return clazz;
    }

    // }}}
    // {{{ setClazz

    /**
     * Set the node's class/package.
     *
     * @return  The class/package to set.
     */
    public void setClazz (String clazz)
    {
        this.clazz = clazz;
    }

    // }}}
    // {{{ getMethod

    /**
     * Get the node's method.
     *
     * @return  The method of the node.
     */
    public String getMethod ()
    {
        return method;
    }

    // }}}
    // {{{ setMethod

    /**
     * Set the node's method.
     *
     * @return  The method to set.
     */
    public void setMethod (String method)
    {
        this.method = method;
    }

    // }}}
    // {{{ getLevel

    /**
     * Get the node's level.
     *
     * @return  The level of the node.
     */
    public Integer getLevel ()
    {
        return level;
    }

    // }}}
    // {{{ setLevel

    /**
     * Set the node's level.
     *
     * @return  The level to set.
     */
    public void setLevel (Integer level)
    {
        this.level = level;
    }

    // }}}
    // {{{ hashCode

    /**
     * Generate a hash code for the NodeData.
     *
     * @return  The hash code.
     */
    public int hashCode()
    {
        return (clazz == null ? 0 : clazz.hashCode()) +
            (method == null ? 0 : method.hashCode());
    }

    // }}}
    // {{{ equals

    /**
     * Check whether the NodeData is equal to some object.
     *
     * @param   o       The object to compare to.
     *
     * @return True if the object is a NodeData and has equal clazz and method
     *              members.
     */
    public boolean equals(Object o)
    {
        if (! (o instanceof NodeData)) {
            return false;
        }
        NodeData node = (NodeData) o;
        return ((clazz == null && node.clazz == null)
                || (clazz != null && clazz.equals(node.clazz)))
            && ((method == null && node.method == null)
                || (method != null && method.equals(node.method)));
    }

    // }}}
    // {{{ compareTo

   /**
     * Compare the data to some other object.
     *
     * @param   o       The Object to compare to. Must be a NodeData.
     *
     * @return  -1, 0 or 1 if the NodeData row is less than, equal to or
     *          greater than the object.
     */
    public int compareTo (Object o)
    {
        NodeData data = (NodeData) o;

        // class nodes without package come after package nodes
        boolean pkg1 = clazz.indexOf(".") >= 0;
        boolean pkg2 = data.getClazz().indexOf(".") >= 0;
        if (pkg1 != pkg2) {
            return pkg1 ? -1 : 1;
        }

        int cl = coll.compare(clazz, data.getClazz());
        if (cl != 0) {
            return cl;
        }
        if (method == null) {
            return data.getMethod() == null ? 0 : -1;
        } else if (data.getMethod() == null) {
            return 1;
        }
        return coll.compare(method, data.getMethod());
    }

    // }}}
}
