// {{{ copyright

/********************************************************************
 *
 * $Id: CommandSaveConfiguration.java,v 1.6 2000/07/05 14:07:45 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog.logview.command;

// {{{ imports

import java.awt.Frame;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import java.util.Properties;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;
import javax.swing.tree.TreePath;

import de.qfs.lib.command.CancelCommandException;
import de.qfs.lib.command.Command;
import de.qfs.lib.command.CommandException;
import de.qfs.lib.command.Invokable;
import de.qfs.lib.command.MessageCommandException;
import de.qfs.lib.config.Configuration;
import de.qfs.lib.gui.Message;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.Logger;

import de.qfs.apps.qflog.logview.LogFilterTreeModel;
import de.qfs.apps.qflog.logview.LogLevelTreeModel;
import de.qfs.apps.qflog.logview.LogFileFilter;
import de.qfs.apps.qflog.logview.LogFrame;
import de.qfs.apps.qflog.logview.LogView;

// }}}

/**
 * Save or restore a LogView's or LogFrame's Configuration.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.6 $
 */
public class CommandSaveConfiguration
    extends Command
    implements Invokable
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger (CommandSaveConfiguration.class);

    /**
     * True to save, false to restore.
     */
    private boolean save;

    /**
     * The LogFrame whose configuration should be saved.
     */
    private LogFrame frame;

    /**
     * The LogView whose configuration should be saved.
     */
    private LogView view;

    /**
     * Another frame parent for the dialog.
     */
    private Frame parent;

    // }}}
    // {{{ constructors

    /**
     * Create a new CommandSaveConfiguration for a LogFrame.
     *
     * @param   source  The source of the command.
     * @param   frame   The LogFrame whose configuration should be saved.
     * @param   save    True to save, false to restore.
     */
    public CommandSaveConfiguration(Object source, LogFrame frame,
                                    boolean save)
    {
        super(source);
        parent = frame;
        this.frame = frame;
        view = frame.getView();
        this.save = save;
    }

    /**
     * Create a new CommandSaveConfiguration for a LogView.
     *
     * @param   source  The source of the command.
     * @param   parent  A parent frame for the file dialog.
     * @param   view    The LogView whose configuration should be saved.
     * @param   save    True to save, false to restore.
     */
    public CommandSaveConfiguration(Object source, Frame parent, LogView view,
                                    boolean save)
    {
        super(source);
        this.parent = parent;
        this.view = view;
        this.save = save;
    }

    // }}}

    // {{{ invoke

    /**
     * Invoke the command.
     *
     * @return  Always null.
     *
     * @throws  CommandException        If an I/O error occurs.
     */
    public Object invoke()
        throws CommandException
    {
        boolean excluded = ! view.isLoggingEnabled();
        if (excluded) {
            Log.excludeThread();
        }
        try {
            MyConfiguration cfg = new MyConfiguration ();

            if (frame != null) {
                frame.getConfigurator().copy().register(cfg);
            }
            view.getConfigurator().copy().register(cfg);

            if (save) {
                save(cfg);
            } else {
                restore(cfg);
            }
            return null;
        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}

    // {{{ save

    /**
     * Save the configuration.
     *
     * @param   cfg     The Configuration to save.
     *
     * @throws  CommandException        If an I/O error occurs.
     */
    private void save(Configuration cfg)
        throws CommandException
    {
        JFileChooser fc = new JFileChooser (view.getConfigurationDirectory());

        LogFileFilter filter = new LogFileFilter
            (view.getResources().getString("logView.fileFilter.qlc",
                                           "qflog configuration file"),
             ".qlc");
        fc.setFileFilter(filter);

        if (fc.showSaveDialog(parent) != fc.APPROVE_OPTION) {
            throw new CancelCommandException ();
        }
        File file = filter.forceExtension(fc.getSelectedFile());

        if (file.exists()) {
            if (Message.showMessage(parent, "logView.askOverwrite",
                                    new String[] {file.getName()})
                != JOptionPane.OK_OPTION) {
                throw new CancelCommandException ();
            }
        }

        boolean failed = true;
        if (view.getStatusLine() != null) {
            view.getStatusLine().pushMessage
                (view.getResources().getString
                 ("logView.status.savingConfig",
                  "Saving configuration..."));
        }
        try {
            cfg.save(new FileOutputStream (file));
            failed = false;
            String dir = file.getParent();
            if (dir != null) {
                view.setConfigurationDirectory(new File (dir));
            }
        } catch (IOException ex) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "save(Configuration)", ex);
            }
            throw new MessageCommandException
                (ex.getMessage(), parent, "logView.saveError",
                 new String[] {file.getPath()});
        } finally {
            if (view.getStatusLine() != null) {
                view.getStatusLine().popMessage();
                view.getStatusLine().setTemporaryMessage
                    (view.getResources().getString
                     (failed ? "logView.status.savingConfig.failed"
                      : "logView.status.savingConfig.done",
                      "Saving configuration..." +
                      (failed ? "failed" : "done")));
            }
        }
    }

    // }}}
    // {{{ restore

    /**
     * Restore the configuration.
     *
     * @param   cfg     The Configuration to restore.
     *
     * @throws  CommandException        If an I/O error occurs.
     */
    private void restore(Configuration cfg)
        throws CommandException
    {
        JFileChooser fc = new JFileChooser (view.getConfigurationDirectory());

        LogFileFilter filter = new LogFileFilter
            (view.getResources().getString("logView.fileFilter.qlc",
                                        "qflog configuration file"),
             ".qlc");
        fc.setFileFilter(filter);

        if (fc.showOpenDialog(parent) != fc.APPROVE_OPTION) {
            throw new CancelCommandException ();
        }
        File file = fc.getSelectedFile();

        boolean failed = true;
        view.getStatusLine().pushMessage
            (view.getResources().getString
             ("logView.status.loadingConfig",
              "Restoring configuration..."));
        // If the restore fails we need to reset the current levels
        LogFilterTreeModel filterModel = view.getFilterView().getModel();
        Properties oops1 = filterModel.getConfigurableState();
        filterModel.removeLevel
            (new TreePath (new Object[] {filterModel.getRoot()}), true);
        Properties oops2 = null;
        LogLevelTreeModel levelModel = null;
        if (view.getLevelView() != null) {
            levelModel = view.getLevelView().getModel();
            oops2 = levelModel.getConfigurableState();
            levelModel.removeLogLevel
                (new TreePath (new Object[] {levelModel.getRoot()}), true);
        }
        try {
            cfg.restore(new FileInputStream (file));
            failed = false;
            String dir = file.getParent();
            if (dir != null) {
                view.setConfigurationDirectory(new File (dir));
            }
            view.getFilterView().showSetLevels();
            if (view.getLevelView() != null) {
                view.getLevelView().showSetLevels();
            }
        } catch (IOException ex) {
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "restore(Configuration)", ex);
            }
            filterModel.setConfigurableState(oops1);
            if (levelModel != null) {
                levelModel.setConfigurableState(oops2);
            }
            throw new MessageCommandException
                (ex.getMessage(), parent, "logView.loadError",
                 new String[] {file.getPath()});
        } finally {
            view.getStatusLine().popMessage();
            view.getStatusLine().setTemporaryMessage
                (view.getResources().getString
                 (failed ? "logView.status.loadingConfig.failed"
                  : "logView.status.loadingConfig.done",
                  "Loading configuration..." +
                  (failed ? "failed" : "done")));
        }
    }

    // }}}

    // {{{ class Configuration

    /**
     * Helper class that allows Construction of a separate Configuration
     * instance.
     */
    private class MyConfiguration
        extends Configuration
    {
        /**
         * Create a new MyConfiguration.
         */
        public MyConfiguration()
        {
        }
    }

    // }}}
}
