// {{{ copyright

/********************************************************************
 *
 * $Id: CommandSaveLog.java,v 1.7 2000/07/05 14:07:45 gs Exp $
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is qfs.de code.
 *
 * The Initial Developer of the Original Code is Gregor Schmid.
 * Portions created by Gregor Schmid are
 * Copyright (C) 1999 Quality First Software, Gregor Schmid.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 *******************************************************************/

// }}}

package de.qfs.apps.qflog.logview.command;

// {{{ imports

import java.awt.Frame;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.FileReader;
import java.io.PrintWriter;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;

import java.lang.reflect.InvocationTargetException;

import java.util.Properties;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JTable;
import javax.swing.ProgressMonitor;
import javax.swing.filechooser.FileFilter;

import de.qfs.lib.command.CancelCommandException;
import de.qfs.lib.command.Command;
import de.qfs.lib.command.CommandException;
import de.qfs.lib.command.Invokable;
import de.qfs.lib.command.MessageCommandException;
import de.qfs.lib.gui.Message;
import de.qfs.lib.gui.SwingUtil;
import de.qfs.lib.log.Log;
import de.qfs.lib.log.LogEntry;
import de.qfs.lib.log.Logger;

import de.qfs.apps.qflog.logview.LogFileFilter;
import de.qfs.apps.qflog.logview.LogFrame;
import de.qfs.apps.qflog.logview.LogParser;
import de.qfs.apps.qflog.logview.LogTableModel;
import de.qfs.apps.qflog.logview.LogView;

// }}}

/**
 * Save or restore a LogView's or LogFrame's Logs.
 *
 * @author      Gregor Schmid
 * @version     $Revision: 1.7 $
 */
public class CommandSaveLog
    extends Command
    implements Invokable
{
    // {{{ variables

    /**
     * The Logger used for logging.
     */
    private final static Logger logger =
        new Logger (CommandSaveLog.class);

    /**
     * True to save, false to restore.
     */
    private boolean save;

    /**
     * True to save all messages, false to save only the visible ones.
     */
    private boolean all;

    /**
     * The LogFrame whose logs should be saved.
     */
    private LogFrame frame;

    /**
     * The LogView whose logs should be saved.
     */
    private LogView view;

    /**
     * Another frame parent for the dialog.
     */
    private Frame parent;

    /**
     * The file to save in / load from.
     */
    private File file;

    /**
     * The FileChooser for all CommandSaveLog commands.
     */
    private static JFileChooser fileChooser;

    //  /**
     //  * The FileFilter for binary .qfl files.
     //  */
    //  private static LogFileFilter qflFilter;

    /**
     * The FileFilter for text .log files.
     */
    private static LogFileFilter logFilter;


    // }}}

    // {{{ constructors

    /**
     * Create a new CommandSaveLog for a LogFrame.
     *
     * @param   source  The source of the command.
     * @param   frame   The LogFrame whose logs should be saved.
     * @param   file    The file to save in / load from.
     * @param   save    True to save, false to restore.
     * @param   all     True to save all messages, false to save only
     *                  the visible ones.
     */
    public CommandSaveLog(Object source, LogFrame frame, File file,
                          boolean save, boolean all)
    {
        super(source);
        parent = frame;
        this.frame = frame;
        view = frame.getView();
        this.file = file;
        this.save = save;
        this.all = all;
    }

    /**
     * Create a new CommandSaveLog for a LogView.
     *
     * @param   source  The source of the command.
     * @param   parent  A parent frame for the file dialog.
     * @param   view    The LogView whose logs should be saved.
     * @param   file    The file to save in / load from.
     * @param   save    True to save, false to restore.
     * @param   all     True to save all messages, false to save only
     *                  the visible ones.
     */
    public CommandSaveLog(Object source, Frame parent, LogView view,
                          File file, boolean save, boolean all)
    {
        super(source);
        this.parent = parent;
        this.view = view;
        this.file = file;
        this.save = save;
        this.all = all;
    }

    // }}}

    // {{{ invoke

    public Object invoke()
        throws CommandException
    {
        boolean excluded = ! view.isLoggingEnabled();
        if (excluded) {
            Log.excludeThread();
        }
        try {
            if (save) {
                save();
            } else {
                restore();
            }
            return null;
        } finally {
            if (excluded) {
                Log.includeThread();
            }
        }
    }

    // }}}

    // {{{ save

    private void save()
        throws CommandException
    {
        setupFileChooser();
        //  FileFilter filter;
        if (file == null) {
            fileChooser.setSelectedFile(null);
            fileChooser.rescanCurrentDirectory();
            if (fileChooser.showSaveDialog(parent)
                != fileChooser.APPROVE_OPTION) {
                throw new CancelCommandException ();
            }

            file = fileChooser.getSelectedFile();
            FileFilter filter = fileChooser.getFileFilter();
            if (file.getName().indexOf(".") < 0
                && filter instanceof LogFileFilter) {
                file = ((LogFileFilter) filter).forceExtension(file);
            }
            if (file.exists()) {
                if (Message.showMessage(parent, "logView.askOverwrite",
                                        new String[] {file.getName()})
                    != JOptionPane.YES_OPTION) {
                    throw new CancelCommandException ();
                }
            }
        //  } else {
            //  filter = file.getName().endsWith(".qfl") ? qflFilter : logFilter;
        }

        boolean failed = true;
        if (view.getStatusLine() != null) {
            view.getStatusLine().pushMessage
                (view.getResources().getString
                 ("logView.status.savingLog",
                  "Saving log file..."));
        }
        try {
            LogEntry[] data;
            if (all) {
                data = view.getTableModel().getData();
            } else {
                JTable table = view.getTableView().getTable();
                int rows = table.getRowCount();
                data = new LogEntry[rows];
                for (int i = 0; i < rows; i++) {
                    data[i] = (LogEntry)
                        table.getValueAt(i, LogTableModel.COL_ENTRY);
                }
            }
            //  if (filter == qflFilter) {
                //  // binary
                //  ObjectOutputStream oos = new ObjectOutputStream
                    //  (new FileOutputStream (file));
                //  oos.writeInt(view.SAVE_FILE_VERSION);
                //  oos.writeObject(new Object[] {view.getClientName()});
                //  for (int i = 0; i < data.length; i++) {
                    //  oos.writeObject(data[i]);
                //  }
                //  oos.close();
                //  failed = false;
            //  } else {
                // text
                PrintWriter pw = new PrintWriter
                    (new FileWriter (file));
                if (view.getClientName() != null) {
                    pw.println("client name: " + view.getClientName());
                }
                for (int i = 0; i < data.length; i++) {
                    pw.println(data[i].toString());
                }
                pw.close();
                failed = false;
            //  }
        } catch (IOException ex) {
            if (logger.level >= Log.ERR) {
                logger.log("save()", ex);
            }
            throw new MessageCommandException
                (ex.getMessage(), parent, "logView.saveError",
                 new String[] {file.getPath()});
        } finally {
            if (view.getStatusLine() != null) {
                view.getStatusLine().popMessage();
                view.getStatusLine().setTemporaryMessage
                    (view.getResources().getString
                     (failed ? "logView.status.savingLog.failed"
                      : "logView.status.savingLog.done",
                      "Saving log file..." + (failed ? "failed" : "done")));
            }
        }
    }

    // }}}
    // {{{ restore

    private void restore()
        throws CommandException
    {
        setupFileChooser();
        //  FileFilter filter = null;
        if (file == null) {
            fileChooser.setSelectedFile(null);
            fileChooser.rescanCurrentDirectory();
            if (fileChooser.showOpenDialog(parent)
                != fileChooser.APPROVE_OPTION) {
                throw new CancelCommandException ();
            }

            file = fileChooser.getSelectedFile();
            //  filter = fileChooser.getFileFilter();
        }
        if (file == null) {
                throw new CancelCommandException ();
        }
        if (!file.exists()) {
            throw new MessageCommandException
                ("No such file: " + file.getPath(), parent,
                 "logView.noSuchFile", new String[] {file.getPath()});
        }

        boolean failed = true;
        if (view.getStatusLine() != null) {
            view.getStatusLine().pushMessage
                (view.getResources().getString
                 ("logView.status.loadingLog",
                  "Loading log file..."));
        }
        try {
            //  if ((filter == qflFilter)
                //  || (filter != logFilter
                    //  && ! file.getName().endsWith(".log"))) {
                //  // try binary
                //  if (readBinary(file)) {
                    //  // ok
                    //  failed = false;
                    //  return;
                //  }
                //  System.err.println("binary failed");
                //  if (filter == qflFilter
                    //  || (filter != logFilter
                        //  && file.getName().endsWith(".qfl"))) {
                    //  // no use to try text in this case
                    //  System.err.println("giving up filter:" + filter +
                                       //  ", file: " + file);
                    //  throw new MessageCommandException
                        //  ("error loading " + file.getPath(),
                         //  parent, "logView.loadError",
                         //  new String[] {file.getPath()});
                //  }
            //  }
            // try text
            if (readText(file)) {
                // ok
                failed = false;
                view.setState(view.STATE_FILE);
                view.setFile(file);
                return;
            }
            throw new MessageCommandException
                ("cannot load file " + file.getPath(),
                 parent, "logView.loadError",
                 new String[] {file.getPath()});
        } finally {
            if (view.getStatusLine() != null) {
                view.getStatusLine().popMessage();
                view.getStatusLine().setTemporaryMessage
                    (view.getResources().getString
                     (failed ? "logView.status.loadingLog.failed"
                      : "logView.status.loadingLog.done",
                      "Loading log file..." + (failed ? "failed" : "done")));
            }
        }
    }

    // }}}
    // {{{ readBinary

    /**
     * Try to restore data from a binary file.
     *
     * @param   file    The file to restore from.
     *
     * @return  True if the restore completed, false if it failed.
     */
    //  private boolean readBinary(File file)
    //  {
        //  try {
            //  ArrayList data = new ArrayList ();
            //  ObjectInputStream ois = new ObjectInputStream
                //  (new FileInputStream (file));
            //  int version = ois.readInt();
            //  Object[] info = (Object[]) ois.readObject();
            //  LogEntry entry = (LogEntry) ois.readObject();
            //  while (entry != null) {
                //  data.add(entry);
                //  entry = (LogEntry) ois.readObject();
            //  }
            //  ois.close();
            //  view.setClientName((String) info[0]);
            //  view.getTableModel().setData
                //  ((LogEntry[]) data.toArray(new LogEntry[0]));
        //  } catch (Exception ex) {
            //  // IOException, ClassCastException or ClassNotFoundException
            //  if (logger.level >= Log.WRN) {
                //  logger.log(Log.WRN, "readBinary(File)", ex);
            //  }
            //  return false;
        //  }
        //  return true;
    //  }

    // }}}
    // {{{ readText

    /**
     * Try to restore data from a text file.
     *
     * @param   file    The file to restore from.
     *
     * @return  True if the restore completed, false if it failed.
     */
    private boolean readText(File file)
    {
        int max = (int) file.length();
        final ProgressMonitor pm =
            new ProgressMonitor (parent, "parsing...", null, 0, max);

        try {
            final BufferedReader br = new BufferedReader
                (new FileReader (file));
            br.mark(4090);
            String line = br.readLine();
            if (line == null) {
                // can't even start
                return false;
            }
            String name = null;
            if (line.startsWith("client name: ")) {
                name = line.substring("client name: ".length());
            } else {
                br.reset();
            }
            final LogParser parser = new LogParser ();
            parser.setProgressCallback(new LogParser.ProgressCallback() {
                private int count = 0;
                public boolean progress(int lines, int chars)
                {
                    if (count++ % 50 == 0) {
                        pm.setProgress(chars);
                        Thread.currentThread().yield();
                        return ! pm.isCanceled();
                    } else {
                        return true;
                    }
                }
            });
            LogEntry[] data = (LogEntry[])
                SwingUtil.withEvents(new SwingUtil.UnsafeRunnable () {
                    public Object run()
                    throws IOException
                    {
                        return
                        parser.parse(br);
                    }
                }, SwingUtil.NO_EVENTS, SwingUtil.NO_EVENTS);
            br.close();
            view.setClientName(name);
            view.getTableModel().setMaxRows(0);
            view.getTableModel().setData(data);
        } catch (IOException ex) {
            ex.printStackTrace();
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "readText(File)", ex);
            }
            return false;
        } catch (InvocationTargetException ex) {
            ex.getTargetException().printStackTrace();
            if (logger.level >= Log.WRN) {
                logger.log(Log.WRN, "readText(File)", ex.getTargetException());
            }
            return false;
        } finally {
            pm.close();
        }
        return true;
    }

    // }}}

    // {{{ setupFileChooser

    /**
     * Prepare the common FileChooser.
     */
    private void setupFileChooser()
    {
        if (fileChooser != null) {
            return;
        }

        fileChooser = new JFileChooser (new File ("."));

        //  qflFilter = new LogFileFilter
            //  (view.getResources().getString("logView.fileFilter.qfl",
                                           //  "qflog binary log file"),
             //  ".qfl");
        logFilter = new LogFileFilter
            (view.getResources().getString("logView.fileFilter.log",
                                           "qflog text log file"),
             ".log");
        //  fileChooser.addChoosableFileFilter(qflFilter);
        fileChooser.setFileFilter(logFilter);
    }

    // }}}
}
